#include "graphics.hh"

// std::ofstream *gDebug;

const int GAME_WIDTH = 1024;
const int GAME_HEIGHT = 768;

const double LINE_WIDTH = 1.5;
const int FONT_SIZE = 18;

std::vector<Window *> gWindows;

std::map<std::string, cairo_surface_t *> gImages;

Question *gActiveQuestion = NULL;

SDL_AudioDeviceID gAudioRecordingDevice = 0, gAudioPlaybackDevice = 0;

SDL_AudioSpec gRecordingAudioSpec, gPlaybackAudioSpec;

OpusDecoder *gDecoder = NULL;
OpusDecoder *gDecoderPlayer2 = NULL;
OpusEncoder *gEncoder = NULL;

// std::ofstream gAudioFile[2];
WAV gAudioFile[2];

unsigned long gSilenceThreshold = 20;
std::array<float,100> gMaxSignals;
std::array<float,100> gSortedMaxSignals;
unsigned long gCurrentMaxSignal = 0;

void dumpAudioSettings(const std::string &type, const char *deviceName, const SDL_AudioSpec &spec) {
  std::cout << std::endl << type << " Audio Settings" << std::endl;
    std::cout << "+---------+---------------" << std::endl;
    std::cout << "|  device | "<< deviceName << std::endl;
    std::cout << "|  format | "<< spec.format << std::endl;
    std::cout << "|    freq | "<< spec.freq << std::endl;
    std::cout << "|channels | "<< int(spec.channels) << std::endl;
    std::cout << "| samples | "<< spec.samples << std::endl;
    std::cout << "+-----------------------" << std::endl;
}

void listAudioDevices() {
  int playbackCount = SDL_GetNumAudioDevices(SDL_FALSE);
  std::cout << std::endl << "Audio Devices" << std::endl;
  std::cout << "+-----+-----+-------------------------------------" << std::endl;
  std::cout << "|Type | Idx | Device Name" << std::endl;
  std::cout << "+-----+-----+-------------------------------------" << std::endl;
  for (int i = 0; i < playbackCount; ++i) {
    std::cout << "|Play | " << std::setfill(' ') << std::setw(3) << i << std::setw(0) << " | " << SDL_GetAudioDeviceName(i, SDL_FALSE) << std::endl;
  }
  int recordCount = SDL_GetNumAudioDevices(SDL_TRUE);
  for (int i = 0; i < recordCount; ++i) {
    std::cout << "|Rec  | " << std::setfill(' ') << std::setw(3) << i << std::setw(0) << " | " << SDL_GetAudioDeviceName(i, SDL_TRUE) << std::endl;
  }
  std::cout << "+-----+-----+-------------------------------------" << std::endl;
}

bool initOpus() {
  int err;
  gEncoder = opus_encoder_create( 48000, 1, OPUS_APPLICATION_VOIP, &err );
  if( err || gEncoder == NULL ) {
    std::cout << "Error initializing encoder: " << err << std::endl;
    return false;
  }

  gDecoder = opus_decoder_create( 48000, 1, &err );
  if( err || gDecoder == NULL ) {
    std::cout << "Error initializing decoder: " << err << std::endl;
    return false;
  }
  gDecoderPlayer2 = opus_decoder_create( 48000, 1, &err );
  if( err || gDecoderPlayer2 == NULL ) {
    std::cout << "Error initializing player 2 decoder: " << err << std::endl;
    return false;
  }


  return true;
}

void initAudio(int playIdx, int recIdx, int freq, int format, int samples) {
  SDL_AudioSpec want;

  gMaxSignals.fill(0.0);

  listAudioDevices();

  SDL_memset(&want, 0, sizeof(want)); /* or SDL_zero(want) */
  want.freq = freq; //48000;
  // want.freq = 8192;
  // want.format = AUDIO_F32;
  want.format = format; // AUDIO_U8;
  want.channels = 1; // 2
  want.samples = samples; // 2048; // 4096;
  want.callback = NULL;

  if (playIdx < 0 || playIdx >= SDL_GetNumAudioDevices(SDL_FALSE)) playIdx = 0;
  if (recIdx < 0 || recIdx >= SDL_GetNumAudioDevices(SDL_TRUE)) recIdx = 0;

  gAudioRecordingDevice = SDL_OpenAudioDevice(SDL_GetAudioDeviceName(recIdx, SDL_TRUE), SDL_TRUE, &want, &gRecordingAudioSpec, SDL_TRUE);
  if (gAudioRecordingDevice == 0) {
    SDL_Log("Failed to open audio recording device: %s", SDL_GetError());
  } else {
    dumpAudioSettings( "Recording", SDL_GetAudioDeviceName(recIdx, SDL_TRUE), gRecordingAudioSpec );
  }
  SDL_PauseAudioDevice( gAudioRecordingDevice, SDL_TRUE);

  gAudioPlaybackDevice = SDL_OpenAudioDevice(SDL_GetAudioDeviceName(playIdx, SDL_FALSE), SDL_FALSE, &want, &gPlaybackAudioSpec, SDL_TRUE);
  if (gAudioPlaybackDevice == 0) {
    SDL_Log("Failed to open audio playback device: %s", SDL_GetError());
  } else {
    dumpAudioSettings( "Playback", SDL_GetAudioDeviceName(playIdx, SDL_FALSE), gPlaybackAudioSpec );
  }
  SDL_PauseAudioDevice( gAudioPlaybackDevice, SDL_TRUE);
  initOpus();
}

void shutdownAudio() {
  if (gAudioRecordingDevice != 0) SDL_CloseAudioDevice( gAudioRecordingDevice );
  if (gAudioPlaybackDevice != 0) SDL_CloseAudioDevice( gAudioPlaybackDevice );

  if( gEncoder ) opus_encoder_destroy( gEncoder );
  if( gDecoder ) opus_decoder_destroy( gDecoder );
  if( gDecoderPlayer2 ) opus_decoder_destroy( gDecoderPlayer2 );
  gEncoder = NULL;
  gDecoder = NULL;
  gDecoderPlayer2 = NULL;
}

void unpauseAudioRecording(const v8::FunctionCallbackInfo<v8::Value> &args) {
  if (gAudioRecordingDevice == 0) return;
  SDL_PauseAudioDevice( gAudioRecordingDevice, SDL_FALSE);
  SDL_ClearQueuedAudio( gAudioPlaybackDevice );
  SDL_PauseAudioDevice( gAudioPlaybackDevice, SDL_FALSE);
}

void pauseAudioRecording(const v8::FunctionCallbackInfo<v8::Value> &args) {
  if (gAudioRecordingDevice == 0) return;
  SDL_PauseAudioDevice( gAudioRecordingDevice, SDL_TRUE);
  SDL_PauseAudioDevice( gAudioPlaybackDevice, SDL_TRUE);
}

void unpauseAudioPlayback(const v8::FunctionCallbackInfo<v8::Value> &args) {
  if (gAudioPlaybackDevice == 0) return;
  SDL_PauseAudioDevice( gAudioPlaybackDevice, SDL_FALSE);
}

void pauseAudioPlayback(const v8::FunctionCallbackInfo<v8::Value> &args) {
  if (gAudioPlaybackDevice == 0) return;
  SDL_PauseAudioDevice( gAudioPlaybackDevice, SDL_FALSE);
}

void startTextInput(const v8::FunctionCallbackInfo<v8::Value> &args) {
  SDL_StartTextInput();
}

void stopTextInput(const v8::FunctionCallbackInfo<v8::Value> &args) {
  SDL_StopTextInput();
}

static std::string audioStatus(SDL_AudioDeviceID dev) {
    switch (SDL_GetAudioDeviceStatus(dev))
    {
    case SDL_AUDIO_STOPPED: return "stopped";
    case SDL_AUDIO_PLAYING: return "playing";
    case SDL_AUDIO_PAUSED: return "paused";
    default: return "???";
    }
}

void playAudioData(const float *samples, int numSamples) {
  if (gAudioPlaybackDevice == 0) return;
  bool addToQueue = true;

  if( SDL_GetQueuedAudioSize( gAudioPlaybackDevice ) > gPlaybackAudioSpec.size * 2) {
    if( SDL_GetQueuedAudioSize( gAudioPlaybackDevice ) > gPlaybackAudioSpec.size * 10 ) {
      // std::cout << "Too much audio backlog. clear the queue!" << std::endl;
      SDL_ClearQueuedAudio( gAudioPlaybackDevice );
    } else {
      // Audio playback is falling behind. We need to catch up in a way
      // that preserves audio quality. So, if the new audio samples are
      // considered silence, then just don't queue them.
      if( gPlaybackAudioSpec.format == AUDIO_F32 ) {
        // float *samples = (float *)data.data();
        // int n = data.size() / 4;
        float maxv = 0;
        for (int i=0; i<numSamples; i++) {
          float s = abs(samples[i]);
          if(s > maxv) maxv = s;
        }
        gMaxSignals[gCurrentMaxSignal] = maxv;
        gCurrentMaxSignal += 1;
        if (gCurrentMaxSignal >= gMaxSignals.size()) gCurrentMaxSignal = 0;

        std::copy(gMaxSignals.begin(), gMaxSignals.end(), gSortedMaxSignals.begin());
        std::sort(gSortedMaxSignals.begin(), gSortedMaxSignals.end());
        addToQueue = maxv > gSortedMaxSignals[gSilenceThreshold];

        // std::cout << "Skip audio " << maxv << " : " << gSortedMaxSignals[0] << " - " << gSortedMaxSignals[99] << std::endl;

        // std::cout << audioStatus(gAudioPlaybackDevice) << " - " << SDL_GetQueuedAudioSize( gAudioPlaybackDevice ) << " " << gPlaybackAudioSpec.size << " min: " << minv << " max: " << maxv << std::endl;
        // addToQueue = !silence;
      } else {
        // TODO: calculate average for other audio formats.
        addToQueue = false;
      }
    }
  }
  if( addToQueue ) {
    SDL_QueueAudio( gAudioPlaybackDevice, (void*)samples, numSamples * sizeof(float) );
  } else {
    // std::cout << "Skip this audio sample" << std::endl;
  }
}

void playAudioData(const std::string &data) {
  float *samples = (float *)data.data();
  int n = data.size() / 4;
  playAudioData(samples, n);
}

void initGraphics(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();

  v8::Local<v8::String> arg1 = v8::Local<v8::String>::Cast(args[0]);
  std::string fontpath = std::string(*v8::String::Utf8Value(isolate, arg1->ToString()));
  v8::Local<v8::Object> audio = v8::Local<v8::Object>::Cast(args[1]);
  std::cout << "Loading font from: " << fontpath << "\n";
  loadCairoFont(fontpath.c_str());
  initWireframes();

  if( SDL_Init( SDL_INIT_VIDEO | SDL_INIT_AUDIO ) < 0 ) {
    std::cout << "SDL could not initialize! SDL_Error: " << SDL_GetError() << "\n";
    return;
  }

  int recIdx = 0, playIdx = 0, freq = 48000, format = AUDIO_F32, samples = 4096;

  if (!audio->Has(v8::String::NewFromUtf8(isolate, "disabled", v8::NewStringType::kNormal).ToLocalChecked()) ||
      audio->Get(v8::String::NewFromUtf8(isolate, "disabled"))->IsFalse()) {

    if( audio->Has(v8::String::NewFromUtf8(isolate, "freq", v8::NewStringType::kNormal).ToLocalChecked())) {
      freq = v8::Local<v8::Number>::Cast(audio->Get(v8::String::NewFromUtf8(isolate, "freq")))->Value();
    }

    if( audio->Has(v8::String::NewFromUtf8(isolate, "format", v8::NewStringType::kNormal).ToLocalChecked())) {
      std::string tmp = std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::Number>::Cast(audio->Get(v8::String::NewFromUtf8(isolate, "format")))));
      if (tmp == "u8") format = AUDIO_U8;
      else if (tmp == "s16") format = AUDIO_S16;
      else if (tmp == "f32") format = AUDIO_F32;
    }

    if( audio->Has(v8::String::NewFromUtf8(isolate, "samples", v8::NewStringType::kNormal).ToLocalChecked())) {
      int tmp = v8::Local<v8::Number>::Cast(audio->Get(v8::String::NewFromUtf8(isolate, "samples")))->Value();
      if (tmp > 0) samples = tmp;
    }

    if( audio->Has(v8::String::NewFromUtf8(isolate, "playbackDevice", v8::NewStringType::kNormal).ToLocalChecked())) {
      int tmp = v8::Local<v8::Number>::Cast(audio->Get(v8::String::NewFromUtf8(isolate, "playbackDevice")))->Value();
      if (tmp > 0) playIdx = tmp;
    }

    if( audio->Has(v8::String::NewFromUtf8(isolate, "recordDevice", v8::NewStringType::kNormal).ToLocalChecked())) {
      int tmp = v8::Local<v8::Number>::Cast(audio->Get(v8::String::NewFromUtf8(isolate, "recordDevice")))->Value();
      if (tmp > 0) recIdx = tmp;
    }

    if( audio->Has(v8::String::NewFromUtf8(isolate, "threshold", v8::NewStringType::kNormal).ToLocalChecked())) {
      float tmp = v8::Local<v8::Number>::Cast(audio->Get(v8::String::NewFromUtf8(isolate, "threshold")))->Value();
      if (tmp > 0) gSilenceThreshold = tmp;
    }


    initAudio(playIdx, recIdx, freq, format, samples);
  }

  // if( !( IMG_Init( IMG_INIT_PNG ) & IMG_INIT_PNG )) {
  //   gDebug << "SDL_image could not initialize! SDL_image Error: %s\n", IMG_GetError() );
  //   return;
  // }

  args.GetReturnValue().Set(v8::String::NewFromUtf8(isolate, "graphics init done"));
}

void shutdownGraphics(const v8::FunctionCallbackInfo<v8::Value> &args) {

    for (size_t i=0; i<gWindows.size(); i++) {
      delete gWindows[i];
    }
    gWindows.clear();

    shutdownAudio();

    //Quit SDL subsystems
    SDL_Quit();
}

void createWindows(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::Boolean> fullscreen = v8::Local<v8::Boolean>::Cast(args[0]);
  std::string basename = std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(args[1])->ToString()));
  v8::Local<v8::Boolean> sync = v8::Local<v8::Boolean>::Cast(args[2]);

  int displays = SDL_GetNumVideoDisplays();
  int n = std::min(displays, 2);

  //Create window
  for (int i=0; i<n; i++) {
    std::string name = basename;
    name += '1' + i;
    Window *w = new Window(name, GAME_WIDTH, GAME_HEIGHT, fullscreen->Value(), sync->Value());
    gWindows.push_back(w);
  }

  SDL_ShowCursor( SDL_DISABLE );

  // v8::Isolate* isolate = args.GetIsolate();
  // args.GetReturnValue().Set(v8::String::NewFromUtf8(isolate, "bam window"));
}

void setFullscreen(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Local<v8::Boolean> fullscreen = v8::Local<v8::Boolean>::Cast(args[0]);
  for( size_t i=0; i<gWindows.size(); i++ ) {
    gWindows[i]->setFullscreen( fullscreen->Value() );
  }
}

void numWindows(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();

  args.GetReturnValue().Set(v8::Number::New(isolate, gWindows.size()));
}

void handleSizeChange() {
  for (size_t i=0; i<gWindows.size(); i++) {
    gWindows[i]->handleSizeChange();
  }
}

void clearActiveQuestion ( bool force ) {
  if( gActiveQuestion ) {
    gActiveQuestion->finish( force );
    delete gActiveQuestion;
    gActiveQuestion = NULL;
  }
}

void processEvents(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::Function> cb = v8::Local<v8::Function>::Cast(args[0]);
  const unsigned argc = 1;

  SDL_Event event;
  while (SDL_PollEvent(&event)) {
    if( gActiveQuestion && gActiveQuestion->handleEvent( event )) {
      QuestionDoneCode c = gActiveQuestion->getDoneCode();

      if( c == QuestionSubmitted ) {
        v8::Local<v8::Object> obj = v8::Object::New(isolate);
        obj->Set(v8::String::NewFromUtf8(isolate, "type"), v8::String::NewFromUtf8(isolate, "questionSubmitted"));
        obj->Set(v8::String::NewFromUtf8(isolate, "question"), v8::String::NewFromUtf8(isolate, gActiveQuestion->getName().c_str()));
        obj->Set(v8::String::NewFromUtf8(isolate, "answer"), v8::String::NewFromUtf8(isolate, gActiveQuestion->getAnswer().c_str()));
        v8::Local<v8::Value> argv[argc] = { obj };

        clearActiveQuestion( false );
        cb->Call(v8::Null(isolate), argc, argv);
      } else if (c == QuestionGoToPrevious ) {
        v8::Local<v8::Object> obj = v8::Object::New(isolate);
        obj->Set(v8::String::NewFromUtf8(isolate, "type"), v8::String::NewFromUtf8(isolate, "questionGoToPrevious"));
        obj->Set(v8::String::NewFromUtf8(isolate, "question"), v8::String::NewFromUtf8(isolate, gActiveQuestion->getName().c_str()));
        v8::Local<v8::Value> argv[argc] = { obj };

        clearActiveQuestion( false );
        cb->Call(v8::Null(isolate), argc, argv);
      }
    } else {
      v8::Local<v8::Object> obj = v8::Object::New(isolate);
      if (event.type == SDL_WINDOWEVENT) {
        obj->Set(v8::String::NewFromUtf8(isolate, "type"), v8::String::NewFromUtf8(isolate, "windowEvent"));
        obj->Set(v8::String::NewFromUtf8(isolate, "windowID"), v8::Number::New(isolate, event.window.windowID));
        if (event.window.event == SDL_WINDOWEVENT_SHOWN) {
          obj->Set(v8::String::NewFromUtf8(isolate, "windowEvent"), v8::String::NewFromUtf8(isolate, "shown"));
        } else if (event.window.event == SDL_WINDOWEVENT_HIDDEN) {
          obj->Set(v8::String::NewFromUtf8(isolate, "windowEvent"), v8::String::NewFromUtf8(isolate, "hidden"));
        } else if (event.window.event == SDL_WINDOWEVENT_EXPOSED) {
          obj->Set(v8::String::NewFromUtf8(isolate, "windowEvent"), v8::String::NewFromUtf8(isolate, "exposed"));
        } else if (event.window.event == SDL_WINDOWEVENT_MOVED) {
          obj->Set(v8::String::NewFromUtf8(isolate, "windowEvent"), v8::String::NewFromUtf8(isolate, "moved"));
        } else if (event.window.event == SDL_WINDOWEVENT_RESIZED) {
          obj->Set(v8::String::NewFromUtf8(isolate, "windowEvent"), v8::String::NewFromUtf8(isolate, "resized"));
        } else if (event.window.event == SDL_WINDOWEVENT_SIZE_CHANGED) {
          handleSizeChange();
          obj->Set(v8::String::NewFromUtf8(isolate, "windowEvent"), v8::String::NewFromUtf8(isolate, "changed"));
        } else if (event.window.event == SDL_WINDOWEVENT_MINIMIZED) {
          obj->Set(v8::String::NewFromUtf8(isolate, "windowEvent"), v8::String::NewFromUtf8(isolate, "minimized"));
        } else if (event.window.event == SDL_WINDOWEVENT_MAXIMIZED) {
          obj->Set(v8::String::NewFromUtf8(isolate, "windowEvent"), v8::String::NewFromUtf8(isolate, "maximized"));
        } else if (event.window.event == SDL_WINDOWEVENT_RESTORED) {
          obj->Set(v8::String::NewFromUtf8(isolate, "windowEvent"), v8::String::NewFromUtf8(isolate, "restored"));
        } else if (event.window.event == SDL_WINDOWEVENT_ENTER) {
          obj->Set(v8::String::NewFromUtf8(isolate, "windowEvent"), v8::String::NewFromUtf8(isolate, "enter"));
        } else if (event.window.event == SDL_WINDOWEVENT_LEAVE) {
          obj->Set(v8::String::NewFromUtf8(isolate, "windowEvent"), v8::String::NewFromUtf8(isolate, "leave"));
        } else if (event.window.event == SDL_WINDOWEVENT_FOCUS_GAINED) {
          obj->Set(v8::String::NewFromUtf8(isolate, "windowEvent"), v8::String::NewFromUtf8(isolate, "focusGained"));
        } else if (event.window.event == SDL_WINDOWEVENT_FOCUS_LOST) {
          obj->Set(v8::String::NewFromUtf8(isolate, "windowEvent"), v8::String::NewFromUtf8(isolate, "focusLost"));
        } else if (event.window.event == SDL_WINDOWEVENT_CLOSE) {
          obj->Set(v8::String::NewFromUtf8(isolate, "windowEvent"), v8::String::NewFromUtf8(isolate, "close"));
        } else if (event.window.event == SDL_WINDOWEVENT_TAKE_FOCUS) {
          obj->Set(v8::String::NewFromUtf8(isolate, "windowEvent"), v8::String::NewFromUtf8(isolate, "takeFocus"));
        } else if (event.window.event == SDL_WINDOWEVENT_HIT_TEST) {
          obj->Set(v8::String::NewFromUtf8(isolate, "windowEvent"), v8::String::NewFromUtf8(isolate, "hitTest"));
        } else {
          obj->Set(v8::String::NewFromUtf8(isolate, "windowEvent"), v8::Number::New(isolate, event.window.event));
        }
      } else if (event.type == SDL_QUIT) {
        obj->Set(v8::String::NewFromUtf8(isolate, "type"), v8::String::NewFromUtf8(isolate, "quit"));
      } else if (event.type == SDL_TEXTINPUT) {
        obj->Set(v8::String::NewFromUtf8(isolate, "type"), v8::String::NewFromUtf8(isolate, "textInput"));
        obj->Set(v8::String::NewFromUtf8(isolate, "text"), v8::String::NewFromUtf8(isolate, event.text.text));
      } else if (event.type == SDL_TEXTEDITING) {
        obj->Set(v8::String::NewFromUtf8(isolate, "type"), v8::String::NewFromUtf8(isolate, "textEditing"));
        obj->Set(v8::String::NewFromUtf8(isolate, "text"), v8::String::NewFromUtf8(isolate, event.edit.text));
        obj->Set(v8::String::NewFromUtf8(isolate, "start"), v8::Integer::New(isolate, event.edit.start));
        obj->Set(v8::String::NewFromUtf8(isolate, "length"), v8::Integer::New(isolate, event.edit.length));
      } else if (event.type == SDL_KEYDOWN || event.type == SDL_KEYUP) {
        if (event.type == SDL_KEYDOWN)
          obj->Set(v8::String::NewFromUtf8(isolate, "type"), v8::String::NewFromUtf8(isolate, "keyDown"));
        else
          obj->Set(v8::String::NewFromUtf8(isolate, "type"), v8::String::NewFromUtf8(isolate, "keyUp"));
        obj->Set(v8::String::NewFromUtf8(isolate, "sym"), v8::Number::New(isolate, event.key.keysym.sym));
        obj->Set(v8::String::NewFromUtf8(isolate, "name"), v8::String::NewFromUtf8(isolate, SDL_GetKeyName(event.key.keysym.sym)));
        obj->Set(v8::String::NewFromUtf8(isolate, "repeat"), v8::Boolean::New(isolate, event.key.repeat));
      } else if (event.type == SDL_MOUSEMOTION) {
        obj->Set(v8::String::NewFromUtf8(isolate, "type"), v8::String::NewFromUtf8(isolate, "mouseMotion"));
        obj->Set(v8::String::NewFromUtf8(isolate, "state"), v8::Number::New(isolate, event.motion.state));
        obj->Set(v8::String::NewFromUtf8(isolate, "x"), v8::Number::New(isolate, event.motion.x));
        obj->Set(v8::String::NewFromUtf8(isolate, "y"), v8::Number::New(isolate, event.motion.y));
        obj->Set(v8::String::NewFromUtf8(isolate, "xrel"), v8::Number::New(isolate, event.motion.xrel));
        obj->Set(v8::String::NewFromUtf8(isolate, "yrel"), v8::Number::New(isolate, event.motion.yrel));
      } else if (event.type == SDL_MOUSEBUTTONDOWN || event.type == SDL_MOUSEBUTTONUP) {
        if (event.type == SDL_MOUSEBUTTONDOWN)
          obj->Set(v8::String::NewFromUtf8(isolate, "type"), v8::String::NewFromUtf8(isolate, "mouseDown"));
        else
          obj->Set(v8::String::NewFromUtf8(isolate, "type"), v8::String::NewFromUtf8(isolate, "mouseUp"));
        obj->Set(v8::String::NewFromUtf8(isolate, "button"), v8::Number::New(isolate, event.button.button));
        obj->Set(v8::String::NewFromUtf8(isolate, "x"), v8::Number::New(isolate, event.button.x));
        obj->Set(v8::String::NewFromUtf8(isolate, "y"), v8::Number::New(isolate, event.button.y));
      } else {
        obj->Set(v8::String::NewFromUtf8(isolate, "type"), v8::Number::New(isolate, event.type));
      }

      v8::Local<v8::Value> argv[argc] = { obj };
      cb->Call(v8::Null(isolate), argc, argv);
    }
  }
}

void setupGameArea(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Local<v8::Integer> i = v8::Local<v8::Integer>::Cast(args[0]);

  cairo_translate(gWindows[i->Value()]->mCtx, gWindows[i->Value()]->getWidth()/2, gWindows[i->Value()]->getHeight()/2);
  cairo_scale(gWindows[i->Value()]->mCtx, 1, -1);
}

void takeDownGameArea(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Local<v8::Integer> i = v8::Local<v8::Integer>::Cast(args[0]);

  cairo_scale(gWindows[i->Value()]->mCtx, 1, -1);
  cairo_translate(gWindows[i->Value()]->mCtx, -gWindows[i->Value()]->getWidth()/2, -gWindows[i->Value()]->getHeight()/2);
}


void startDrawing(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Local<v8::Integer> i = v8::Local<v8::Integer>::Cast(args[0]);

  clearActiveQuestion( true );

  gWindows[i->Value()]->createContext();
  cairo_set_line_width( gWindows[i->Value()]->mCtx, LINE_WIDTH);
}

void clearScreen(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Local<v8::Integer> i = v8::Local<v8::Integer>::Cast(args[0]);
  v8::Local<v8::Integer> color = v8::Local<v8::Integer>::Cast(args[1]);

  cairo_set_source_rgb( gWindows[i->Value()]->mCtx, ((color->Value() >> 16)&0xFF)/256.0, ((color->Value() >> 8)&0xFF)/256.0, (color->Value()&0xFF)/256.0 );
  cairo_paint( gWindows[i->Value()]->mCtx );
}

void finishDrawing(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Local<v8::Integer> i = v8::Local<v8::Integer>::Cast(args[0]);
  gWindows[i->Value()]->destroyContext();
}

void flip(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Local<v8::Integer> window = v8::Local<v8::Integer>::Cast(args[0]);
  SDL_RenderPresent(gWindows[window->Value()]->mRenderer);
}

void clipDrawingArea(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Local<v8::Integer> window = v8::Local<v8::Integer>::Cast(args[0]);
  v8::Local<v8::Number> x = v8::Local<v8::Number>::Cast(args[1]);
  v8::Local<v8::Number> y = v8::Local<v8::Number>::Cast(args[2]);
  v8::Local<v8::Number> width = v8::Local<v8::Number>::Cast(args[3]);
  v8::Local<v8::Number> height = v8::Local<v8::Number>::Cast(args[4]);

  Window *w = gWindows[window->Value()];

  cairo_save( w->mCtx );
  cairo_rectangle( w->mCtx, x->Value(), y->Value(), width->Value(), height->Value() );
  cairo_clip( w->mCtx );
}

void unclipDrawingArea(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Local<v8::Integer> window = v8::Local<v8::Integer>::Cast(args[0]);
  Window *w = gWindows[window->Value()];

  cairo_restore( w->mCtx );
}

void cairo_drawWireFrame( Window *w, const WireFrame *wf, const double x, const double y, const double angle, int color ) {
  int i;
  cairo_save( w->mCtx );
  cairo_translate( w->mCtx, x, y );
  cairo_rotate( w->mCtx, deg2rad( angle ));
  cairo_set_line_width( w->mCtx, LINE_WIDTH);
  cairo_set_source_rgb( w->mCtx, ((color >> 16)&0xFF)/256.0, ((color >> 8)&0xFF)/256.0, (color&0xFF)/256.0 );

  for (i=0; i<wf->lineCount; i++) {
    cairo_move_to( w->mCtx, wf->points[wf->lines[i].from].mX, wf->points[wf->lines[i].from].mY );
    cairo_line_to( w->mCtx, wf->points[wf->lines[i].to].mX, wf->points[wf->lines[i].to].mY );
  }

  cairo_stroke( w->mCtx );
  cairo_restore( w->mCtx );
}

void drawWireFrame(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::Integer> window = v8::Local<v8::Integer>::Cast(args[0]);
  v8::Local<v8::String> object = v8::Local<v8::String>::Cast(args[1]);
  v8::Local<v8::Number> x = v8::Local<v8::Number>::Cast(args[2]);
  v8::Local<v8::Number> y = v8::Local<v8::Number>::Cast(args[3]);
  v8::Local<v8::Number> angle = v8::Local<v8::Number>::Cast(args[4]);
  v8::Local<v8::Integer> color = v8::Local<v8::Integer>::Cast(args[5]);

  WireFrame *wf;
  std::string name = std::string(*v8::String::Utf8Value(isolate, object->ToString()));

  if (name == "fortress") {
    wf = &fortressWireFrame;
  } else if (name == "ship") {
    wf = &shipWireFrame;
  } else if (name == "missile") {
    wf = &missileWireFrame;
  } else if (name == "shell") {
    wf = &shellWireFrame;
  } else {
    std::cout << "cant find it " << name << "\n";
    return;
  }

  cairo_drawWireFrame(gWindows[window->Value()], wf, x->Value(), y->Value(), angle->Value(), color->Value());
}

void cairo_drawExplosion( Window *w, const double x, const double y ) {
  int radius, angle, ofs;
  /* cairo_set_line_width( mctx, 1.4 ); */
  // cairo_set_line_width( mctx, mLineWidth );
  ofs = 0;
  for (radius=15; radius<70; radius += 8 ) {
    ofs += 3;
    if( radius < 60 ) {
      cairo_set_source_rgb( w->mCtx, 1,1,0 );
    } else {
      cairo_set_source_rgb( w->mCtx, 1,0,0 );
    }
    for (angle=0; angle<360; angle += 30 ) {
      cairo_arc( w->mCtx, x, y, radius, deg2rad(angle+ofs),deg2rad(angle+ofs+10));
      cairo_stroke( w->mCtx );
    }
  }
  cairo_set_source_rgb( w->mCtx, 1, 1, 0);
  cairo_arc(w->mCtx, x, y, 7, 0, M_PI*2);
  cairo_stroke( w->mCtx );
}

void drawExplosion(const v8::FunctionCallbackInfo<v8::Value> &args) {
  // v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::Integer> window = v8::Local<v8::Integer>::Cast(args[0]);
  v8::Local<v8::Number> x = v8::Local<v8::Number>::Cast(args[1]);
  v8::Local<v8::Number> y = v8::Local<v8::Number>::Cast(args[2]);

  cairo_drawExplosion(gWindows[window->Value()], x->Value(), y->Value());
}

void cairo_hexagonPath( Window *w, const double radius, const double angle, bool full ) {
  static double lastRadius = -1;
  static Vector points[6];

  if (lastRadius != radius) {
    double x1 = floor(-radius);
    double x2 = floor(-radius*0.5);
    double x3 = floor(+radius*0.5);
    double x4 = floor(+radius);
    double y1 = 0;
    double y2 = floor(-radius*sin(M_PI*2/3));
    double y3 = floor(+radius*sin(M_PI*2/3));
    points[0].mX = x1;
    points[0].mY = y1;
    points[1].mX = x2;
    points[1].mY = y2;
    points[2].mX = x3;
    points[2].mY = y2;
    points[3].mX = x4;
    points[3].mY = y1;
    points[4].mX = x3;
    points[4].mY = y3;
    points[5].mX = x2;
    points[5].mY = y3;
  }

  if (full) {
    cairo_move_to( w->mCtx, points[0].mX, points[0].mY );
    for (int i=1; i<6; i++) {
      cairo_line_to( w->mCtx, points[i].mX, points[i].mY );
    }
    cairo_close_path( w->mCtx );
  } else {
    cairo_move_to( w->mCtx, points[1].mX, points[1].mY );
    for (int i=2; i<6; i++) {
      cairo_line_to( w->mCtx, points[i].mX, points[i].mY );
    }
  }
}


void cairo_drawHexagon( Window *w, const double radius, const double x, const double y, const double angle, bool full, const int color ) {

  cairo_save( w->mCtx );
  cairo_set_source_rgb( w->mCtx, ((color >> 16)&0xFF)/256.0, ((color >> 8)&0xFF)/256.0, (color&0xFF)/256.0 );
  cairo_translate(w->mCtx, x, y);
  cairo_rotate(w->mCtx, deg2rad(angle));

  cairo_hexagonPath(w, radius, angle, full);
}

void drawPolygon(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::Integer> window = v8::Local<v8::Integer>::Cast(args[0]);
  v8::Local<v8::Array> points = v8::Local<v8::Array>::Cast(args[1]);
  v8::Local<v8::Integer> color = v8::Local<v8::Integer>::Cast(args[2]);
  v8::Local<v8::Boolean> fill = v8::Local<v8::Boolean>::Cast(args[3]);

  Window *w = gWindows[window->Value()];

  cairo_save( w->mCtx );

  for (uint32_t i=0; i<points->Length(); i++) {
    v8::Local<v8::Object> p = v8::Local<v8::Object>::Cast(points->Get(i));
    v8::Local<v8::Number> x = v8::Local<v8::Number>::Cast(p->Get(v8::String::NewFromUtf8(isolate, "x")));
    v8::Local<v8::Number> y = v8::Local<v8::Number>::Cast(p->Get(v8::String::NewFromUtf8(isolate, "y")));
    if (i == 0)
      cairo_move_to( w->mCtx, x->Value(), y->Value() );
    else
      cairo_line_to( w->mCtx, x->Value(), y->Value() );
  }

  cairo_set_source_rgb( w->mCtx, ((color->Value() >> 16)&0xFF)/256.0, ((color->Value() >> 8)&0xFF)/256.0, (color->Value()&0xFF)/256.0 );
  if (fill->Value()) cairo_fill( w->mCtx );
  else cairo_stroke( w->mCtx );
  cairo_restore( w->mCtx );
}

void fillBubble(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Local<v8::Integer> window = v8::Local<v8::Integer>::Cast(args[0]);
  double x = v8::Local<v8::Number>::Cast(args[1])->Value();
  double y = v8::Local<v8::Number>::Cast(args[2])->Value();
  double width = v8::Local<v8::Number>::Cast(args[3])->Value();
  double height = v8::Local<v8::Number>::Cast(args[4])->Value();
  double corner_radius = v8::Local<v8::Number>::Cast(args[5])->Value();
  v8::Local<v8::Integer> color = v8::Local<v8::Integer>::Cast(args[6]);

  Window *w = gWindows[window->Value()];

  cairo_save( w->mCtx );

  double aspect = 1.0;
  // double corner_radius = height / 10.0;
  double radius = corner_radius / aspect;
  double degrees = M_PI / 180.0;

  cairo_new_sub_path (w->mCtx);
  cairo_arc (w->mCtx, x + width - radius, y + radius, radius, -90 * degrees, 0 * degrees);
  cairo_arc (w->mCtx, x + width - radius, y + height - radius, radius, 0 * degrees, 90 * degrees);
  cairo_arc (w->mCtx, x + radius, y + height - radius, radius, 90 * degrees, 180 * degrees);
  cairo_arc (w->mCtx, x + radius, y + radius, radius, 180 * degrees, 270 * degrees);
  cairo_close_path (w->mCtx);

  cairo_set_source_rgb( w->mCtx, ((color->Value() >> 16)&0xFF)/256.0, ((color->Value() >> 8)&0xFF)/256.0, (color->Value()&0xFF)/256.0 );
  cairo_fill (w->mCtx);
  // cairo_set_line_width (w->mCtx, 10.0);
  // cairo_stroke( w->mCtx );

  cairo_restore( w->mCtx );
}

void drawHexagon(const v8::FunctionCallbackInfo<v8::Value> &args) {
  // v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::Integer> window = v8::Local<v8::Integer>::Cast(args[0]);
 v8::Local<v8::Number> radius = v8::Local<v8::Number>::Cast(args[1]);
  v8::Local<v8::Number> x = v8::Local<v8::Number>::Cast(args[2]);
  v8::Local<v8::Number> y = v8::Local<v8::Number>::Cast(args[3]);
  v8::Local<v8::Number> angle = v8::Local<v8::Number>::Cast(args[4]);
  v8::Local<v8::Boolean> full = v8::Local<v8::Boolean>::Cast(args[5]);
  v8::Local<v8::Integer> color = v8::Local<v8::Integer>::Cast(args[6]);

  cairo_drawHexagon(gWindows[window->Value()], radius->Value(), x->Value(), y->Value(), angle->Value(), full->Value(), color->Value());
  cairo_stroke( gWindows[window->Value()]->mCtx );
  cairo_restore( gWindows[window->Value()]->mCtx );
}

void fillHexagon(const v8::FunctionCallbackInfo<v8::Value> &args) {
  // v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::Integer> window = v8::Local<v8::Integer>::Cast(args[0]);
 v8::Local<v8::Number> radius = v8::Local<v8::Number>::Cast(args[1]);
  v8::Local<v8::Number> x = v8::Local<v8::Number>::Cast(args[2]);
  v8::Local<v8::Number> y = v8::Local<v8::Number>::Cast(args[3]);
  v8::Local<v8::Number> angle = v8::Local<v8::Number>::Cast(args[4]);
  v8::Local<v8::Integer> color = v8::Local<v8::Integer>::Cast(args[6]);

  cairo_drawHexagon(gWindows[window->Value()], radius->Value(), x->Value(), y->Value(), angle->Value(), true, color->Value());
  cairo_fill( gWindows[window->Value()]->mCtx );
  cairo_restore( gWindows[window->Value()]->mCtx );
}

void cairo_drawScore( Window *w, const std::string label, const int value, const double y ) {
  int label_width = 89;
  int label_height = 32;
  int start = (w->getWidth() - label_width)/2;

  cairo_save( w->mCtx );

  cairo_set_font_face( w->mCtx, gCairoFont );
  cairo_set_font_size( w->mCtx, FONT_SIZE );

  cairo_set_source_rgb(w->mCtx, 0, 1, 0);
  cairo_rectangle(w->mCtx, start, y, label_width, label_height * 2);
  cairo_move_to(w->mCtx, start, y+label_height);
  cairo_line_to(w->mCtx, start+label_width, y+label_height);
  cairo_stroke(w->mCtx);

  cairo_set_source_rgb(w->mCtx, 0, 1, 0);

  drawText(w->mCtx, "Points", start + label_width/2, y + label_height/2, 0, 0);
  cairo_stroke( w->mCtx );
  std::ostringstream s;
  s << value;
  cairo_set_source_rgb(w->mCtx, 1, 1, 0);
  drawText(w->mCtx, s.str(), start + label_width/2, y + label_height + label_height/2, 0, 0);
  cairo_stroke( w->mCtx );

  cairo_restore( w->mCtx );
}

void old_cairo_drawFlatScore( Window *w, const std::string label, const int value ) {
  int label_width = 89;
  int label_height = 32;
  int start = (w->getWidth() - label_width*2)/2;
  int y = w->getHeight()/2 + 315;

  cairo_save( w->mCtx );

  cairo_set_font_face( w->mCtx, gCairoFont );
  cairo_set_font_size( w->mCtx, FONT_SIZE );

  cairo_set_line_width(w->mCtx, 4);
  cairo_set_source_rgb(w->mCtx, 0, 0.4, 0);
  cairo_rectangle(w->mCtx, start, y, label_width*2, label_height);
  cairo_move_to(w->mCtx, start+label_width, y);
  cairo_line_to(w->mCtx, start+label_width, y+label_height);
  cairo_stroke(w->mCtx);

  cairo_set_line_width(w->mCtx, 4);
  cairo_set_source_rgb(w->mCtx, 0, 1, 0);

  drawText(w->mCtx, "Points", start + label_width/2, y + label_height/2, 0, 0);
  cairo_stroke( w->mCtx );
  std::ostringstream s;
  s << value;
  cairo_set_source_rgb(w->mCtx, 1, 1, 0);
  drawText(w->mCtx, s.str(), start + label_width + label_width/2, y + label_height/2, 0, 0);
  cairo_stroke( w->mCtx );

  cairo_restore( w->mCtx );
}

void cairo_drawFlatScore( Window *w, const std::string label, const int value ) {
  int label_width = 89;
  int label_height = 32;
  int start = (w->getWidth() - label_width)/2;
  int y = w->getHeight()/2 + 310;

  cairo_save( w->mCtx );

  cairo_set_font_face( w->mCtx, gCairoFont );
  cairo_set_font_size( w->mCtx, FONT_SIZE );

  cairo_set_line_width(w->mCtx, 4);
  cairo_set_source_rgb(w->mCtx, 0, 0.4, 0);
  cairo_rectangle(w->mCtx, start, y, label_width, label_height);
  cairo_move_to(w->mCtx, start+label_width, y);
  cairo_line_to(w->mCtx, start+label_width, y+label_height);
  cairo_stroke(w->mCtx);

  // cairo_set_line_width(w->mCtx, 1);
  // cairo_set_source_rgb(w->mCtx, 0, 1, 0);

  // drawText(w->mCtx, "Points", start + label_width/2, y + label_height/2, 0, 0);
  // cairo_stroke( w->mCtx );
  std::ostringstream s;
  s << value;
  cairo_set_source_rgb(w->mCtx, 1, 1, 0);
  drawText(w->mCtx, s.str(), start + label_width/2, y + label_height/2, 0, 0);
  cairo_stroke( w->mCtx );

  cairo_restore( w->mCtx );
}


void drawScore(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Local<v8::Integer> window = v8::Local<v8::Integer>::Cast(args[0]);
  v8::Local<v8::Integer> score = v8::Local<v8::Integer>::Cast(args[1]);
  v8::Local<v8::Number> y = v8::Local<v8::Number>::Cast(args[2]);

  cairo_drawScore(gWindows[window->Value()], "Score", score->Value(), y->Value());
}

void drawFlatScore(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Local<v8::Integer> window = v8::Local<v8::Integer>::Cast(args[0]);
  v8::Local<v8::Integer> score = v8::Local<v8::Integer>::Cast(args[1]);

  cairo_drawFlatScore(gWindows[window->Value()], "Score", score->Value());
}

void cairo_drawGameArea(Window *w) {
  cairo_save(w->mCtx);

  cairo_set_source_rgb(w->mCtx, 0, .4, 0);
  cairo_set_line_width(w->mCtx, 4);

  cairo_rectangle(w->mCtx, -355, -310, 710, 620);
  cairo_stroke(w->mCtx);

  // cairo_set_line_width(w->mCtx, 1);
  // double x = -350+2, y = -315+2;
  // crispXY(w->mCtx, x, y);
  // cairo_rectangle(w->mCtx, x, y, 710-4, 626-4);
  // cairo_set_source_rgb(w->mCtx, 0, 1, 0);
  // cairo_stroke(w->mCtx);

  cairo_restore(w->mCtx);
}

void drawGameArea(const v8::FunctionCallbackInfo<v8::Value> &args) {
  // v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::Integer> window = v8::Local<v8::Integer>::Cast(args[0]);

  cairo_drawGameArea(gWindows[window->Value()]);
}


void cairo_drawText(Window *w, std::string txt, double x, double y, double size, int color ) {
  cairo_set_font_face( w->mCtx, gCairoFont );
  cairo_set_font_size( w->mCtx, size );
  cairo_set_source_rgb( w->mCtx, ((color >> 16)&0xFF)/256.0, ((color >> 8)&0xFF)/256.0, (color&0xFF)/256.0 );

  cairo_move_to( w->mCtx, x, y );
  cairo_show_text(w->mCtx, txt.c_str());
}

void drawText(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::Integer> window = v8::Local<v8::Integer>::Cast(args[0]);
  v8::Local<v8::String> object = v8::Local<v8::String>::Cast(args[1]);
  v8::Local<v8::Number> x = v8::Local<v8::Number>::Cast(args[2]);
  v8::Local<v8::Number> y = v8::Local<v8::Number>::Cast(args[3]);
  v8::Local<v8::Number> size = v8::Local<v8::Integer>::Cast(args[4]);
  v8::Local<v8::Integer> color = v8::Local<v8::Integer>::Cast(args[5]);

  std::string s = std::string(*v8::String::Utf8Value(isolate, object->ToString()));
  if (s == "") s = " ";

  cairo_drawText(gWindows[window->Value()],
                 s,
                 x->Value(),
                 y->Value(),
                 size->Value(),
                 color->Value());
}

void fontSize(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::Integer> window = v8::Local<v8::Integer>::Cast(args[0]);
  v8::Local<v8::Number> size = v8::Local<v8::Number>::Cast(args[1]);

  double ascent, descent, height;
  fontSize(gWindows[window->Value()], size->Value(), ascent, descent, height);

  v8::Local<v8::Object> obj = v8::Object::New(isolate);
  obj->Set(v8::String::NewFromUtf8(isolate, "ascent"), v8::Number::New(isolate, ascent));
  obj->Set(v8::String::NewFromUtf8(isolate, "descent"), v8::Number::New(isolate, descent));
  obj->Set(v8::String::NewFromUtf8(isolate, "height"), v8::Number::New(isolate, height));

  args.GetReturnValue().Set(obj);
}

void textSize(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::Integer> window = v8::Local<v8::Integer>::Cast(args[0]);
  v8::Local<v8::String> object = v8::Local<v8::String>::Cast(args[1]);
  v8::Local<v8::Number> size = v8::Local<v8::Number>::Cast(args[2]);

  std::string s = std::string(*v8::String::Utf8Value(isolate, object->ToString()));
  // if (s == "") s = " ";
  double w, h;
  textSize(gWindows[window->Value()], s, size->Value(), w, h);

  v8::Local<v8::Object> obj = v8::Object::New(isolate);
  obj->Set(v8::String::NewFromUtf8(isolate, "w"), v8::Number::New(isolate, w));
  obj->Set(v8::String::NewFromUtf8(isolate, "h"), v8::Number::New(isolate, h));

  args.GetReturnValue().Set(obj);
}

void screenSize(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::Integer> window = v8::Local<v8::Integer>::Cast(args[0]);

  v8::Local<v8::Object> obj = v8::Object::New(isolate);
  obj->Set(v8::String::NewFromUtf8(isolate, "w"), v8::Number::New(isolate, gWindows[window->Value()]->getWidth()));
  obj->Set(v8::String::NewFromUtf8(isolate, "h"), v8::Number::New(isolate, gWindows[window->Value()]->getHeight()));

  args.GetReturnValue().Set(obj);
};

cairo_surface_t *getImage(const std::string &file) {
  std::map<std::string, cairo_surface_t *>::iterator it;
  it = gImages.find(file);
  if (it == gImages.end()) {
    std::cout << "loading " << file << "\n";
    gImages[file] = cairo_image_surface_create_from_png( file.c_str() );
  }
  return gImages[file];
};

void imageSize(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  std::string file = std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(args[0])));

  cairo_surface_t *surf = getImage( file );

  v8::Local<v8::Object> obj = v8::Object::New(isolate);
  obj->Set(v8::String::NewFromUtf8(isolate, "w"), v8::Integer::New(isolate, cairo_image_surface_get_width( surf )));
  obj->Set(v8::String::NewFromUtf8(isolate, "h"), v8::Integer::New(isolate, cairo_image_surface_get_height( surf )));

  args.GetReturnValue().Set(obj);
}

void drawImage(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::Integer> window = v8::Local<v8::Integer>::Cast(args[0]);
  std::string file = std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(args[1])));
  v8::Local<v8::Number> x = v8::Local<v8::Number>::Cast(args[2]);
  v8::Local<v8::Number> y = v8::Local<v8::Number>::Cast(args[3]);
  std::string xOrient = std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(args[4])));
  std::string yOrient = std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(args[5])));

  cairo_surface_t *img = getImage( file );
  // if ( cairo_surface_status( img ) == CAIRO_STATUS_SUCCESS ) {
  int w = cairo_image_surface_get_width( img );
  int h = cairo_image_surface_get_height( img );
  double cx = x->Value(), cy = y->Value();

  if (xOrient == "center") cx -= w/2.0;
  else if (xOrient == "right") cx -= w;
  if (yOrient == "center") cy -= h/2.0;
  else if (xOrient == "bottom") cy -= h;

  crispXY(gWindows[window->Value()]->mCtx, cx, cy);

  // std::cout << "image " << file << " " << cx << " " << cy << " " << w << " " << h << "\n";

  // cairo_save( gWindows[window->Value()]->mCtx );

  cairo_set_source_surface( gWindows[window->Value()]->mCtx, img, cx, cy );
  // cairo_translate( gWindows[window->Value()]->mCtx, cx, cy );
  cairo_paint( gWindows[window->Value()]->mCtx );

  // cairo_restore( gWindows[window->Value()]->mCtx );
  // }
}

void debug(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::String> object = v8::Local<v8::String>::Cast(args[0]);
  std::string s = std::string(*v8::String::Utf8Value(isolate, object->ToString()));

  std::cout << "raz: " << s;
}

void mainMenu(const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::String> warning = v8::Local<v8::String>::Cast(args[0]);
  v8::Local<v8::Array> js_sessions = v8::Local<v8::Array>::Cast(args[1]);
  // v8::Local<v8::Array> js_conditions = v8::Local<v8::Array>::Cast(args[2]);

  std::string s = std::string(*v8::String::Utf8Value(isolate, warning->ToString()));
  std::vector<std::string> sessions, conditions;

  for (uint32_t i=0; i<js_sessions->Length(); i++) {
    sessions.push_back(std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(js_sessions->Get(i)))));
  }

  Menu menu(gWindows[0], sessions, conditions, 0, s);
  menu.run();

  v8::Local<v8::Object> obj = v8::Object::New(isolate);
  obj->Set(v8::String::NewFromUtf8(isolate, "mode"), v8::String::NewFromUtf8(isolate, menu.modeName().c_str()));
  obj->Set(v8::String::NewFromUtf8(isolate, "id"), v8::String::NewFromUtf8(isolate, menu.mSubjectId.c_str()));
  obj->Set(v8::String::NewFromUtf8(isolate, "debug"), v8::Boolean::New(isolate, menu.mDebug));
  obj->Set(v8::String::NewFromUtf8(isolate, "fullscreen"), v8::Boolean::New(isolate, menu.mScreenMode));
  obj->Set(v8::String::NewFromUtf8(isolate, "session"), v8::Integer::New(isolate, menu.mSession));

  args.GetReturnValue().Set(obj);
};

// Networking
Server *gServer = NULL;
BroadcastClient *gBroadcastClient = NULL;
Client *gClient = NULL;

void startServer( const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Local<v8::Integer> port = v8::Local<v8::Integer>::Cast(args[0]);
  gServer = new Server( port->Value() );
  gServer->start();
}

void stopServer( const v8::FunctionCallbackInfo<v8::Value> &args) {
  gServer->shutdown();
  delete gServer;
  gServer = NULL;
}

void checkForNewClients( const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();

  bool ret = gServer->checkForNewClients();

  args.GetReturnValue().Set(v8::Boolean::New(isolate, ret));
}

void getNumClients( const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  args.GetReturnValue().Set(v8::Integer::New(isolate, gServer->mClients.size()));
};

void broadcastAvailability( const v8::FunctionCallbackInfo<v8::Value> &args) {
  gServer->broadcastAvailability();
}

void serverSend( const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::String> msg = v8::Local<v8::String>::Cast(args[0]);
  std::string s = std::string(*v8::String::Utf8Value(isolate, msg->ToString()));
  if (args.Length() > 1) {
    v8::Local<v8::Integer> client = v8::Local<v8::Integer>::Cast(args[1]);
    if( client->Value() >= 0 && client->Value() < gServer->mClients.size() )
      gServer->sendMessage( s, client->Value() );
  } else {
    gServer->sendMessage( s );
  }
}

void serverStartLoggingAudio( const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  std::string path[2] = {std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(args[0]))),
                         std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(args[1])))};

  if( gAudioFile[0].isOpen() ) gAudioFile[0].close();
  if( gAudioFile[1].isOpen() ) gAudioFile[1].close();

  // FIXME: hardcoded values that are configurable on the client.
  gAudioFile[0].open( path[0].c_str(), 48000, 1 );
  gAudioFile[1].open( path[1].c_str(), 48000, 1 );
}

void serverStopLoggingAudio( const v8::FunctionCallbackInfo<v8::Value> &args) {
  if( gAudioFile[0].isOpen() ) gAudioFile[0].close();
  if( gAudioFile[1].isOpen() ) gAudioFile[1].close();
}

void serverRead( const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::Integer> client = v8::Local<v8::Integer>::Cast(args[0]);

  std::vector<Uint8> types;
  std::vector<std::string> messages;

  gServer->getClientMessages(client->Value(), types, messages);

  if( messages.size() > 0 ) {
    v8::Local<v8::Array> ret = v8::Array::New(isolate, messages.size());
    int idx = 0;
    for (size_t i=0; i<messages.size(); i++) {
      if ( types[i] == MESSAGE_TYPE_STRING ) {
      ret->Set( idx, v8::String::NewFromUtf8( isolate, messages[i].c_str() ));
      idx += 1;
      } else if( types[i] == MESSAGE_TYPE_AUDIO ) {
        // std::cout << "audio " << client->Value() << " " << messages[i].size() << std::endl;
        // Hardcoded for 2 players
        int other = 0;
        if (client->Value() == 0) other = 1;
        gServer->sendAudioData( messages[i], other );
        if( gAudioFile[client->Value()].isOpen() ) {
          // Help synchronize the audio streams.
          if( gAudioFile[client->Value()].mBytesWritten == 0 ) {
            ret->Set( idx, v8::String::NewFromUtf8( isolate, "{\"type\":\"first-audio-packet\"}" ));
            idx += 1;
          }
          gAudioFile[client->Value()].write(messages[i].data(), messages[i].size());
        }
      }
    }

    // std::cout << "<<< Done" << std::endl;

    ret->Set(v8::String::NewFromUtf8(isolate, "length"), v8::Integer::New(isolate, idx));
    args.GetReturnValue().Set(ret);
  }
}

void serverIsConnected( const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();

  args.GetReturnValue().Set(v8::Boolean::New(isolate, gServer->mConnected));
}

void findServer( const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();

  if( gBroadcastClient == NULL ) gBroadcastClient = new BroadcastClient();

  if( gBroadcastClient->checkForServerBroadcasts() ) {
    std::ostringstream s;
    s << ((gBroadcastClient->mLastServer.host >>  0) & 0xFF) << "."
      << ((gBroadcastClient->mLastServer.host >>  8) & 0xFF) << "."
      << ((gBroadcastClient->mLastServer.host >> 16) & 0xFF) << "."
      << ((gBroadcastClient->mLastServer.host >> 24) & 0xFF);
    v8::Local<v8::Object> obj = v8::Object::New(isolate);
    obj->Set(v8::String::NewFromUtf8(isolate, "host"), v8::String::NewFromUtf8(isolate, s.str().c_str()));
    obj->Set(v8::String::NewFromUtf8(isolate, "port"), v8::Integer::New(isolate, gBroadcastClient->mLastServer.port));
    args.GetReturnValue().Set(obj);
  }
}

void startClient( const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::String> host = v8::Local<v8::String>::Cast(args[0]);
  v8::Local<v8::Integer> port = v8::Local<v8::Integer>::Cast(args[1]);
  std::string s = std::string(*v8::String::Utf8Value(isolate, host->ToString()));

  gClient = new Client(s, port->Value());
  bool ret = gClient->connect();
  args.GetReturnValue().Set(v8::Boolean::New(isolate, ret));
}

void stopClient( const v8::FunctionCallbackInfo<v8::Value> &args) {
  delete gClient;
  gClient = NULL;
}

void clientSend( const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::String> host = v8::Local<v8::String>::Cast(args[0]);
  std::string s = std::string(*v8::String::Utf8Value(isolate, host->ToString()));

  gClient->sendMessage(s);
}

void clientSendBuffer( const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();

  if (args.Length() == 1 && args[0]->IsArrayBufferView()) {
    v8::Local<v8::ArrayBufferView> buf = v8::Local<v8::ArrayBufferView>::Cast(args[0]);
    v8::Local<v8::ArrayBuffer> b = buf->Buffer();
    v8::ArrayBuffer::Contents c = b->GetContents();
    gClient->sendData(MESSAGE_TYPE_BUFFER, (Uint8*)c.Data() + buf->ByteOffset(), buf->ByteLength());
  }
}


void clientSendAudioData( const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();

  if (args.Length() == 0 && args[0]->IsArrayBufferView()) {
    v8::Local<v8::ArrayBufferView> buf = v8::Local<v8::ArrayBufferView>::Cast(args[0]);
    v8::Local<v8::ArrayBuffer> b = buf->Buffer();
    v8::ArrayBuffer::Contents c = b->GetContents();
    gClient->sendAudioData((Uint8*)c.Data() + buf->ByteOffset(), buf->ByteLength());
  }
}

void clientSendAudio( const v8::FunctionCallbackInfo<v8::Value> &args ) {
  const int bufferSize = 2880; //4096 * 32;
  static float data[bufferSize];
  const int outputSize = 4000; // 1500
  static unsigned char outputBuf[outputSize];

  // memset(data, 0, bufferSize);

  if (gAudioRecordingDevice == 0) return;

  Uint32 len = SDL_DequeueAudio( gAudioRecordingDevice, (void *)data, bufferSize*sizeof(float));

  if (len > 0) {
    // std::cout << SDL_GetTicks() << ": Send Audio " << len << std::endl;
    int sz = opus_encode_float(gEncoder, (float *)data, len/sizeof(float), outputBuf, outputSize);
    if (sz < 0) {
      std::cout << "Encoding error: " << sz << std::endl;
    } else if (sz <= 2) {
      // no need to transmit
    } else {
      gClient->sendAudioData(outputBuf, sz);
    }
  }
}

void getAudio( const v8::FunctionCallbackInfo<v8::Value> &args ) {
  if (gAudioRecordingDevice == 0) return;
  const int bufferSize = 4096 * 2;
  // static Uint8 data[bufferSize];
  // memset(data, 0, bufferSize);

  v8::Local<v8::ArrayBuffer> buf = v8::ArrayBuffer::New( args.GetIsolate(), bufferSize);
  v8::ArrayBuffer::Contents c = buf->GetContents();
  Uint32 len = SDL_DequeueAudio( gAudioRecordingDevice, c.Data(), c.ByteLength());
  if (len > 0) {
    v8::Local<v8::Uint8Array> ret = v8::Uint8Array::New( buf, 0, len);
    args.GetReturnValue().Set( ret );
  }
}

void queueAudio( const v8::FunctionCallbackInfo<v8::Value> &args ) {
  // v8::Isolate* isolate = args.GetIsolate();

  if (args.Length() == 0) return;
  if (gAudioPlaybackDevice == 0) return;

  if (args[0]->IsArrayBufferView()) {
    v8::Local<v8::ArrayBufferView> buf = v8::Local<v8::ArrayBufferView>::Cast(args[0]);
    v8::Local<v8::ArrayBuffer> b = buf->Buffer();
    v8::ArrayBuffer::Contents c = b->GetContents();

    // std::cout << "Queuing " << buf->ByteLength() << " bytes" << " ofs " << buf->ByteOffset() << std::endl;
    SDL_QueueAudio( gAudioPlaybackDevice, (Uint8*)c.Data() + buf->ByteOffset(), buf->ByteLength());

    // std::cout << audioStatus(gAudioPlaybackDevice) << " - " << SDL_GetQueuedAudioSize( gAudioPlaybackDevice ) << " " << gPlaybackAudioSpec.size << std::endl;

  } else if (args[0]->IsArrayBuffer()) {
    v8::Local<v8::ArrayBuffer> buf = v8::Local<v8::ArrayBuffer>::Cast(args[0]);
    v8::ArrayBuffer::Contents c = buf->GetContents();

    // std::cout << "Queuing " << c.ByteLength() << " bytes" << std::endl;
    SDL_QueueAudio( gAudioPlaybackDevice, c.Data(), c.ByteLength());

    std::cout << audioStatus(gAudioPlaybackDevice) << " - " << SDL_GetQueuedAudioSize( gAudioPlaybackDevice ) << " " << gPlaybackAudioSpec.size << std::endl;

  }
}

void mixAndQueueAudio( const v8::FunctionCallbackInfo<v8::Value> &args ) {
  // v8::Isolate* isolate = args.GetIsolate();

  if (args.Length() != 2) return;
  if (gAudioPlaybackDevice == 0) return;

  if (args[0]->IsArrayBufferView() && args[1]->IsArrayBufferView()) {
    v8::Local<v8::ArrayBufferView> buf1 = v8::Local<v8::ArrayBufferView>::Cast(args[0]);
    v8::Local<v8::ArrayBufferView> buf2 = v8::Local<v8::ArrayBufferView>::Cast(args[1]);
    v8::Local<v8::ArrayBuffer> b1 = buf1->Buffer();
    v8::Local<v8::ArrayBuffer> b2 = buf2->Buffer();
    v8::ArrayBuffer::Contents c1 = b1->GetContents();
    v8::ArrayBuffer::Contents c2 = b2->GetContents();

    if (buf1->ByteLength() == buf2->ByteLength()) {
      SDL_MixAudioFormat((Uint8*)c1.Data()+buf1->ByteOffset(), (Uint8*)c2.Data() + buf2->ByteOffset(), gPlaybackAudioSpec.format, buf1->ByteLength(), SDL_MIX_MAXVOLUME);
      SDL_QueueAudio( gAudioPlaybackDevice, (Uint8*)c1.Data() + buf1->ByteOffset(), buf1->ByteLength());
    } else {
      std::cout << "buffers to be mixed are different lengths! " << buf1->ByteLength() << " != " << buf2->ByteLength() << std::endl;
    }

    // std::cout << audioStatus(gAudioPlaybackDevice) << " - " << SDL_GetQueuedAudioSize( gAudioPlaybackDevice ) << " " << gPlaybackAudioSpec.size << std::endl;

  } else {//  if (args[0]->IsArrayBuffer()) {
    // v8::Local<v8::ArrayBuffer> buf = v8::Local<v8::ArrayBuffer>::Cast(args[0]);
    // v8::ArrayBuffer::Contents c = buf->GetContents();

    // // std::cout << "Queuing " << c.ByteLength() << " bytes" << std::endl;
    // SDL_QueueAudio( gAudioPlaybackDevice, c.Data(), c.ByteLength());

    // std::cout << audioStatus(gAudioPlaybackDevice) << " - " << SDL_GetQueuedAudioSize( gAudioPlaybackDevice ) << " " << gPlaybackAudioSpec.size << std::endl;
    std::cout << "Uh oh the buffers to be mixed weren't array views!" << std::endl;

  }
  //  else {
  //   v8::Local<v8::String> tmp = args[0]->TypeOf(isolate);
  //   std::string type = *v8::String::Utf8Value(isolate, tmp);
  //   std::cout << "first argument is " << type << " " << args[0]->IsUint8Array() << std::endl;
  // }
}

void clearAudioQueues( const v8::FunctionCallbackInfo<v8::Value> &args ) {
  SDL_ClearQueuedAudio( gAudioPlaybackDevice );
  SDL_ClearQueuedAudio( gAudioRecordingDevice );
}

void audioPlaybackQueueSize( const v8::FunctionCallbackInfo<v8::Value> &args ) {
  if (gAudioPlaybackDevice == 0) return;
  v8::Isolate* isolate = args.GetIsolate();
  v8::Local<v8::Integer> ret = v8::Integer::New(isolate, SDL_GetQueuedAudioSize( gAudioPlaybackDevice ));
  args.GetReturnValue().Set(ret);
}


v8::Local<v8::Value> decodeAudioDataIntoBuffer(v8::Isolate *isolate, OpusDecoder *decoder, const std::string &msg, int player) {
  const int frameSize = 2880 * 10;
  static float decodeBuf[frameSize];

  int samples = opus_decoder_get_nb_samples( decoder,
                                             (unsigned char *)msg.data(),
                                             msg.size() );

  // std::cout << "got packet with " << samples << " samples." << std::endl;

  int ret = opus_decode_float( decoder,
                               (unsigned char *)msg.data(),
                               msg.size(),
                               decodeBuf,
                               frameSize,
                               0 );

  if (ret > 0) {
    v8::MaybeLocal<v8::Object> maybe = node::Buffer::New( isolate, ret * sizeof(float) );
    v8::Local<v8::Object> buf;
    if (maybe.ToLocal(&buf)) {
      memcpy(node::Buffer::Data(buf), decodeBuf, ret * sizeof(float));
      v8::Local<v8::Object> obj = v8::Object::New(isolate);
      obj->Set(v8::String::NewFromUtf8(isolate, "screen"), v8::String::NewFromUtf8(isolate, "spyAudio"));
      obj->Set(v8::String::NewFromUtf8(isolate, "player"),  v8::Integer::New(isolate, player));
      obj->Set(v8::String::NewFromUtf8(isolate, "audio"),  buf);
      return obj;
    }
  }
  // something failed.
  return v8::Null(isolate);
}

void clientRead( const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();

  std::vector<Uint8> types;
  std::vector<std::string> messages;

  gClient->getServerMessages( types, messages );

  if( messages.size() > 0 ) {
    v8::Local<v8::Array> ret = v8::Array::New(isolate, messages.size());
    int idx = 0;
    for (size_t i=0; i<messages.size(); i++) {
      if (types[i] == MESSAGE_TYPE_STRING) {
        ret->Set( idx, v8::String::NewFromUtf8( isolate, messages[i].c_str() ));
        idx += 1;
      } else if (types[i] == MESSAGE_TYPE_BUFFER) {
        // v8::Local<v8::ArrayBuffer> buf = v8::ArrayBuffer::New( args.GetIsolate(), messages[i].size());
        // v8::ArrayBuffer::Contents c = buf->GetContents();
        v8::MaybeLocal<v8::Object> maybe = node::Buffer::New( isolate, messages[i].size() );
        v8::Local<v8::Object> buf;
        if (maybe.ToLocal(&buf)) {
          memcpy(node::Buffer::Data(buf), messages[i].c_str(), messages[i].size());
          ret->Set( idx, buf );
          idx += 1;
        }
        // std::cout << "client got a buffer " << messages[i].size() << std::endl;

      } else if (types[i] == MESSAGE_TYPE_AUDIO) {
        // std::cout << "Boing! audio " << messages[i].size() << std::endl;

        const int frameSize = 2880 * 10;
        static float decodeBuf[frameSize];

        int samples = opus_decoder_get_nb_samples( gDecoder,
                                                   (unsigned char *)messages[i].data(),
                                                   messages[i].size() );

        // std::cout << "got packet with " << samples << " samples." << std::endl;

        int ret = opus_decode_float( gDecoder,
                                     (unsigned char *)messages[i].data(),
                                     messages[i].size(),
                                     decodeBuf,
                                     frameSize,
                                     0 );

        // std::cout << "decoded " << ret << " samples." << std::endl;

        if (ret > 0) playAudioData(decodeBuf, ret);
        else std::cout << "Decoding error: " << ret << std::endl;
      } else if (types[i] == MESSAGE_TYPE_SPY_AUDIO_0) {
        v8::Local<v8::Value> obj = decodeAudioDataIntoBuffer(isolate, gDecoder, messages[i], 0);
        if (!obj->IsNullOrUndefined()) {
          ret->Set( idx, obj );
          idx += 1;
        }
      } else if (types[i] == MESSAGE_TYPE_SPY_AUDIO_1) {
        v8::Local<v8::Value> obj = decodeAudioDataIntoBuffer(isolate, gDecoderPlayer2, messages[i], 1);
        if (!obj->IsNullOrUndefined()) {
          ret->Set( idx, obj );
          idx += 1;
        }
      }
    }
    // std::cout << "message length before: " << ret->Length() << std::endl;
    ret->Set(v8::String::NewFromUtf8(isolate, "length"), v8::Integer::New(isolate, idx));
    // std::cout << "message length after: " << ret->Length() << std::endl;
    args.GetReturnValue().Set(ret);
  }
}

void clientIsConnected( const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();

  args.GetReturnValue().Set(v8::Boolean::New(isolate, gClient && gClient->mConnected));
}

void prepDataForQuestion( const v8::FunctionCallbackInfo<v8::Value> &args, std::string &name, std::string &question, std::vector<std::string> &options, std::vector<std::string> &strings ) {
  v8::Isolate* isolate = args.GetIsolate();
  name = std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(args[0])));
  question = std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(args[1])));
  v8::Local<v8::Array> js_options = v8::Local<v8::Array>::Cast(args[2]);

  for (uint32_t i=0; i<js_options->Length(); i++) {
    if( js_options->Get(i)->IsArray() ) {
      v8::Local<v8::Array> pair = v8::Local<v8::Array>::Cast(js_options->Get(i));

      if( pair->Length() >= 2 ) {
        options.push_back(std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(pair->Get(0)))));
        strings.push_back(std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(pair->Get(1)))));
      }
    } else {
      std::cout << "Element " << i << " is not an array\n";
    }
  }
}

void askRadioQuestion( const v8::FunctionCallbackInfo<v8::Value> &args ) {
  std::string name, question;
  std::vector<std::string> options;
  std::vector<std::string> strings;

  prepDataForQuestion( args, name, question, options, strings );
  clearActiveQuestion( true );

  RadioQuestion *q = new RadioQuestion ( gWindows[0], name, question, options, strings );
  q->setCanGoToPrevious( v8::Local<v8::Boolean>::Cast(args[3])->Value() );
  if( !args[4]->IsUndefined() ) {
    v8::Local<v8::Integer> choice = v8::Local<v8::Integer>::Cast(args[4]);
    q->select( choice->Value() );
  }
  gActiveQuestion = q;
  gActiveQuestion->start();
}

void askCheckboxQuestion( const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  std::string name, question;
  std::vector<std::string> options;
  std::vector<std::string> strings;

  prepDataForQuestion( args, name, question, options, strings );
  clearActiveQuestion( true );
  CheckboxQuestion *q= new CheckboxQuestion ( gWindows[0], name, question, options, strings );
  q->setCanGoToPrevious( v8::Local<v8::Boolean>::Cast(args[3])->Value() );

  if( args[4]->IsArray() ) {
    v8::Local<v8::Array> checked = v8::Local<v8::Array>::Cast(args[4]);
    for (uint32_t i=0; i<checked->Length(); i++) {
      std::string option = std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(checked->Get(i))));
      std::vector<std::string>::iterator m = std::find( options.begin(), options.end(), option );
      if( m != options.end() ) {
        q->check( std::distance( options.begin(), m ), true);
      }
    }
  }
  gActiveQuestion = q;
  gActiveQuestion->start();
}

void askOneLineQuestion( const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  std::string name = std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(args[0])));
  std::string question = std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(args[1])));

  clearActiveQuestion( true );
  OneLineQuestion *q = new OneLineQuestion( gWindows[0], name, question );
  q->setCanGoToPrevious( v8::Local<v8::Boolean>::Cast(args[2])->Value() );
  if( !args[3]->IsUndefined() ) {
    std::string answer = std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(args[3])));
    q->setAnswer( answer );
  }
  gActiveQuestion = q;
  gActiveQuestion->start();
}

void askMultiLineQuestion( const v8::FunctionCallbackInfo<v8::Value> &args) {
  v8::Isolate* isolate = args.GetIsolate();
  std::string name = std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(args[0])));
  std::string question = std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(args[1])));

  clearActiveQuestion( true );
  MultiLineQuestion *q = new MultiLineQuestion ( gWindows[0], name, question );
  q->setCanGoToPrevious( v8::Local<v8::Boolean>::Cast(args[2])->Value() );
  if( !args[3]->IsUndefined() ) {
    std::string answer = std::string(*v8::String::Utf8Value(isolate, v8::Local<v8::String>::Cast(args[3])));
    q->setAnswer( answer );
  }
  gActiveQuestion = q;
  gActiveQuestion->start();
}

void initModule(v8::Local<v8::Object> exports) {
  std::cout << "init graphics.cc\n";

  NODE_SET_METHOD(exports, "initGraphics", initGraphics);
  NODE_SET_METHOD(exports, "shutdownGraphics", shutdownGraphics);
  NODE_SET_METHOD(exports, "createWindows", createWindows);
  NODE_SET_METHOD(exports, "setFullscreen", setFullscreen);
  NODE_SET_METHOD(exports, "numWindows", numWindows);
  NODE_SET_METHOD(exports, "startDrawing", startDrawing);
  NODE_SET_METHOD(exports, "clearScreen", clearScreen);
  NODE_SET_METHOD(exports, "setupGameArea", setupGameArea);
  NODE_SET_METHOD(exports, "takeDownGameArea", takeDownGameArea);
  NODE_SET_METHOD(exports, "finishDrawing", finishDrawing);
  NODE_SET_METHOD(exports, "flip", flip);
  NODE_SET_METHOD(exports, "clipDrawingArea", clipDrawingArea);
  NODE_SET_METHOD(exports, "unclipDrawingArea", unclipDrawingArea);
  NODE_SET_METHOD(exports, "drawWireFrame", drawWireFrame);
  NODE_SET_METHOD(exports, "drawExplosion", drawExplosion);
  NODE_SET_METHOD(exports, "drawHexagon", drawHexagon);
  NODE_SET_METHOD(exports, "fillHexagon", fillHexagon);
  NODE_SET_METHOD(exports, "drawGameArea", drawGameArea);
  NODE_SET_METHOD(exports, "drawScore", drawScore);
  NODE_SET_METHOD(exports, "drawFlatScore", drawFlatScore);
  NODE_SET_METHOD(exports, "drawText", drawText);
  NODE_SET_METHOD(exports, "drawImage", drawImage);
  NODE_SET_METHOD(exports, "imageSize", imageSize);
  NODE_SET_METHOD(exports, "fontSize", fontSize);
  NODE_SET_METHOD(exports, "textSize", textSize);
  NODE_SET_METHOD(exports, "screenSize", screenSize);
  NODE_SET_METHOD(exports, "processEvents", processEvents);

  NODE_SET_METHOD(exports, "mainMenu", mainMenu);

  NODE_SET_METHOD(exports, "debug", debug);

  NODE_SET_METHOD(exports, "startServer", startServer);
  NODE_SET_METHOD(exports, "stopServer", stopServer);
  NODE_SET_METHOD(exports, "checkForNewClients", checkForNewClients);
  NODE_SET_METHOD(exports, "getNumClients", getNumClients);
  NODE_SET_METHOD(exports, "broadcastAvailability", broadcastAvailability);
  NODE_SET_METHOD(exports, "serverSend", serverSend);
  NODE_SET_METHOD(exports, "serverRead", serverRead);
  NODE_SET_METHOD(exports, "serverIsConnected", serverIsConnected);

  NODE_SET_METHOD(exports, "findServer", findServer);

  NODE_SET_METHOD(exports, "startClient", startClient);
  NODE_SET_METHOD(exports, "stopClient", stopClient);
  NODE_SET_METHOD(exports, "clientSend", clientSend);
  NODE_SET_METHOD(exports, "clientSendBuffer", clientSendBuffer);
  NODE_SET_METHOD(exports, "clientSendAudio", clientSendAudio);
  NODE_SET_METHOD(exports, "clientRead", clientRead);
  NODE_SET_METHOD(exports, "clientIsConnected", clientIsConnected);

  NODE_SET_METHOD(exports, "getAudio", getAudio);
  NODE_SET_METHOD(exports, "queueAudio", queueAudio);
  NODE_SET_METHOD(exports, "mixAndQueueAudio", mixAndQueueAudio);
  NODE_SET_METHOD(exports, "clearAudioQueues", clearAudioQueues);
  NODE_SET_METHOD(exports, "audioPlaybackQueueSize", audioPlaybackQueueSize);

  NODE_SET_METHOD(exports, "drawPolygon", drawPolygon);
  NODE_SET_METHOD(exports, "fillBubble", fillBubble);

  // questions

  NODE_SET_METHOD(exports, "askRadioQuestion", askRadioQuestion);
  NODE_SET_METHOD(exports, "askCheckboxQuestion", askCheckboxQuestion);
  NODE_SET_METHOD(exports, "askOneLineQuestion", askOneLineQuestion);
  NODE_SET_METHOD(exports, "askMultiLineQuestion", askMultiLineQuestion);

  // Audio

  NODE_SET_METHOD(exports, "unpauseAudioRecording", unpauseAudioRecording);
  NODE_SET_METHOD(exports, "pauseAudioRecording", pauseAudioRecording);
  NODE_SET_METHOD(exports, "unpauseAudioPlayback", unpauseAudioPlayback);
  NODE_SET_METHOD(exports, "pauseAudioPlayback", pauseAudioPlayback);

  NODE_SET_METHOD(exports, "serverStartLoggingAudio", serverStartLoggingAudio);
  NODE_SET_METHOD(exports, "serverStopLoggingAudio", serverStopLoggingAudio);

  NODE_SET_METHOD(exports, "startTextInput", startTextInput);
  NODE_SET_METHOD(exports, "stopTextInput", stopTextInput);
}

NODE_MODULE(NODE_GYP_MODULE_NAME, initModule)
