'use strict';

const args = require('yargs')
      // .option('condition', {
      //     describe: "Specify the default condition for incoming connections",
      //     choices: ['autoturn', 'youturn', 'flappy_autoturn', 'flappy_youturn'],
      //     default: 'flappy_youturn'})
      // .option('log-format', {
      //     describe: "Specify the default log format for incoming connections",
      //     choices: ['native', 'flappy', 'sf'],
      //     default: 'native'})
      .option('debug', {
          describe: "enable debug features",
          type: 'boolean',
          default: false})
      .option('id', {
          describe: "Specify the unique identifier for this experiment session",
          type: 'string',
          default: '1'})
      .option('line-ending', {
          describe: "Line endings to use.",
          choices: ['crlf', 'lf'],
          default: 'crlf'})
      .option('port', {
          describe: "The port to connect on",
          type: 'number',
          default: 3000})
      .option('logs', {
          describe: "The directory used to store the log files",
          default: 'logs/'})
      .option('nohumans', {
          describe: "This model server will ignore all keyboard input.",
          type: 'boolean',
          default: false})
      .option('bot', {
          describe: 'Set the player to be controlled by the bot. 0 means disabled.',
          choices: [0,1,2],
          type: 'number',
          default: 0})
      .option('nodrawing', {
          describe: "This model server will not do any graphics drawing.",
          type: 'boolean',
          default: false})
      .option('stepping', {
          describe: "How the game logic is stepped forward.",
          options: ['timer', 'model'],
          default: 'timer'})
      .option('disconnect-action', {
          describe: "What to do if a model disconnects. Close the server or return to lobby?",
          options: ['lobby', 'quit'],
          default: 'lobby'})
      .help()
      .argv;

const net = require('net');
const path = require('path');
var game = require('./game');
var server = require('./server');
var graphics = require('./serverGraphics');
var JSONStream = require('JSONStream');
var build = require('./version').build;
var ModelServer = require('./modelserver');

const lineEndingTranslation = {'crlf': '\r\n', 'lf': '\n'};

const newLine = lineEndingTranslation[args.lineEnding];

const keyMapping = {};
keyMapping['Space'] = {player: 0, key: game.FIRE};
keyMapping['Left'] = {player: 0, key: game.LEFT};
keyMapping['Right'] = {player: 0, key: game.RIGHT};
keyMapping['Up'] = {player: 0, key: game.THRUST};
keyMapping['Left Shift'] = {player: 1, key: game.FIRE};
keyMapping['Right Shift'] = {player: 1, key: game.FIRE};
keyMapping['A'] = {player: 1, key: game.LEFT};
keyMapping['D'] = {player: 1, key: game.RIGHT};
keyMapping['W'] = {player: 1, key: game.THRUST};

const optionsMapping = {'1': 0,
                        '2': 1};

function GraphicsModelServer(options) {
    ModelServer.call(this, options);
    this.grHelper = new graphics.GraphicsHelper('.')
}

GraphicsModelServer.prototype = Object.create(ModelServer.prototype);

GraphicsModelServer.prototype.broadcastBuffer = function (state) {
    this.broadcastScreenState(this.reconstructGameState(state));
}

GraphicsModelServer.prototype.broadcastScreenState = function (state) {
    this.state = state;
    // No drawing override
    if (state.screen === 'game' && !this.drawing) return;
    this.grHelper.drawScreenState(this.state);
};

GraphicsModelServer.prototype.inputEvent = function (event) {
    this.lg('event', event);
    if (event.type === 'quit' || (event.type === 'windowEvent' && event.windowEvent === 'close')) {
        this.shutdown();
    } else if (this.mode === 'game') {
        this.game.handleInputEvent(event);
    }
};

GraphicsModelServer.prototype.processEvents = function () {
    if (this.mode === 'lobby') {
        graphics.processEvents((e) => {
            if (e.type === 'quit') {
                this.shutdown();
            } else if (e.type === 'windowEvent') {
                // console.log('windowevent', e.windowEvent)
                if (e.windowEvent === 'exposed') {
                    this.grHelper.drawScreenState(this.lobbyScreenState);
                } else if (e.windowEvent === 'close') {
                    this.shutdown();
                }
            } else if (!args.nohumans && e.type === 'keyDown' && !e.repeat && e.name in optionsMapping) {
                if (this.agentPlayer === optionsMapping[e.name]) {
                    this.playersReady[this.agentPlayer] = false;
                    this.agentPlayer = -1;
                } else {
                    if (this.agentPlayer >= 0) this.playersReady[this.agentPlayer] = false;
                    this.agentPlayer = optionsMapping[e.name];
                    this.playersReady[this.agentPlayer] = true;
                }
                this.updateLobbyScreenState();
                this.broadcastScreenState(this.lobbyScreenState);
            } else if (!args.nohumans && e.type === 'keyDown' && !e.repeat && keyMapping[e.name]) {
                var map = keyMapping[e.name];
                if (map.key === game.FIRE) {
                    this.maybeStartLogging();
                    this.playersReady[map.player] = true;
                    this.updateLobbyScreenState();
                    this.broadcastScreenState(this.lobbyScreenState);
                    this.maybeStartGame();
                }
            }
        });
    } else if (this.mode === 'game') {
        graphics.processEvents((e) => {
            this.inputEvent(e);
            if (e.type === 'windowEvent') {
                // console.log('windowevent', e.windowEvent)
                if (e.windowEvent === 'exposed') {
                    this.grHelper.drawScreenState(this.state);
                }
            } else if (!args.nohumans && e.type === 'keyDown' && !e.repeat && e.name === 'Escape') {
                this.startLobby();
            } else if (!args.nohumans && e.type === 'keyDown' && !e.repeat && keyMapping[e.name]) {
                var map = keyMapping[e.name];
                if (!args.nohumans)
                    this.inputEvent({type: 'gameKeyDown', key: map.key, player: map.player});
            } else if (!args.nohumans && e.type === 'keyUp' && keyMapping[e.name]) {
                var map = keyMapping[e.name];
                if (!args.nohumans)
                    this.inputEvent({type: 'gameKeyUp', key: map.key, player: map.player});
            }
        });
    } else if (this.mode === 'score') {
        graphics.processEvents((e) => {
            if (e.type === 'quit') {
                this.shutdown();
            } else if (e.type === 'windowEvent') {
                // console.log('windowevent', e.windowEvent)
                if (e.windowEvent === 'exposed') {
                    this.grHelper.drawScreenState(this.scoreScreenState);
                } else if (e.windowEvent === 'close') {
                    this.shutdown();
                }
            } else if (!args.nohumans && e.type === 'keyDown' && !e.repeat && e.name in optionsMapping) {
                if (this.agentPlayer === optionsMapping[e.name]) {
                    this.playersReady[this.agentPlayer] = false;
                    this.agentPlayer = -1;
                } else {
                    if (this.agentPlayer >= 0) this.playersReady[this.agentPlayer] = false;
                    this.agentPlayer = optionsMapping[e.name];
                    this.playersReady[this.agentPlayer] = true;
                }
                this.updateScoreScreenState();
                this.broadcastScreenState(this.scoreScreenState);
            } else if (!args.nohuman && e.type === 'keyDown' && !e.repeat && keyMapping[e.name]) {
                var map = keyMapping[e.name];
                if (map.key === game.FIRE) {
                    this.playersReady[map.player] = true;
                    this.updateScoreScreenState();
                    this.broadcastScreenState(this.scoreScreenState);
                    this.maybeStartGame();
                }
            }
        });
    }

    this.eventTimer = setTimeout(this.processEvents.bind(this), 0);
};

GraphicsModelServer.prototype.init = function () {
    ModelServer.prototype.init.call(this);
    this.processEvents();
};

GraphicsModelServer.prototype.shutdown = function () {
    ModelServer.prototype.shutdown.call(this, () => { clearTimeout(this.eventTimer); });
};

var logdir, font;
// A hack to detect the platform and context it's running in.
var p = path.parse(process.argv[0]);
var d = p.dir.split(path.sep);
if (process.pkg &&
    d.length >= 3 &&
    d[d.length-1] === 'MacOS' &&
    d[d.length-2] === 'Contents') {
    // OSX inside a .app
    logdir = path.join(path.dirname(process.argv[0]), '../../../logs/');
    font = path.join(path.dirname(process.argv[0]), '../Resources/', 'freesansbold.ttf');
    console.log('osx app', path.dirname(process.argv[0]));
} else {
    // We're either on windows or running the executable
    // outside a .app directory structure.
    logdir = args.logs;
    font = 'freesansbold.ttf';
}

console.log('logdir:', logdir);
console.log('font:', font);

graphics.initGraphics(font, {disabled: true});
graphics.createWindows(false, 'Model Server ', false);

args.logdir = logdir;
var m = new GraphicsModelServer(args);

m.init();
