const net = require('net');
const screens = require('./screens');
const path = require('path');
var game = require('./game');
var server = require('./server');
var JSONStream = require('JSONStream');
var build = require('./version').build;
const SkilledAgent = require('./skilledagent');

const lineEndingTranslation = {'crlf': '\r\n', 'lf': '\n'};

const keyMapping = {};
keyMapping['Space'] = {player: 0, key: game.FIRE};
keyMapping['Left'] = {player: 0, key: game.LEFT};
keyMapping['Right'] = {player: 0, key: game.RIGHT};
keyMapping['Up'] = {player: 0, key: game.THRUST};
keyMapping['Left Shift'] = {player: 1, key: game.FIRE};
keyMapping['Right Shift'] = {player: 1, key: game.FIRE};
keyMapping['A'] = {player: 1, key: game.LEFT};
keyMapping['D'] = {player: 1, key: game.RIGHT};
keyMapping['W'] = {player: 1, key: game.THRUST};

function ModelServer(options) {
    server.Server.call(this, options.id, {modelRun: true}, options.logs, [], false);
    this.newLine = lineEndingTranslation[options.lineEnding];
    this.clients = [];
    this.state = {screen:'blank'};
    this.lobbyScreenState = {screen:'message',
                             lines: [`Listening on port ${options.port} for models`,
                                     "",
                                     "Press '1' or '2' to toggle a bot player",
                                     "",
                                     "Player 1 Human Controls",
                                     "Left and Right Arrows - Turn CCW and CW",
                                     "Up Arrow - Thrust",
                                     "Space - Fire",
                                     "",
                                     "Player 2 Human Controls",
                                     "A and D - Turn CCW and CW",
                                     "W - Thrust",
                                     "Left Shift - Fire",
                                     "",
                                     "Press ESC to prematurely end a game."],
                             corner: 'v'+build,
                             continue: ['','']};
    this.scoreScreenState = {screen:'message',
                             lines: ["Score: TODO"],
                             continue: ['Player 1 - Press FIRE to Continue',
                                        'Player 2 - Press FIRE to Continue']};
    this.noDrawGameScreenState = {screen:'message',
                                  lines: ["Game"],
                                  continue: ['','']};
    // Used for reconstructing binary game states
    this.gameState = {screen: 'game',
                      ships: [{alive: true, x:0, y:0, angle:0, color: 0x00FFFF},
                              {alive: true, x:0, y:0, angle:0, color: 0xFF00FF}],
                      fortress: {alive: true, x:0, y:0, angle: 0, explosion: false},
                      missiles: [],
                      shells: [],
                      smallhex: {radius: 40, x: 0, y: 0, angle: 0, full: true},
                      bighex: {radius: 200, x: 0, y: 0, angle: 0, full: true},
                      points: 0,
                      vulnerability: 0,
                      debug: false,
                      latency: [0,0]
                     };
    this.playersReady = new Array(2);
    this.stepping = options.stepping;
    this.drawing = !options.nodrawing;
    this.port = options.port;
    this.huntForPort = options['hunt-for-port'];
    this.disconnectAction = options['disconnect-action'] || 'lobby';

    if (options.bot === 1) this.agentPlayer = 0;
    else if (options.bot === 2) this.agentPlayer = 1;
    else this.agentPlayer = -1;

    this.resetPlayersReady();

    this.verbose = options.verbose;
    this.debug = options.debug;
    if (this.debug) {
	this.dumpStream = fs.createWriteStream(options['debug-output'], {flags:'a'});
    }
}

ModelServer.prototype = Object.create(server.Server.prototype);

ModelServer.prototype.debugLog = function (s) {
    if (this.debug) {
	this.dumpStream.write(s);
	this.dumpStream.write('\n');
    }
}

ModelServer.prototype.updateScoreScreenState = function () {
    for (let i=0; i<this.scoreScreenState.continue.length; i++) {
        var txt = this.playersReady[i] ? 'Ready':'Press FIRE to Continue';
        let player = this.agentPlayer === i ? 'Bot':'Player';
        this.scoreScreenState.continue[i] = `${player} ${(i+1)} - ${txt}`;
    }
};

ModelServer.prototype.updateLobbyScreenState = function () {
    for (let i=0; i<this.lobbyScreenState.continue.length; i++) {
        var txt = this.playersReady[i] ? 'Ready':'Press FIRE to Join';
        let player = this.agentPlayer === i ? 'Bot':'Player';
        this.lobbyScreenState.continue[i] = `${player} ${(i+1)} - ${txt}`;
    }
};

ModelServer.prototype.reconstructProjectiles = function(dest, buf, offset, num) {
    if (dest.length < num) {
        while (dest.length < num) {
            dest.push({x:0, y:0, angle: 0});
        }
    } else {
        dest.length = num;
    }
    for (let i=0; i<num; i++) {
        dest[i].x = buf.readFloatLE(offset);
        dest[i].y = buf.readFloatLE(offset+4);
        dest[i].angle = buf.readFloatLE(offset+8);
        offset += 12;
    }
};

ModelServer.prototype.reconstructGameState = function(buf) {
    this.gameState.points = buf.readUInt16LE(0);

    this.gameState.ships[0].alive = buf.readUInt8(2) !== 0;
    this.gameState.ships[0].x = buf.readFloatLE(3);
    this.gameState.ships[0].y = buf.readFloatLE(7);
    this.gameState.ships[0].angle = buf.readFloatLE(11);

    this.gameState.ships[1].alive = buf.readUInt8(15) !== 0;
    this.gameState.ships[1].x = buf.readFloatLE(16);
    this.gameState.ships[1].y = buf.readFloatLE(20);
    this.gameState.ships[1].angle = buf.readFloatLE(24);

    this.gameState.fortress.alive = buf.readUInt8(28) !== 0;
    this.gameState.fortress.angle = buf.readFloatLE(29);
    this.gameState.smallhex.angle = this.gameState.fortress.angle;
    this.gameState.smallhex.full = buf.readUInt8(33) !== 0;
    this.gameState.fortress.explosion = buf.readUInt8(34) !== 0;

    var numMissiles = buf.readUInt8(35);
    var numShells = buf.readUInt8(36);

    this.reconstructProjectiles(this.gameState.missiles, buf, 37, numMissiles);
    this.reconstructProjectiles(this.gameState.shells, buf, 37+numMissiles*3*4, numShells);

    this.gameState.debug = buf.readUInt8(37+(numMissiles+numShells)*3*4) !== 0;
    if (this.gameState.debug) {
        this.gameState.latency[0] = buf.readUInt32LE(37+(numMissiles+numShells)*3*4+1);
        this.gameState.latency[1] = buf.readUInt32LE(37+(numMissiles+numShells)*3*4+5);
    }

    if (this.gameState.debug && this.debug) {
        this.gameState.sendDelay = this.sendDelay;
    }
    // console.log('state', this.gameState);
    return this.gameState;
}

ModelServer.prototype.broadcastBuffer = function (state) {
}

ModelServer.prototype.broadcastScreenState = function (state) {
    this.state = state;
};

ModelServer.prototype.inputEvent = function (event) {
    this.lg('event', event);
    if (this.mode === 'game') {
        this.game.handleInputEvent(event);
    }
};

ModelServer.prototype.resetPlayersReady = function () {
    this.playersReady[0] = false;
    this.playersReady[1] = false;
    if (this.agentPlayer >= 0) this.playersReady[this.agentPlayer] = true;
};

ModelServer.prototype.numPlayersReady = function () {
    var c = 0;
    for (let i=0; i<this.playersReady.length; i++) {
        if (this.playersReady[i]) c += 1;
    }
    return c;
}

ModelServer.prototype.arePlayersReady = function () {
    var ready = true;
    for (let i=0; i<this.playersReady.length; i++) {
        if (!this.playersReady[i]) ready = false;
    }
    return ready;
}

ModelServer.prototype.maybeStartLogging = function () {
    if (!this.logStream)
        this.lgStartExperiment({build: build, date: new Date().toString(), sessionId: this.sessionId});
};

// Game calls this in timer mode when the game is over
ModelServer.prototype.nextScreen = function () {
    if (this.mode === 'game') this.startScore();
    else if (this.verbose) console.log('(:message-type :error :text "impossible nextScreen")');
}

ModelServer.prototype.maybeStartGame = function () {
    if (this.arePlayersReady()) {
        this.startGame();
    }
};

ModelServer.prototype.startTCPServer = function () {
    this.serverSocket = net.createServer();
    this.serverSocket.on('connection', this.onConnect.bind(this));
    this.serverSocket.on('error', (err) => {
        if (err.code === 'EADDRINUSE' && this.huntForPort) {
            this.port += 1;
            this.serverSocket.listen(this.port);
            this.debugLog(`--- Port already in use. Attempting ${this.port}.`);
        } else {
            console.log(`(:message-type :error :text "${err}")`);
            this.debugLog(`--- Error starting server: ${err}`);
        }
    });
    this.serverSocket.on('listening', () => {
        console.log(`(:message-type :ready :port ${this.port})`);
        this.debugLog(`--- Listening on port ${this.port}`);
    });

    this.serverSocket.listen(this.port);
}

ModelServer.prototype.init = function () {
    server.Server.prototype.init.call(this);
    this.mode = 'lobby';
    this.game = null;

    this.startTCPServer();
    this.clients = [];

    process.on('SIGINT', () => { this.shutdown(); });

    this.updateLobbyScreenState();
    this.broadcastScreenState(this.lobbyScreenState);
    this.debugLog('--- Start');
};

ModelServer.prototype.doDisconnectAction = function () {
    if (this.disconnectAction === 'lobby') {
        if (this.mode === 'game') this.nextScreen();
    } else if (this.disconnectAction === 'quit') {
        this.shutdown();
    } else {
        console.log(`(:message-type :error :text "Unknown disconnect action: ${this.disconnectAction}")`);
        this.shutdown();
    }
};

ModelServer.prototype.startLobby = function () {
    if (this.verbose) console.log(`(:message-type :state-change :state "lobby" :clients ${this.clients.length})`);
    this.mode = 'lobby';
    if (this.game) this.game.cleanup();
    this.game = null;
    this.lgEndExperiment({'bonus': this.bonus, 'finished': true});
    for (let i=0; i<this.clients.length; i++) {
        var o = this.addConfigSettings(this.clients[i], {msgType: 'stateChange',
                                                         state: 'config'});
        this.clientEmit(this.clients[i], o);
        this.clients[i].state = 'config';
    }
    this.resetPlayersReady();
    this.updateLobbyScreenState();
    this.broadcastScreenState(this.lobbyScreenState);
};

ModelServer.prototype.updateGame = function () {
    if (this.agentPlayer >= 0) this.agent.update();
    this.game.updateHelper();
    var s = this.game.getModelState();
    for (let i=0; i<this.clients.length; i++) {
        this.clientEmit(this.clients[i], {msgType: 'gameStateUpdate', gameState: s});
    }
    if (this.game.game.isFinished()) {
        this.startScore();
    } else {
        this.game.requestUpdate();
    }
};

ModelServer.prototype.startGame = function () {
    if (this.verbose) console.log('(:message-type :state-change :state "game")');
    this.mode = 'game';
    this.game = new screens.Game(1, false, 'shield+shell');
    this.game.initHelper(this);
    if (this.agentPlayer >= 0) {
        this.agent = new SkilledAgent(this.game.game, this.game.game.players[this.agentPlayer]);
    }
    if (!this.drawing) this.broadcastScreenState(this.noDrawGameScreenState);
    if (this.stepping === 'timer') {
        this.game.update = this.updateGame.bind(this);
        this.game.requestUpdate();
    }

    this.resetPlayersReady();

    var s = this.game.getModelState();
    for (let i=0; i<this.clients.length; i++) {
        this.clients[i].state = 'game';
        this.clientEmit(this.clients[i], {msgType: 'stateChange',
                                          state: 'game',
                                          gameState: s});
    }
};

ModelServer.prototype.startScore = function () {
    if (this.verbose) console.log('(:message-type :state-change :state "score")');
    this.mode = 'score';
    this.game.cleanup();
    this.resetPlayersReady();
    this.scoreScreenState.lines[0] = "Score: " + this.game.game.points.toString();
    this.updateScoreScreenState();
    this.broadcastScreenState(this.scoreScreenState);
    for (let i=0; i<this.clients.length; i++) {
        this.clients[i].state = 'game';
        this.clientEmit(this.clients[i], {msgType: 'stateChange',
                                          state: 'score',
                                          points: this.game.game.points,
                                          rawPoints: this.game.game.rawPoints,
                                          bonus: this.game.game.bonus,
                                          totalBonus: this.bonus});
    }
};

ModelServer.prototype.addConfigSettings = function (client, obj) {
    obj.stepping = this.stepping;
    obj.id = this.sessionId;
    obj.player = client.player;
    obj.bot = this.agentPlayer;
    obj.fps = 60;
    obj.build = build;
    return obj;
};

function Client(socket, pnum) {
    this.socket = socket;
    this.state = 'config';
    this.player = pnum;
    this.id = 'model1';
}

Client.prototype = {};

ModelServer.prototype.clientEmit = function(client, data) {
    this.debugLog(`>>> ${client.player} ${JSON.stringify(data)}`);
    client.socket.write(JSON.stringify(data) + this.newLine);
}

ModelServer.prototype.onConnect = function (socket) {
    var client = new Client(socket, this.clients.length%2);

    this.clients.push(client);
    socket.setNoDelay(true);

    this.debugLog(`--- New client ${client.player} from ${socket.remoteAddress}`);

    var stream = JSONStream.parse();
    socket.pipe(stream);

    stream.on('data', (data) => { this.onData(client, data); });
    stream.on('error', (err) => { this.onParseError(client, err); });
    socket.on('error', (err) => { this.onError(client, err); });
    socket.on('close', (data) => { this.onClose(client); });
    this.clientEmit(client, this.addConfigSettings(client,
						   {msgType: 'stateChange',
						    state: 'config'}));
    if (this.verbose) console.log(`(:message-type :info :text "Client connected")`);
};

ModelServer.prototype.delClient = function (client) {
    client.socket.end();
    this.clients.splice(this.clients.indexOf(client), 1);
};

ModelServer.prototype.translateKeyName = function (key) {
    if (key === game.FIRE ||
        key === game.LEFT ||
        key === game.RIGHT ||
        key === game.THRUST)
        return key;
    else if (key === 'fire') return game.FIRE;
    else if (key === 'thrust') return game.THRUST;
    else if (key === 'left') return game.LEFT;
    else if (key === 'right') return game.RIGHT;
    else return null;
}

ModelServer.prototype.handleClientKeyEvent = function (client, eventName, data) {
    var k = this.translateKeyName(data.value);
    if (k === null) {
        this.clientEmit(client, {msgType: "response",
                                 cmd: data.cmd,
                                 token: data.token,
                                 success: false,
                                 error: "unknown key: "+data.cmd});
    } else {
        this.inputEvent({type: eventName,
                         key: k,
                         player: client.player});
        this.clientEmit(client, {msgType: "response",
                                 cmd: data.cmd,
                                 token: data.token,
                                 success: true});
    }
};

ModelServer.prototype.onParseError = function (client, err) {
    console.log('(:message-type :error :text "failed to parse json")');
    // var stream = JSONStream.parse();
    // socket.pipe(stream);

    // stream.on('data', (data) => { this.onData(client, data); });
    // stream.on('error', (err) => { this.onParseError(client, err); });
    // socket.on('close', (data) => { this.onClose(client); });

}

ModelServer.prototype.onData = function (client, data) {
    this.debugLog(`<<< ${client.player} ${JSON.stringify(data)}`);
    var error = null;
    if (data.cmd === undefined) {
        this.clientEmit(client, {msgType: "response",
				 token: data.token,
				 error: "no command",
				 success: false});
        return;
    } else if (data.cmd === 'state') {
        // You can always request the current state. Maybe useful for
        // debugging?
        var r = {msgType: "response",
                 token: data.token,
                 cmd: data.cmd,
                 success: true};
        if (client.state === 'config' || client.state === 'ready')
            r.state = client.state;
        else
            r.state = this.mode;
        this.clientEmit(client, r);
        return;
    } if (data.cmd === 'drawing') {
        if (data.value) this.drawing = true;
        else this.drawing = false;
        this.clientEmit(client, {msgType:'response',
                                 token: data.token,
                                 cmd: data.cmd,
                                 success: true,
                                 value: this.drawing});
        if (this.mode === 'game' && !this.drawing)
            this.broadcastScreenState(this.noDrawGameScreenState);
        return;
    }

    if (client.state === 'config') {
        if (data.cmd === 'continue') {
            client.state = 'ready';
            this.clientEmit(client, {msgType: "response",
                                     token: data.token,
                                     cmd: data.cmd,
                                     clientState: "ready",
                                     success: true});
            this.maybeStartLogging();
            this.playersReady[client.player] = true;
            this.updateLobbyScreenState();
            this.broadcastScreenState(this.lobbyScreenState);
            this.maybeStartGame();
            return;
        } else if (data.cmd === 'quit') {
            this.delClient(client);
            if (this.disconnectAction === 'quit') {
                this.shutdown();
            }
            return;
        } else if (data.cmd === 'stepping') {
            if (data.value === 'timer') this.stepping = 'timer';
            else if (data.value === 'model') this.stepping = 'model';
        } else if (data.cmd === 'id') {
            if (data.value) {
                this.sessionId = data.value.toString();
            } else {
                error = 'Must supply a value for "id"';
            }
        } else if (data.cmd === 'player') {
            if (data.value === 0 || data.value === 1) {
                client.player = data.value;
            } else {
                error = 'Player must be 0 or 1';
            }
        } else if (data.cmd === 'bot') {
            if (data.value === 0 || data.value === 1 || data.value === -1) {
                if (this.agentPlayer >= 0) this.playersReady[this.agentPlayer] = false;
                this.agentPlayer = data.value;
                if (this.agentPlayer >= 0) this.playersReady[this.agentPlayer] = true;
                this.updateLobbyScreenState();
                this.broadcastScreenState(this.lobbyScreenState);
            } else {
                error = 'Bot player must be 0 or 1. -1 will disable the bot player.';
            }
        } else {
            error = 'unknown command';
        }
        var response = this.addConfigSettings(client, {msgType: 'response',
                                                       cmd: data.cmd,
                                                       token: data.token});
        if (error) {
            response.success = false;
            response.error = error;
        } else {
            response.success = true;
        }
        this.clientEmit(client, response);
    } if (client.state === 'ready') {
        if (data.cmd === 'quit') {
            this.delClient(client);
            if (this.disconnectAction === 'quit') {
                this.shutdown();
            }
            return;
        }
        this.clientEmit(client, {msgType: 'response',
                                 cmd: data.cmd,
                                 clientState: "ready",
                                 token: data.token,
                                 success: false,
                                 error: 'unknown command: ' + data.cmd});
    } else if (client.state === 'game') {
        if (data.cmd === 'quit') {
            this.delClient(client);
            this.doDisconnectAction();
        }
        if (this.mode === 'game') {
            if (data.cmd === 'keydown') {
                this.handleClientKeyEvent(client, 'gameKeyDown', data);
            } else if (data.cmd === 'keyup') {
                this.handleClientKeyEvent(client, 'gameKeyUp', data);
            } else if (data.cmd === 'gameState') {
                this.clientEmit(client, {msgType: 'response',
                                         cmd: data.cmd,
                                         token: data.token,
                                         success: false,
                                         error: "The gameState command is obsolete"});
            } else if (data.cmd === 'step') {
                if (this.stepping === 'model') {
                    this.playersReady[client.player] = true;
                    if (this.arePlayersReady() || this.clients.length === 1) {
                        if (this.agentPlayer >= 0) this.agent.update();
                        this.game.updateHelper();
                        var s = this.game.getModelState();
                        for (let i=0; i<this.clients.length; i++)
                            this.clientEmit(this.clients[i], {msgType: 'response',
                                                              cmd: data.cmd,
                                                              // token: 'TODO',
                                                              success: true,
                                                              gameState: s});
                        this.resetPlayersReady();
                        if (this.game.game.isFinished()) this.startScore();
                    }
                } else {
                    this.clientEmit(client, {msgType: 'response',
                                             cmd: data.cmd,
                                             token: data.token,
                                             success: false,
                                             error: 'stepping mode is not set to "model"'});
                }
            } else {
                this.clientEmit(client, {msgType: 'response',
                                         cmd: data.cmd,
                                         token: data.token,
                                         success: false,
                                         error: 'unknown command'});
            }
        } else if (this.mode === 'score') {
            if (data.cmd === 'quit') {
                this.delclient(client);
                this.doDisconnectAction();
            } else if (data.cmd === 'continue') {
                this.playersReady[client.player] = true;
                this.clientEmit(client, {msgType: 'response',
                                         cmd: data.cmd,
                                         token: data.token,
                                         success: true});
                this.updateScoreScreenState();
                this.broadcastScreenState(this.scoreScreenState);
                this.maybeStartGame();
            } else {
                this.clientEmit(client, {msgType: 'response',
                                         cmd: data.cmd,
                                         token: data.token,
                                         success: false,
                                         error: 'unknown command'});
            }
        }
    }
};

ModelServer.prototype.onError = function (client, err) {
    console.log(`(:message-type :error :text "socket error: ${err}")`);
    if (this.clients.indexOf(client) >= 0) {
        this.delClient(client);
        this.doDisconnectAction();
    }
};

ModelServer.prototype.onClose = function (client) {
    if (this.clients.indexOf(client) >= 0) {
        this.delClient(client);
        this.doDisconnectAction();
    }
};

ModelServer.prototype.shutdown = function (extrafn) {
    clearTimeout(this.shutdownTimer);
    this.shutdownTimer = setTimeout(() => {
        if (this.mode === 'game') this.game.cleanup();
        if (this.logStream) this.lgEndExperiment({'bonus': this.bonus, 'finished': true});
        //clearTimeout(this.eventTimer);
        if (extrafn) extrafn();
        this.serverSocket.close();
        for (let i=0; i<this.clients.length; i++) {
            // this.clients[i].socket.end();
            this.clients[i].socket.destroy();
        }
        this.clients = [];
    if (this.verbose) console.log('(:message-type :info :text "shutdown complete.")');
    }, 0);
    this.debugLog('--- Shutdown complete');
};


module.exports = ModelServer;
