#include "graphics.hh"

static bool debug = false;

Uint32 checksum(const Uint8 *data, Uint32 bytes) {
  Uint32 sum = 0;
  for (Uint32 i=0; i<bytes; i++) sum += data[i];
  return sum;
}

void hexDump(const Uint8 *data, Uint32 bytes) {
  std::cout << std::hex;
  for (Uint32 i=0; i<bytes; i++) {
    std::cout << (int)data[i];
    if (i % 90 == 0) std::cout << std::endl;
    else std::cout << " ";
  }
  std::cout << std::dec;
}

bool TCPRead( TCPsocket socket, const Uint8 *buf, int bytesToRead ) {
  // memset(buf, 0xFF, header.size);
  // int bytes = 0;
  while (bytesToRead > 0) {
    int ret = SDLNet_TCP_Recv( socket, (void *)buf, bytesToRead );
    if (ret <= 0) {
      if (ret < 0)
        std::cout << "SDLNet_TCP_Recv: " << ret << " " << SDLNet_GetError() << std::endl;
      else
        std::cout << "SDLNet_TCP_Recv: Connection closed by peer" << std::endl;
      return false;
    }
    buf = buf + ret;
    bytesToRead -= ret;
  }
  return true;
}

bool TCPSend( TCPsocket socket, const void *data, Uint32 bytes ) {
  int ret = SDLNet_TCP_Send( socket, data, bytes );
  if (ret < bytes ) {
    std::cout << "SDLNet_TCP_Send: " << SDLNet_GetError() << "\n";
    return false;
  }
  return true;
}


Server::Server(int port) {
  mPort = port;
  mSocket = NULL;
  mSet = SDLNet_AllocSocketSet(2);
  mConnected = true;
}

Server::~Server() {
  shutdown();
}

bool Server::start() {
  if( SDLNet_ResolveHost( &mIP, NULL, mPort )) {
    std::cout << "SDLNet_ResolveHost: " << SDLNet_GetError() << "\n";
    return false;
  }

  mSocket = SDLNet_TCP_Open( &mIP );
  if( !mSocket ) {
    std::cout << "SDLNet_TCP_Open: " << SDLNet_GetError() << "\n";
    return false;
  }

  // SDLNet_ResolveHost( &ip, "127.0.0.1", BROADCAST_PORT );
  mBroadcastPacket = SDLNet_AllocPacket( 1 + sizeof( Uint16 ));
  SDLNet_ResolveHost( &mBroadcastPacket->address, "255.255.255.255", BROADCAST_PORT );
  mBroadcastPacket->len = 1 + sizeof( Uint16 );
  mBroadcastPacket->data[0] = BROADCAST_MAGIC;
  SDLNet_Write16(mPort, (mBroadcastPacket->data+1));
  mBroadcastSocket = SDLNet_UDP_Open( 0 );

  if( !mBroadcastSocket ) {
    std::cout << "SDLNet_UDP_Open: " << SDLNet_GetError() << "\n";
  }

  return true;
}

bool Server::broadcastAvailability() {
  int ret;
  if (( ret = SDLNet_UDP_Send( mBroadcastSocket, -1, mBroadcastPacket )) == 0) {
    std::cout << "SDLNet_UDP_Send: "<< SDLNet_GetError() << "\n";
    return false;
  } else {
    // std::cout << "SDLNet_UDP_Send success "<<  ret << " len: " << mBroadcastPacket->len << "\n";
    return true;
  }
}

bool Server::checkForNewClients() {
  TCPsocket s;

  s = SDLNet_TCP_Accept( mSocket );
  if( s ) {
    mClients.push_back( s );
    SDLNet_TCP_AddSocket( mSet, s );
    return true;
  } else {
    return false;
  }
}

void Server::shutdown() {
  for( size_t i=0; i<mClients.size(); i++ ) {
    SDLNet_TCP_Close( mClients[i] );
  }
  SDLNet_TCP_Close( mSocket );
  SDLNet_FreeSocketSet( mSet );
}

void Server::getClientMessages(size_t clientIndex, std::vector<Uint8> &types, std::vector<std::string> &messages) {
  if( clientIndex >= mClients.size() ) return;

  while( SDLNet_CheckSockets( mSet, 0 ) > 0) {
    if( SDLNet_SocketReady( mClients[clientIndex] )) {
      MessageHeader header;

      if( !TCPRead( mClients[clientIndex], (Uint8 *)&header, sizeof( header )) ) {
        mConnected = false;
        break;
      }

      if( debug ) std::cout << "<<< " << std::hex << header.magic << " - " << header.type << " C " << header.checksum << " - " << std::dec << header.size << std::endl;

      if( header.magic != MESSAGE_MAGIC ) {
        std::cout << "Magic mismatch: " << std::hex << MESSAGE_MAGIC << " != " << header.magic << std::dec << std::endl;
        mConnected = false;
        break;
      }
      if( header.type != MESSAGE_TYPE_AUDIO &&
          header.type != MESSAGE_TYPE_STRING &&
          header.type != MESSAGE_TYPE_BUFFER ) {
        std::cout << "Unknown Message Type: " << header.type << std::endl;
        mConnected = false;
        break;
      }
      if( header.size > 0 ) {
        char *buf = new char[header.size];
        if( !TCPRead( mClients[clientIndex], (Uint8 *)buf, header.size )) {
          mConnected = false;
          delete [] buf;
          break;
        }
        Uint32 chk = checksum((Uint8 *)buf, header.size);
        if( header.checksum != chk ) {
          std::cout << "checksum mismatch: " << std::hex << header.checksum << " != " << chk << std::endl;
          hexDump((Uint8*)buf, header.size);
          mConnected = false;
          delete[] buf;
          break;
        }

        std::string s( buf, header.size );
        delete[] buf;

        types.push_back(header.type);
        messages.push_back(s);
      }
    } else {
      // FIXME
      return;
    }
  }
}

bool Server::sendMessage( std::string &msg, Uint8 player ) {
  MessageHeader header;

  header.magic = MESSAGE_MAGIC;
  header.type = MESSAGE_TYPE_STRING;
  header.checksum = checksum((Uint8*)msg.data(), msg.size());
  header.size = msg.size();

  if( !TCPSend( mClients[player], &header, sizeof(header) )) {
    mConnected = false;
    return false;
  }
  if( !TCPSend( mClients[player], msg.data(), msg.size() )) {
    mConnected = false;
    return false;
  }

  return true;
}

bool Server::sendMessage(std::string &msg) {
  MessageHeader header;

  header.magic = MESSAGE_MAGIC;
  header.type = MESSAGE_TYPE_STRING;
  header.checksum = checksum((Uint8*)msg.data(), msg.size());
  header.size = msg.size();

  if( debug ) std::cout << ">>> " << std::hex << header.magic << std::dec << " - " << header.type << " - " << header.size << std::endl;

  for (size_t i=0; i<mClients.size(); i++) {
    if( !TCPSend( mClients[i], &header, sizeof( header ))) {
      mConnected = false;
      return false;
    }
    if( !TCPSend( mClients[i], msg.data(), msg.size() )) {
      mConnected = false;
      return false;
    }
  }

  return true;
}

bool Server::sendAudioData(const std::string &msg, Uint8 player ) {
  MessageHeader header;

  header.magic = MESSAGE_MAGIC;
  header.type = MESSAGE_TYPE_AUDIO;
  header.checksum = checksum((Uint8*)msg.data(), msg.size());
  header.size = msg.size();

  if( debug ) std::cout << ">>> " << std::hex << header.magic << " - " << header.type << " C " << header.checksum << " - " << std::dec << header.size << std::endl;

  if( !TCPSend( mClients[player], &header, sizeof( header ))) {
    mConnected = false;
    return false;
  }
  if( !TCPSend( mClients[player], msg.data(), msg.size() )) {
    mConnected = false;
    return false;
  }

  return true;
}

BroadcastClient::BroadcastClient() {
  mPacket = SDLNet_AllocPacket (1+sizeof(Uint16));
  mSocket = SDLNet_UDP_Open (BROADCAST_PORT);
  mLastServer.host = INADDR_NONE;
  mLastServer.port = 0;

  if( !mSocket ) {
    if( debug ) std::cout << "SDLNet_UDP_Open: " << SDLNet_GetError() << "\n";
  }
}

BroadcastClient::~BroadcastClient() {
  SDLNet_FreePacket( mPacket );
  SDLNet_UDP_Close( mSocket );
}

bool BroadcastClient::checkForServerBroadcasts() {
  int ret = SDLNet_UDP_Recv( mSocket, mPacket );
  if( ret <= 0 ) {
    // std::cout << "SDLNet_UDP_Recv: " << ret << " " << SDLNet_GetError() << "\n";
    return false;
  } else {
    // bool magicMatch = mPacket->len == 1 && mPacket->data[0] == BROADCAST_MAGIC;

    std::cout << "broadcast packet " << mPacket->len << "\n";

    if (mPacket->len == (1+sizeof(Uint16)) && mPacket->data[0] == BROADCAST_MAGIC) {
      int magic = mPacket->data[0];
      std::cout << "Got a broadcast from " << mPacket->address.host << ":" << mPacket->address.port << " " << mPacket->len << " - " << std::hex << magic << std::dec << "\n";

      mLastServer.host = mPacket->address.host;
      mLastServer.port = (Uint16)SDLNet_Read16(mPacket->data+1);
      return true;
    } else {
      return false;
    }
  }
}

Client::Client(const std::string &host, int port) {
  mSocket = NULL;
  mHost = host;
  mPort = port;
  mSet = SDLNet_AllocSocketSet(1);
  mConnected = false;
}

Client::~Client() {
  SDLNet_TCP_Close( mSocket );
  SDLNet_FreeSocketSet( mSet );
}

bool Client::connect() {
  if( SDLNet_ResolveHost( &mIP, mHost.c_str(), mPort ) == -1) {
    std::cout << "SDLNet_ResolveHost: " << SDLNet_GetError() << "\n";
    return false;
  }
  mSocket = SDLNet_TCP_Open( &mIP );
  if( mSocket ) {
    std::cout << "ok we're connected " << mSocket << "\n";
    SDLNet_TCP_AddSocket( mSet, mSocket );
    mConnected = true;
    return true;
  } else {
    std::cout << "SDLNet_TCP_Open: " << SDLNet_GetError() << "\n";
    return false;
  }
}

void Client::getServerMessages(std::vector<Uint8> &types, std::vector<std::string> &messages) {
  while( SDLNet_CheckSockets( mSet, 0 ) > 0) {
    MessageHeader header;
    if (!TCPRead( mSocket, (Uint8*)&header, sizeof(header) )) {
      mConnected = false;
      break;
    }

    if( debug ) std::cout << "<<< " << std::hex << header.magic << " - " << header.type << " C " << header.checksum << " - " << std::dec << header.size << std::endl;

    if( header.magic != MESSAGE_MAGIC ) {
      std::cout << "Magic mismatch: " << std::hex << MESSAGE_MAGIC << " != " << header.magic << std::dec << " | " << header.type << " - " << header.size << std::endl;
      mConnected = false;
      break;
    }
    if( header.type != MESSAGE_TYPE_AUDIO &&
        header.type != MESSAGE_TYPE_STRING &&
        header.type != MESSAGE_TYPE_BUFFER &&
        header.type != MESSAGE_TYPE_SPY_AUDIO_0 &&
        header.type != MESSAGE_TYPE_SPY_AUDIO_1) {
      std::cout << "Unknown Message Type: " << header.type << std::endl;
      mConnected = false;
      break;
    }
    if( header.size > 0 ) {
      char *buf = new char[header.size];
      if( !TCPRead( mSocket, (Uint8 *)buf, header.size )) {
        mConnected = false;
        delete[] buf;
        break;
      }
      Uint32 chk = checksum((Uint8*)buf, header.size);
      if( header.checksum != chk ) {
        std::cout << "checksum mismatch: " << std::hex << header.checksum << " != " << chk << std::endl;
        mConnected = false;
        delete[] buf;
        break;
      }

      std::string s( buf, header.size );
      delete[] buf;

      types.push_back(header.type);
      messages.push_back(s);
    }
  }
}

bool Client::sendMessage(std::string &msg) {
  return sendData( MESSAGE_TYPE_STRING, (void*)msg.data(), msg.size() );
}

bool Client::sendAudioData(Uint8 *data, int bytes) {
  return sendData( MESSAGE_TYPE_AUDIO, (void*)data, bytes );
}

bool Client::sendData(Uint8 type, const void *data, Uint32 bytes) {
  MessageHeader header;

  header.magic = MESSAGE_MAGIC;
  header.type = type;
  header.checksum = checksum((Uint8*)data, bytes);
  header.size = bytes;

  if( debug ) std::cout << ">>> " << std::hex << header.magic << " - " << header.type << " C " << header.checksum << " - " << std::dec << header.size << std::endl;

  if( !TCPSend( mSocket, &header, sizeof( header ))) {
    mConnected = false;
    return false;
  }
  if( !TCPSend( mSocket, data, bytes )) {
    mConnected = false;
    return false;
  }

  return true;
}
