const EventEmitter = require('events');
const net = require('net');

const headerSize = 4+4+4+4;
const jsonMessage = 0;
const audioMessage = 1;
const bufferMessage = 2;
const player0AudioSpyMessage = 3;
const player1AudioSpyMessage = 4;

function Network(port) {
    EventEmitter.call(this);
    this.server = net.createServer();
    this.server.on('connection', Network.prototype.onConnection.bind(this));
    this.server.listen(port);
}

Network.prototype = Object.create(EventEmitter.prototype);

Network.prototype.onConnection = function (socket) {
    var client = new Client(socket);
    this.emit('connection', client);
};

function Client(socket) {
    EventEmitter.call(this);
    this.joinTime = new Date().getTime();
    Client.prototype.resetSocket.call(this, socket);
}

Client.prototype = Object.create(EventEmitter.prototype);

Client.prototype.removeListeners = function () {
    console.log('Removing listeners for client from', this.socket.remoteAddress);
    this.socket.off('data', this.socketCallbacks.data);
    this.socket.off('error', this.socketCallbacks.error);
    this.socket.off('close', this.socketCallbacks.close);
}

Client.prototype.resetSocket = function (socket) {
    if (this.socket) {
        this.removeListeners();
        if (!this.socket.destroyed) this.socket.destroy();
    }
    this.socket = socket;
    this.lastPacketTime= new Date().getTime();
    this.buffer = Buffer.alloc(0);
    this.socketCallbacks = {data: this.onData.bind(this),
                            error: this.onError.bind(this),
                            close: this.onClose.bind(this)};

    this.socket.on('data', this.socketCallbacks.data);
    this.socket.on('error', this.socketCallbacks.error);
    this.socket.on('close', this.socketCallbacks.close);

    this.socket.setNoDelay(true);
}

Client.prototype.onClose = function () {
    this.socket = null;
    this.emit('close');
}

Client.prototype.onError = function (data) {
    // An error
    console.log('client error', data);
}

Client.prototype.onData = function (data) {
    // console.log('client data', data);
    this.buffer = Buffer.concat([this.buffer, data]);

    while (this.buffer.length >= headerSize) {
        if(this.buffer.readUInt32LE(0) === 0xDEADBEEF) {
            var type = this.buffer.readUInt32LE(4);
            var size = this.buffer.readUInt32LE(12);
            if (this.buffer.length >= headerSize + size) {
                if (type === jsonMessage) {
                    var text = this.buffer.toString('utf8', headerSize, headerSize+size);
                    // console.log('packet:', text);
                    var json = JSON.parse(text);
                    this.buffer = this.buffer.slice(headerSize+size);
                    this.emit('json', json);
                } else if (type === audioMessage) {
                    var audio = this.buffer.slice(headerSize, headerSize+size);
                    this.buffer = this.buffer.slice(headerSize+size);
                    this.emit('audio', audio);
                } else {
                    console.log(`unknown packet type: ${type}`);
                    this.buffer = Buffer.alloc(0);
                    this.socket.destroy();
                }

            } else {
                // console.log('needs remaining data', type, this.buffer.length, headerSize, size);
                return;
            }
        } else {
            console.error(`magic mismatch: ${this.buffer.readInt32LE(0)}`);
            this.buffer = Buffer.alloc(0);
            this.socket.destroy();
        }
    }
}

Client.prototype.checksum = function (buf) {
    var sum = 0;
    for (let i=0; i<buf.length; i++) {
        sum += buf.readUInt8(i);
    }
    return sum;
}

Client.prototype.sendAudio = function (audio) {
    if (!this.socket) return;
    var header = Buffer.allocUnsafe(headerSize);
    header.writeUInt32LE(0xDEADBEEF, 0);
    header.writeUInt32LE(audioMessage, 4);
    header.writeUInt32LE(this.checksum(audio), 8);
    header.writeUInt32LE(audio.length, 12);
    this.socket.write(header);
    this.socket.write(audio);
}

Client.prototype.sendSpyAudio = function (audio, player) {
    if (!this.socket) return;
    var header = Buffer.allocUnsafe(headerSize);
    header.writeUInt32LE(0xDEADBEEF, 0);
    header.writeUInt32LE(player0AudioSpyMessage+player, 4);
    header.writeUInt32LE(this.checksum(audio), 8);
    header.writeUInt32LE(audio.length, 12);
    this.socket.write(header);
    this.socket.write(audio);
}

Client.prototype.sendBuffer = function (buf) {
    if (!this.socket) return;
    var header = Buffer.allocUnsafe(headerSize);
    header.writeUInt32LE(0xDEADBEEF, 0);
    header.writeUInt32LE(bufferMessage, 4);
    header.writeUInt32LE(this.checksum(buf), 8);
    header.writeUInt32LE(buf.length, 12);
    this.socket.write(header);
    this.socket.write(buf);
}


Client.prototype.sendText = function (text) {
    if (!this.socket) return;
    var data = Buffer.from(text, 'utf8');
    var header = Buffer.allocUnsafe(headerSize);
    header.writeUInt32LE(0xDEADBEEF, 0);
    header.writeUInt32LE(jsonMessage, 4);
    header.writeUInt32LE(this.checksum(data), 8);
    header.writeUInt32LE(data.length, 12);
    this.socket.write(header);
    this.socket.write(data);
}

Client.prototype.sendJSON = function (json) {
    // if (!this.socket) return;
    this.sendText(JSON.stringify(json));
}

exports.Client = Client;
