#include "graphics.hh"

Question::Question( Window *w, const std::string &name, const std::string &query ) {
  mWindow = w;
  mName = name;
  mQuery = query;
  mFontSize = 18;
  mPad = 10;
  mSubmitDelay = 300;

  mQueryY = mQueryHeight = mQueryWidth = 0;
  mInputY = mInputHeight = mInputWidth = 0;
  mSubmitY = mSubmitHeight = mSubmitWidth = 0;

  mRectX = 0;
  mRectY = 0;
  mRectW = 0;
  mRectH = 0;

  mSubmitHelpText = "Press Enter to submit";
  mPrevHelpText = "Press Enter to go back to the previous question";
  mInputHelpText = "";

  mDone = QuestionInProgress;
  mCanGoToPrevious = true;
}

Question::~Question () {
}

void Question::clearEvents() {
    SDL_Event e;
    while ( SDL_PollEvent( &e ));
}

void Question::setCanGoToPrevious( bool flag ) {
  mCanGoToPrevious = flag;
}

std::string Question::getName() {
  return mName;
}

QuestionDoneCode Question::getDoneCode() {
  return mDone;
}

void Question::draw( bool helpText ) {
  mWindow->createContext();
  doDrawing( helpText );
  mWindow->destroyContext();
  SDL_RenderPresent( mWindow->mRenderer );
}

void Question::drawBoundingRect(double x, double y, double w, double h) {
  mRectX = x;
  mRectY = y;
  mRectW = w;
  mRectH = h;
  cairo_rectangle( mWindow->mCtx, mRectX, mRectY, mRectW, mRectH );
}

// void Question::clearWindow() {
//   mWindow->createContext();
//   cairo_set_source_rgb(mWindow->mCtx, 230/255.0, 240/255.0, 255/255.0);
//   cairo_paint(mWindow->mCtx);
//   mWindow->destroyContext();
//   SDL_RenderPresent( mWindow->mRenderer );
//   SDL_Delay( mSubmitDelay );
//   clearEvents();
// }

void Question::clearWindow() {
  draw( false );

  double border = 15, amount = 10;
  double x = mRectX, y = mRectY, w = mRectW, h = mRectH;

  while (w > 0 && h > 0) {
    mWindow->createContext();

    cairo_set_source_rgb(mWindow->mCtx, 230/255.0, 240/255.0, 255/255.0);
    // cairo_set_source_rgb(mWindow->mCtx, 1, 0, 0);
    cairo_rectangle( mWindow->mCtx, x - border, y-border, w + border*2, border);
    cairo_rectangle( mWindow->mCtx, x - border, y+h, w + border*2, border);
    cairo_rectangle( mWindow->mCtx, x-border, y - border, border, h+border*2);
    cairo_rectangle( mWindow->mCtx, x+w, y - border, border, h+border*2);
    cairo_fill( mWindow->mCtx );
    cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
    cairo_rectangle( mWindow->mCtx, x, y, w, h);
    cairo_stroke( mWindow->mCtx );

    mWindow->destroyContext();
    SDL_RenderPresent( mWindow->mRenderer );

    x += amount;
    y += amount;
    w -= 2*amount;
    h -= 2*amount;
  }
  // Clear the screen.
  mWindow->createContext();
  cairo_set_source_rgb(mWindow->mCtx, 230/255.0, 240/255.0, 255/255.0);
  cairo_paint(mWindow->mCtx);
  mWindow->destroyContext();
  SDL_RenderPresent( mWindow->mRenderer );
  SDL_Delay( 300 );
  clearEvents();
}

void Question::open() {
  // We need to do this once to populate the mRect* values.
  mWindow->createContext();
  doDrawing( false );
  mWindow->destroyContext();

  double border = 5, amount = 10;
  double x, y, w, h;

  if (mRectW > mRectH) {
    x = mRectX + mRectH/2;
    y = mRectY + mRectH/2;
    w = mRectW-mRectH;
    h = 1;
  } else {
    x = mRectX + mRectW/2;
    y = mRectY + mRectW/2;
    w = 1;
    h = mRectH-mRectW;
  }

  while( w < mRectW && h < mRectH ) {
    mWindow->createContext();

    doDrawing( false );

    cairo_set_source_rgb(mWindow->mCtx, 230/255.0, 240/255.0, 255/255.0);
    // cairo_set_source_rgb(mWindow->mCtx, 1, 0, 0);
    cairo_rectangle( mWindow->mCtx, mRectX, mRectY-border, mRectW, y-mRectY+border);
    cairo_rectangle( mWindow->mCtx, mRectX, y+h, mRectW, y-mRectY+border);
    cairo_rectangle( mWindow->mCtx, mRectX-border, mRectY-border, x-mRectX+border, mRectH+border*2);
    cairo_rectangle( mWindow->mCtx, x+w, mRectY-border, x-mRectX+border, mRectH+border*2);

    cairo_fill( mWindow->mCtx );
    cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
    cairo_rectangle( mWindow->mCtx, x, y, w, h);
    cairo_stroke( mWindow->mCtx );

    mWindow->destroyContext();
    SDL_RenderPresent( mWindow->mRenderer );
    clearEvents();
    // SDL_Delay(50);

    x -= amount;
    y -= amount;
    w += 2*amount;
    h += 2*amount;
  }
  draw(true);
}

void Question::updateLayout() {
  updateQueryLayout();
  updateInputLayout();
  updateSubmitLayout();
  updateBoxLayout();
}

void Question::updateQueryLayout() {
  cairo_font_extents_t extents;
  cairo_font_extents( mWindow->mCtx, &extents );

  std::ostringstream line;
  mQueryWidth = 0;
  mQueryHeight = 0;
  for( size_t i=0; i<mQuery.size(); i++ ) {
    bool add = true;
    if (mQuery[i] == ' ') {
      double w, h;
      textSize( mWindow, line.str(), mFontSize, w, h );
      if( w > maxQuestionQueryWidth ) {
        mQueryHeight += extents.height; // h
        if (w > mQueryWidth) mQueryWidth = w;
        line.str("");
        add = false;
      }
    }
    if (add) line << mQuery[i];
  }

  if( line.str().size() > 0 ) {
    double w, h;
    textSize( mWindow, line.str(), mFontSize, w, h );
    if( w > mQueryWidth ) mQueryWidth = w;
    mQueryHeight += extents.height;
  }
  mQueryWidth += mPad * 2;
  // mQueryHeight = extents.height;
}

void Question::updateSubmitLayout() {
  cairo_font_extents_t extents;
  cairo_font_extents( mWindow->mCtx, &extents );

  std::string done = "Submit";
  std::string back = "Previous Question";

  textSize( mWindow, back, mFontSize, mSubmitWidth, mSubmitHeight );
  mSubmitWidth += mPad * 2;
  mSubmitHeight = extents.height * 2 + mPad;
}


void Question::updateBoxLayout() {
  cairo_font_extents_t extents;
  cairo_font_extents( mWindow->mCtx, &extents );

  mRectW = std::max(std::max(std::max(400.0, mQueryWidth), mInputWidth), mSubmitWidth);
  mRectH = mQueryHeight + mPad*2 + mInputHeight + mPad*4 + mSubmitHeight + mPad*2;
  mRectX = mWindow->getWidth()/2 - mRectW/2;
  mRectY = mWindow->getHeight()/2 - mRectH/2;

  // std::cout << "heights " << extents.height << " " << extents.ascent << " " << mPad << " " << mQueryHeight << " " << mInputHeight << " "  << mSubmitHeight << "\n";

  mQueryY = mRectY + extents.ascent + mPad;
  mInputY = mRectY + mQueryHeight + mPad*4 + extents.ascent;
  mSubmitY = mRectY + mQueryHeight + mPad*2 + mInputHeight + mPad*4 + mPad + extents.ascent;

  // std::cout << "Ys " << mRectY << " " << mQueryY << " " << mInputY << " "  << mSubmitY << "\n";
}

void Question::drawCommon( bool helpText ) {
  cairo_font_extents_t extents;
  cairo_font_extents( mWindow->mCtx, &extents );

  cairo_set_source_rgb(mWindow->mCtx, 230/255.0, 240/255.0, 255/255.0);
  cairo_paint(mWindow->mCtx);

  cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
  cairo_rectangle( mWindow->mCtx, mRectX, mRectY, mRectW, mRectH );
  cairo_stroke( mWindow->mCtx );


  cairo_move_to( mWindow->mCtx, mRectX, mInputY - extents.ascent - mPad*2);
  cairo_line_to( mWindow->mCtx, mRectX+mRectW, mInputY - extents.ascent - mPad*2);
  cairo_move_to( mWindow->mCtx, mRectX, mSubmitY - extents.ascent - mPad);
  cairo_line_to( mWindow->mCtx, mRectX+mRectW, mSubmitY - extents.ascent - mPad);

  cairo_move_to( mWindow->mCtx, mRectX, mSubmitY + extents.height + mPad - extents.ascent - mPad/2);
  cairo_line_to( mWindow->mCtx, mRectX+mRectW, mSubmitY + extents.height + mPad - extents.ascent - mPad/2 );

  cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
  cairo_stroke( mWindow->mCtx );

  drawWrappedText( mWindow->mCtx, mQuery, mWindow->getWidth()/2, mQueryY, maxQuestionQueryWidth, 0, -1 );

  cairo_set_source_rgb( mWindow->mCtx, 0.75, 0.75, 0.75 );
  if( isSubmitHighlighted() ) {
    cairo_rectangle( mWindow->mCtx,
                     mRectX + mPad, mSubmitY - extents.ascent - mPad/4,
                     mRectW - mPad*2, extents.height );
    cairo_fill( mWindow->mCtx );
  } else if( isPrevHighlighted() ) {
    cairo_rectangle( mWindow->mCtx,
                     mRectX + mPad, mSubmitY + extents.height + mPad + mPad/4 - extents.ascent,
                     mRectW - mPad*2, extents.height );
    cairo_fill( mWindow->mCtx );
  }
  cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
  drawText( mWindow->mCtx, "Submit", mWindow->getWidth()/2, mSubmitY-mPad/4, 0, -1 );
  if ( mCanGoToPrevious )
    drawText( mWindow->mCtx, "Previous Question", mWindow->getWidth()/2, mSubmitY+mPad/4+mPad+extents.height, 0, -1 );

  if( helpText ) {
    cairo_set_source_rgb( mWindow->mCtx, 0.6, 0.6, 0.6 );
    std::string *txt;
    if( isSubmitHighlighted() ) {
      txt = &mSubmitHelpText;
    } else if( isPrevHighlighted() ) {
      txt = &mPrevHelpText;
    } else {
      txt = &mInputHelpText;
    }
    drawText( mWindow->mCtx, *txt, mWindow->getWidth()/2, mRectY+mRectH + extents.ascent + mPad , 0, -1 );
  }


  // std::cout << "Bang!\n";
}

void Question::doDrawing( bool helpText ) {
  cairo_set_font_face( mWindow->mCtx, gCairoFont );
  cairo_set_font_size( mWindow->mCtx, mFontSize );

  updateLayout();
  drawCommon( helpText );
  drawInput();
}

void Question::drawSubmit() {
  cairo_font_extents_t extents;
  cairo_font_extents( mWindow->mCtx, &extents );

  std::string done = "Submit";
  std::string back = "Previous Question";
  double w,h;
  textSize( mWindow, done, mFontSize, w, h );
  w += mPad * 2;

  if( isSubmitHighlighted() ) {
    cairo_rectangle( mWindow->mCtx,
                     mWindow->getWidth()/2 - mRectW/2 + mPad,
                     mRectY + mPad,
                     mRectW - mPad*2, extents.height );
    cairo_set_source_rgb( mWindow->mCtx, 0.75, 0.75, 0.75 );
    cairo_fill( mWindow->mCtx );
  }
  cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
  drawText( mWindow->mCtx, done, mWindow->getWidth()/2, mSubmitY, 0, -1 );

  if( isPrevHighlighted() ) {
    cairo_rectangle( mWindow->mCtx,
                     mWindow->getWidth()/2 - mRectW/2 + mPad,
                     mSubmitY + extents.height + mPad*2,
                     mRectW - mPad*2, extents.height );
    cairo_set_source_rgb( mWindow->mCtx, 0.75, 0.75, 0.75 );
    cairo_fill( mWindow->mCtx );
  }
  cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
  drawText( mWindow->mCtx, back, mWindow->getWidth()/2, mSubmitY+extents.height+mPad, 0, -1 );
}

RadioQuestion::RadioQuestion(Window *w, const std::string &name, const std::string &query, std::vector<std::string> options, std::vector<std::string> strings) :
  Question( w, name, query ) {
  mOptions = options;
  mStrings = strings;
  mCurrent = 0;
  mSelected = 0;
  mSelectedValid = false;
  mRadioRadius = 10;
  mInputHelpText = "Press Enter to select";
}

RadioQuestion::~RadioQuestion() {
}

void RadioQuestion::select( unsigned int idx ) {
  if( idx >= 0 && idx < mStrings.size() ) {
    mSelected = idx;
    mSelectedValid = true;
  }
}

std::string RadioQuestion::getAnswer() {
  if( mSelectedValid )
    return mOptions[mSelected];
  else
    return "";
}

bool RadioQuestion::isSubmitHighlighted() {
  return mCurrent == mOptions.size();
}

bool RadioQuestion::isPrevHighlighted() {
  return mCurrent == mOptions.size()+1;
}

void RadioQuestion::drawRadio(double x, double y, bool state) {
  cairo_new_path( mWindow->mCtx );
  cairo_arc( mWindow->mCtx, x+mRadioRadius, y+mRadioRadius, mRadioRadius, 0., M_PI*2 );
  cairo_set_line_width( mWindow->mCtx, 2 );
  cairo_set_source_rgb( mWindow->mCtx, 0, 0, 0 );
  cairo_stroke( mWindow->mCtx );

  if( state ) {
    cairo_arc( mWindow->mCtx, x+mRadioRadius, y+mRadioRadius, mRadioRadius-4, 0., M_PI*2 );
    cairo_fill( mWindow->mCtx );
  }
  cairo_set_line_width( mWindow->mCtx, 1 );
}

void RadioQuestion::start() {
  open();
  clearEvents();
}

void RadioQuestion::finish(bool force) {
  if( !force ) clearWindow();
}

bool RadioQuestion::handleEvent(const SDL_Event &e) {
  bool update = false;
  bool handled = false;
  switch( e.type ) {
  case SDL_WINDOWEVENT:
    if( e.window.event == SDL_WINDOWEVENT_EXPOSED ) {
      update = true;
      handled = true;
    }
    break;
  case SDL_KEYDOWN:
    if( e.key.keysym.sym == SDLK_DOWN ) {
      unsigned int max = mCanGoToPrevious ? mStrings.size()+1 : mStrings.size();
      mCurrent += 1;
      if( mCurrent >= max ) mCurrent = max;
      handled = true;
      update = true;
    } else if( e.key.keysym.sym == SDLK_UP ) {
      if( mCurrent > 0 ) mCurrent -= 1;
      handled = true;
      update = true;
    } else if( e.key.keysym.sym == SDLK_RETURN || e.key.keysym.sym == SDLK_SPACE ) {
      if( isSubmitHighlighted() ) {
        if( mSelectedValid  && e.key.keysym.sym == SDLK_RETURN) {
          mDone = QuestionSubmitted;
          handled = true;
        }
      } else if( isPrevHighlighted() ) {
        if( e.key.keysym.sym == SDLK_RETURN ) {
          if( mCanGoToPrevious ) mDone = QuestionGoToPrevious;
          handled = true;
        }
      } else {
        mSelected = mCurrent;
        mSelectedValid = true;
        if( e.key.keysym.sym == SDLK_RETURN )
          mCurrent = mStrings.size();
        update = true;
        handled = true;
      }
    }
  }

  if( update ) draw( true );

  return handled;
}

// void RadioQuestion::doDrawing( bool helpText ) {
//   cairo_set_source_rgb(mWindow->mCtx, 230/255.0, 240/255.0, 255/255.0);
//   cairo_paint(mWindow->mCtx);

//   cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
//   cairo_set_font_face( mWindow->mCtx, gCairoFont );
//   cairo_set_font_size( mWindow->mCtx, mFontSize );

//   cairo_font_extents_t extents;
//   cairo_font_extents( mWindow->mCtx, &extents );

//   double qw, qh;
//   textSize( mWindow, mQuery, mFontSize, qw, qh );

//   drawText( mWindow->mCtx, mQuery, mWindow->getWidth()/2, mWindow->getHeight()/2 - 100, 0, -1 );

//   double totalh = mStrings.size() * (extents.height + mPad);
//   double maxx = 0;
//   for (size_t i=0; i<mStrings.size(); i++) {
//     double w,h;
//     textSize( mWindow, mStrings[i], mFontSize, w, h );
//     if (w > maxx) maxx = w;
//   }
//   maxx += mPad * 2 + 20;
//   double boxWidth = std::max(maxx, std::max(400.0, qw+mPad*2));

//   double y = mWindow->getHeight()/2 - 30;

//   drawBoundingRect( mWindow->getWidth()/2 - boxWidth/2 - mPad,
//                     mWindow->getHeight()/2 - 100 - extents.ascent - mPad,
//                     boxWidth + mPad*2,
//                     totalh + 100 + mPad * 2 );
//   cairo_move_to( mWindow->mCtx,
//                  mWindow->getWidth()/2 - boxWidth/2 - mPad,
//                  mWindow->getHeight()/2 - 100 - extents.ascent + extents.height + mPad );
//   cairo_line_to( mWindow->mCtx,
//                  mWindow->getWidth()/2 + boxWidth/2 + mPad,
//                  mWindow->getHeight()/2 - 100 - extents.ascent + extents.height + mPad );
//   cairo_move_to( mWindow->mCtx,
//                  mWindow->getWidth()/2 - boxWidth/2 - mPad,
//                  mWindow->getHeight()/2 - 60 + totalh );
//   cairo_line_to( mWindow->mCtx,
//                  mWindow->getWidth()/2 + boxWidth/2 + mPad,
//                  mWindow->getHeight()/2 - 60 + totalh );
//   cairo_stroke( mWindow->mCtx );

//   if( helpText ) {
//     cairo_set_source_rgb( mWindow->mCtx, 0.6, 0.6, 0.6 );
//     drawText( mWindow->mCtx, "Press Enter to submit", mWindow->getWidth()/2, mWindow->getHeight()/2 + totalh + 60 + mPad * 2 , 0, -1 );
//   }

//   for (size_t i=0; i<mStrings.size(); i++) {
//     if( i == mCurrent ) {
//       cairo_rectangle( mWindow->mCtx,
//                        mWindow->getWidth()/2 - boxWidth/2,
//                        y - extents.ascent,
//                        boxWidth, extents.height );
//       cairo_set_source_rgb( mWindow->mCtx, 0.6, 0.6, 0.6 );
//       cairo_fill( mWindow->mCtx );
//     }
//     drawRadio( mWindow->getWidth()/2 - maxx/2 + mPad,
//                y - extents.ascent+ extents.height/2-10,
//                mSelectedValid && i == mSelected );

//     cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
//     drawText( mWindow->mCtx, mStrings[i], mWindow->getWidth()/2-maxx/2+mPad+20+mPad, y, 1, -1 );

//     y += extents.height + mPad;
//   }

//   y += 20;

//   drawSubmit( y, boxWidth, mCurrent == mOptions.size(), mCurrent == mOptions.size()+1 );
// }

void RadioQuestion::updateInputLayout() {
  cairo_font_extents_t extents;
  cairo_font_extents( mWindow->mCtx, &extents );

  mInputWidth = 0;
  for (size_t i=0; i<mStrings.size(); i++) {
    double w,h;
    textSize( mWindow, mStrings[i], mFontSize, w, h );
    if (w > mInputWidth) mInputWidth = w;
  }
  mInputWidth += mPad * 5 + mRadioRadius*2;
  mInputHeight = mStrings.size() * (extents.height + mPad) - mPad;
}

void RadioQuestion::drawInput() {
  double y = mInputY;
  cairo_font_extents_t extents;
  cairo_font_extents( mWindow->mCtx, &extents );

  for (size_t i=0; i<mStrings.size(); i++) {
    if( i == mCurrent ) {
      cairo_rectangle( mWindow->mCtx,
                       mWindow->getWidth()/2 - mRectW/2 + mPad,
                       y - extents.ascent,
                       mRectW - mPad*2, extents.height );
      cairo_set_source_rgb( mWindow->mCtx, 0.75, 0.75, 0.75 );
      cairo_fill( mWindow->mCtx );
    }
    drawRadio( mWindow->getWidth()/2 - mRectW/2 + mPad*2,
               y - extents.ascent+ extents.height/2-10,
               mSelectedValid && i == mSelected );

    cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
    drawText( mWindow->mCtx, mStrings[i], mWindow->getWidth()/2-mRectW/2+mPad*2+mRadioRadius*2+mPad, y, 1, -1 );

    y += extents.height + mPad;
  }
}


CheckboxQuestion::CheckboxQuestion(Window *w, const std::string &name, const std::string &query, std::vector<std::string> options, std::vector<std::string> strings) :
  Question( w, name, query ) {
  mOptions = options;
  mStrings = strings;
  mCurrent = 0;
  mCheckboxSize = 20;
  for (size_t i=0; i<options.size(); i++)
    mSelected.push_back(false);
  mInputHelpText = "Press Enter or Space to check";
}

CheckboxQuestion::~CheckboxQuestion() {

}

void CheckboxQuestion::check( unsigned int idx, bool flag ) {
  if( idx < mStrings.size() ) {
    mSelected[idx] = flag;
  }
}

std::string CheckboxQuestion::getAnswer() {
  std::ostringstream ret;
  bool first = true;
  for( size_t i=0; i<mSelected.size(); i++ ) {
    if( mSelected[i] ) {
      if( first ) first = false;
      else ret << ",";
      ret << mOptions[i];
    }
  }
  return ret.str();
}

bool CheckboxQuestion::isSubmitHighlighted() {
  return mCurrent == mOptions.size();
}

bool CheckboxQuestion::isPrevHighlighted() {
  return mCurrent == mOptions.size()+1;
}

void CheckboxQuestion::drawCheckbox(double x, double y, bool state) {
  cairo_rectangle( mWindow->mCtx, x, y, mCheckboxSize, mCheckboxSize );
  cairo_set_line_width( mWindow->mCtx, 1 );
  cairo_set_source_rgb( mWindow->mCtx, 0, 0, 0 );
  cairo_stroke( mWindow->mCtx );

  if( state ) {
    cairo_rectangle( mWindow->mCtx, x+3, y+3, 14, 14 );
    cairo_clip( mWindow->mCtx );
    cairo_move_to( mWindow->mCtx, x, y );
    cairo_line_to( mWindow->mCtx, x+mCheckboxSize, y+mCheckboxSize );
    cairo_move_to( mWindow->mCtx, x+mCheckboxSize, y );
    cairo_line_to( mWindow->mCtx, x, y+mCheckboxSize );
    cairo_set_line_width( mWindow->mCtx, 4 );
    cairo_stroke( mWindow->mCtx );
    cairo_set_line_width( mWindow->mCtx, 1 );
    cairo_reset_clip( mWindow->mCtx );
  }
}

void CheckboxQuestion::updateInputLayout() {
  cairo_font_extents_t extents;
  cairo_font_extents( mWindow->mCtx, &extents );

  mInputWidth = 0;
  for (size_t i=0; i<mStrings.size(); i++) {
    double w,h;
    textSize( mWindow, mStrings[i], mFontSize, w, h );
    if (w > mInputWidth) mInputWidth = w;
  }
  mInputWidth += mPad * 5 + mCheckboxSize;
  mInputHeight = mStrings.size() * (extents.height + mPad) - mPad;
}

void CheckboxQuestion::drawInput() {
  double y = mInputY;
  cairo_font_extents_t extents;
  cairo_font_extents( mWindow->mCtx, &extents );

  for (size_t i=0; i<mStrings.size(); i++) {
    if( i == mCurrent ) {
      cairo_rectangle( mWindow->mCtx,
                       mWindow->getWidth()/2 - mRectW/2 + mPad,
                       y - extents.ascent,
                       mRectW - mPad*2, extents.height );
      cairo_set_source_rgb( mWindow->mCtx, 0.75, 0.75, 0.75 );
      cairo_fill( mWindow->mCtx );
    }
    drawCheckbox( mWindow->getWidth()/2 - mRectW/2 + mPad*2,
                  y - extents.ascent+ extents.height/2-10,
                  mSelected[i] );

    cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
    drawText( mWindow->mCtx, mStrings[i], mWindow->getWidth()/2-mRectW/2+mPad*2+mCheckboxSize+mPad, y, 1, -1 );

    y += extents.height + mPad;
  }
}

// void CheckboxQuestion::doDrawing( bool helpText ) {
//   cairo_set_source_rgb(mWindow->mCtx, 230/255.0, 240/255.0, 255/255.0);
//   cairo_paint(mWindow->mCtx);

//   cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
//   cairo_set_font_face( mWindow->mCtx, gCairoFont );
//   cairo_set_font_size( mWindow->mCtx, mFontSize );

//   cairo_font_extents_t extents;
//   cairo_font_extents( mWindow->mCtx, &extents );

//   double qw, qh;
//   textSize( mWindow, mQuery, mFontSize, qw, qh );

//   drawText( mWindow->mCtx, mQuery, mWindow->getWidth()/2, mWindow->getHeight()/2 - 100, 0, -1 );

//   double totalh = mStrings.size() * (extents.height + mPad);
//   double maxx = 0;
//   for (size_t i=0; i<mStrings.size(); i++) {
//     double w,h;
//     textSize( mWindow, mStrings[i], mFontSize, w, h );
//     if (w > maxx) maxx = w;
//   }
//   maxx += mPad * 3 + 20;
//   double boxWidth = std::max(maxx, std::max(400.0, qw+mPad*2));

//   double y = mWindow->getHeight()/2 - 30;

//   drawBoundingRect( mWindow->getWidth()/2 - boxWidth/2 - mPad,
//                     mWindow->getHeight()/2 - 100 - extents.ascent - mPad,
//                     boxWidth + mPad*2,
//                     totalh + 100 + mPad * 2 );
//   cairo_move_to( mWindow->mCtx,
//                  mWindow->getWidth()/2 - boxWidth/2 - mPad,
//                  mWindow->getHeight()/2 - 100 - extents.ascent + extents.height + mPad );
//   cairo_line_to( mWindow->mCtx,
//                  mWindow->getWidth()/2 + boxWidth/2 + mPad,
//                  mWindow->getHeight()/2 - 100 - extents.ascent + extents.height + mPad );
//   cairo_move_to( mWindow->mCtx,
//                  mWindow->getWidth()/2 - boxWidth/2 - mPad,
//                  mWindow->getHeight()/2 - 60 + totalh );
//   cairo_line_to( mWindow->mCtx,
//                  mWindow->getWidth()/2 + boxWidth/2 + mPad,
//                  mWindow->getHeight()/2 - 60 + totalh );
//   cairo_stroke( mWindow->mCtx );


//   for (size_t i=0; i<mStrings.size(); i++) {
//     if( i == mCurrent ) {
//       cairo_rectangle( mWindow->mCtx,
//                        mWindow->getWidth()/2 - boxWidth/2,
//                        y - extents.ascent,
//                        boxWidth, extents.height );
//       cairo_set_source_rgb( mWindow->mCtx, 0.6, 0.6, 0.6);
//       cairo_fill( mWindow->mCtx );
//     }
//     drawCheckbox( mWindow->getWidth()/2 - maxx/2 + mPad,
//                   y - extents.ascent+ extents.height/2-10,
//                   mSelected[i] );

//     cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
//     drawText( mWindow->mCtx, mStrings[i], mWindow->getWidth()/2-maxx/2+mPad+20+mPad, y, 1, -1 );

//     y += extents.height + mPad;
//   }

//   y += 20;

//   std::string done = "Done";
//   double w,h;
//   textSize( mWindow, done, mFontSize, w, h );
//   w += mPad * 2;

//   if( mCurrent == mOptions.size() ) {
//     cairo_rectangle( mWindow->mCtx,
//                      mWindow->getWidth()/2 - boxWidth/2,
//                      y - extents.ascent,
//                      boxWidth, extents.height );
//     cairo_set_source_rgb( mWindow->mCtx, 0.6, 0.6, 0.6 );
//     cairo_fill( mWindow->mCtx );
//   }
//   cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
//   drawText( mWindow->mCtx, "Done", mWindow->getWidth()/2, y, 0, -1 );
// }

void CheckboxQuestion::start() {
  open();
  clearEvents();
}

void CheckboxQuestion::finish(bool force) {
  if( !force ) clearWindow();
}

bool CheckboxQuestion::handleEvent(const SDL_Event &e) {
  bool handled = false;
  bool update = false;

  switch( e.type ) {
  case SDL_WINDOWEVENT:
    if( e.window.event == SDL_WINDOWEVENT_EXPOSED ) {
      update = true;
      handled = true;
    }
    break;
  case SDL_KEYDOWN:
    if( e.key.keysym.sym == SDLK_DOWN ) {
      unsigned int max = mCanGoToPrevious ? mStrings.size()+1 : mStrings.size();
      mCurrent += 1;
      if( mCurrent > max ) mCurrent = max;
      handled = true;
      update = true;
    } else if( e.key.keysym.sym == SDLK_UP ) {
      handled = true;
      update = true;
      if( mCurrent > 0 ) mCurrent -= 1;
    } else if( e.key.keysym.sym == SDLK_SPACE ) {
      if( mCurrent < mStrings.size() ) {
        mSelected[mCurrent] = !mSelected[mCurrent];
        update = true;
        handled = true;
      }
    } else if( e.key.keysym.sym == SDLK_RETURN ) {
      if( isSubmitHighlighted() ) {
        mDone = QuestionSubmitted;
        handled = true;
      } else if( isPrevHighlighted() ) {
        if( mCanGoToPrevious ) mDone = QuestionGoToPrevious;
        handled = true;
      } else {
        mSelected[mCurrent] = !mSelected[mCurrent];
        update = true;
        handled = true;
      }
    }
  }

  if( update ) draw(true);
  return handled;
}

// std::string CheckboxQuestion::run() {
//   SDL_Event e;
//   bool done = false;

//   open();

//   clearEvents();
//   while( !done ) {
//     bool update = false;
//     if ( SDL_WaitEvent( &e ) == 0 ) {
//       printf( "SDL_WaitEvent SDL_Error: %s\n", SDL_GetError() );
//       shutdown();
//     }

//     switch( e.type ) {
//     case SDL_WINDOWEVENT:
//       if( e.window.event == SDL_WINDOWEVENT_EXPOSED ) {
//         update = true;
//       }
//       break;
//     case SDL_QUIT:
//       std::cout << "FIXME: Quit!\n";
//       break;
//     case SDL_KEYDOWN:
//       if( e.key.keysym.sym == SDLK_DOWN ) {
//         update = true;
//         mCurrent += 1;
//         if( mCurrent > mStrings.size()+1 ) mCurrent = mStrings.size()+1;
//       } else if( e.key.keysym.sym == SDLK_UP ) {
//         update = true;
//         if( mCurrent > 0 ) mCurrent -= 1;
//       } else if( e.key.keysym.sym == SDLK_SPACE ) {
//         if( mCurrent < mStrings.size() ) {
//           mSelected[mCurrent] = !mSelected[mCurrent];
//           update = true;
//         }
//       } else if( e.key.keysym.sym == SDLK_RETURN ) {
//         if( mCurrent == mStrings.size() ) {
//           done = true;
//         } else {
//           mSelected[mCurrent] = !mSelected[mCurrent];
//           update = true;
//         }
//       }
//     }

//     if( update ) draw(true);
//   }

//   std::ostringstream ret;
//   bool first = true;
//   for( size_t i=0; i<mSelected.size(); i++ ) {
//     if( mSelected[i] ) {
//       if( first ) first = false;
//       else ret << ",";
//       ret << mOptions[i];
//     }
//   }

//   clearWindow();

//   return ret.str();
// }

OneLineQuestion::OneLineQuestion(Window *w, const std::string &name, const std::string &query)
  : Question( w, name, query ) {
  mAnswer = "";
  mHighlight = HighlightInput;
  mInputHelpText = "Press Enter when done";
}

void OneLineQuestion::setAnswer( const std::string &answer ) {
  mAnswer = answer;
}

OneLineQuestion::~OneLineQuestion() {
  SDL_StopTextInput();
}

bool OneLineQuestion::isSubmitHighlighted() {
  return mHighlight == HighlightSubmit;
}

bool OneLineQuestion::isPrevHighlighted() {
  return mHighlight == HighlightPrev;
}

void OneLineQuestion::updateInputLayout() {
  cairo_font_extents_t extents;
  cairo_font_extents( mWindow->mCtx, &extents );
  textSize( mWindow, mAnswer, mFontSize, mInputWidth, mInputHeight );

  mInputWidth += mPad*4;

  mInputWidth = std::max(mInputWidth,200.0);
  mInputHeight = extents.height;
}


void OneLineQuestion::drawInput() {
  cairo_font_extents_t extents;
  cairo_font_extents( mWindow->mCtx, &extents );

  // double w, h;
  // textSize( mWindow, mAnswer, mFontSize, w, h );
  // double mx = std::max(w+mPad*2,200.0);

  cairo_rectangle( mWindow->mCtx,
                   mWindow->getWidth()/2 - mRectW/2 + mPad,
                   mInputY - extents.ascent,
                   mRectW - mPad*2, extents.height );
  cairo_set_source_rgb( mWindow->mCtx, 0.75, 0.75, 0.75 );
  if( mHighlight == HighlightInput ) {
    cairo_fill( mWindow->mCtx );
    double bw,bh;
    // std::string before;
    // before = mAnswer[i].substr(0, mCursorX);
    textSize( mWindow, mAnswer, mFontSize, bw, bh );

    cairo_move_to( mWindow->mCtx,
                   mWindow->getWidth()/2 + bw/2 + 1,
                   mInputY - extents.ascent);
    cairo_line_to( mWindow->mCtx,
                   mWindow->getWidth()/2 + bw/2 + 1,
                   mInputY + extents.descent );
    cairo_set_source_rgb( mWindow->mCtx, 0, 0, 0 );
    cairo_stroke( mWindow->mCtx );
  } else {
    cairo_stroke( mWindow->mCtx );
  }

  cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
  drawText( mWindow->mCtx, mAnswer, mWindow->getWidth()/2, mInputY, 0, -1 );
}

// void OneLineQuestion::doDrawing( bool helpText ) {
//   cairo_set_source_rgb(mWindow->mCtx, 230/255.0, 240/255.0, 255/255.0);
//   cairo_paint(mWindow->mCtx);

//   cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
//   cairo_set_font_face( mWindow->mCtx, gCairoFont );
//   cairo_set_font_size( mWindow->mCtx, mFontSize );

//   cairo_font_extents_t extents;
//   cairo_font_extents( mWindow->mCtx, &extents );

//   double qw, qh;
//   textSize( mWindow, mQuery, mFontSize, qw, qh );

//   drawText( mWindow->mCtx, mQuery, mWindow->getWidth()/2, mWindow->getHeight()/2 - 100, 0, -1 );
//   double boxWidth = std::max(400.0, qw + mPad*2);

//   drawBoundingRect( mWindow->getWidth()/2 - boxWidth/2 - mPad,
//                     mWindow->getHeight()/2 - 100 - extents.ascent - mPad,
//                     boxWidth + mPad*2,
//                     100 );
//   cairo_move_to( mWindow->mCtx,
//                  mWindow->getWidth()/2 - boxWidth/2 - mPad,
//                  mWindow->getHeight()/2 - 100 - extents.ascent + extents.height + mPad );
//   cairo_line_to( mWindow->mCtx,
//                  mWindow->getWidth()/2 + boxWidth/2 + mPad,
//                  mWindow->getHeight()/2 - 100 - extents.ascent + extents.height + mPad );
//   cairo_stroke( mWindow->mCtx );

//   if( helpText ) {
//     cairo_set_source_rgb( mWindow->mCtx, 0.6, 0.6, 0.6 );
//     drawText( mWindow->mCtx, "Press Enter to submit", mWindow->getWidth()/2, mWindow->getHeight()/2 + 10 , 0, -1 );
//   }


//   double w, h;
//   textSize( mWindow, mAnswer, mFontSize, w, h );
//   double mx = std::max(w,200.0);

//   cairo_rectangle( mWindow->mCtx,
//                    mWindow->getWidth()/2 - mx/2,
//                    mWindow->getHeight()/2 - 50 - extents.ascent,
//                    mx, extents.height );
//   cairo_set_source_rgb( mWindow->mCtx, 0.75, 0.75, 0.75 );
//   cairo_fill( mWindow->mCtx );

//   cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
//   drawText( mWindow->mCtx, mAnswer, mWindow->getWidth()/2, mWindow->getHeight()/2-50, 0, -1 );

//   cairo_move_to( mWindow->mCtx,
//                  mWindow->getWidth()/2 + w/2 + 1,
//                  mWindow->getHeight()/2 - 50 - extents.ascent);
//   cairo_line_to( mWindow->mCtx,
//                  mWindow->getWidth()/2 + w/2 + 1,
//                  mWindow->getHeight()/2 - 50 + extents.descent );
//   cairo_stroke( mWindow->mCtx );
// }

void OneLineQuestion::start() {
  SDL_StartTextInput();
  open();
  clearEvents();
}

void OneLineQuestion::finish(bool force) {
  SDL_StopTextInput();
  if (!force) clearWindow();
}


std::string OneLineQuestion::getAnswer() {
  return mAnswer;
}

bool OneLineQuestion::handleEvent(const SDL_Event &e) {
  bool handled = false;
  bool update = false;

  switch( e.type ) {
  case SDL_WINDOWEVENT:
    if( e.window.event == SDL_WINDOWEVENT_EXPOSED ) {
      update = true;
    }
    break;
  case SDL_TEXTINPUT:
    if( mHighlight == HighlightInput ) {
      std::string msg( e.text.text );
      mAnswer.append( msg );
      handled = true;
    }
    update = true;
    break;
  case SDL_KEYDOWN:
    if( mHighlight == HighlightInput ) {
      if( e.key.keysym.sym == SDLK_RETURN ) {
        if( mAnswer.size() > 0 ) {
          mHighlight = HighlightSubmit;
          update = true;
        }
        handled = true;
      } else if ( e.key.keysym.sym == SDLK_BACKSPACE ) {
        if (mAnswer.size() > 0)
          mAnswer.erase(mAnswer.end()-1);
        handled = true;
        update = true;
      } else if( e.key.keysym.sym == SDLK_DOWN ) {
        mHighlight = HighlightSubmit;
        handled = true;
        update = true;
      }
      handled = true;
    } else if( mHighlight == HighlightSubmit ) {
      if( e.key.keysym.sym == SDLK_RETURN ) {
        if( mAnswer.size() > 0 ) mDone = QuestionSubmitted;
        handled = true;
      } else if( e.key.keysym.sym == SDLK_UP ) {
        mHighlight = HighlightInput;
        handled = true;
        update = true;
      } else if( e.key.keysym.sym == SDLK_DOWN ) {
        if( mCanGoToPrevious ) mHighlight = HighlightPrev;
        handled = true;
        update = true;
      }
    } else {
      // Prev
      if( e.key.keysym.sym == SDLK_RETURN ) {
        if( mCanGoToPrevious ) mDone = QuestionGoToPrevious;
        handled = true;
      } else if( e.key.keysym.sym == SDLK_UP ) {
        mHighlight = HighlightSubmit;
        update = true;
      }
    }
  }
  if( update ) draw(true);


  // SDL_StopTextInput();
  // clearWindow();

  return handled;
}

MultiLineQuestion::MultiLineQuestion(Window *w, const std::string &name, const std::string &query)
  : Question( w, name, query ) {
  mAnswer.push_back("");
  mCursorX = 0;
  mCursorY = 0;
  mMaxWidth = 550;
  mHighlight = HighlightInput;
  mInputHelpText = "Press Enter when done";
}

MultiLineQuestion::~MultiLineQuestion() {
}

void MultiLineQuestion::setAnswer( const std::string &answer ) {
  mAnswer.clear();
  mAnswer.push_back( answer );
  mCursorX = mAnswer[0].size();
  mCursorY = 0;
}


std::string MultiLineQuestion::getAnswer() {
  std::ostringstream ret;
  for( size_t i=0; i<mAnswer.size(); i++ ) {
    ret << mAnswer[i];
    // if( i < mAnswer.size()-1 ) ret << "\n";
  }

  return ret.str();
}

bool MultiLineQuestion::isSubmitHighlighted() {
  return mHighlight == HighlightSubmit;
}

bool MultiLineQuestion::isPrevHighlighted() {
  return mHighlight == HighlightPrev;
}

void MultiLineQuestion::wrapLines() {
  for (size_t i=0; i<mAnswer.size(); i++) {
    double w, h;
    textSize( mWindow, mAnswer[i], mFontSize, w, h );
    if( w > mMaxWidth ) {
      long lastSpace = -1;
      for (size_t j=0; j<mAnswer[i].size(); j++) {
        textSize( mWindow, mAnswer[i].substr(0,j), mFontSize, w, h );
        if( mAnswer[i][j] == ' ' ) {
          // Ignore a space at the very end.
          if (j < mAnswer[i].size()-1) {
            lastSpace = j;
            if( w > mMaxWidth ) break;
          }
        }
      }

      if( lastSpace >= 0 && lastSpace < mAnswer[i].size()-1 ) {
        std::string extra = mAnswer[i].substr(lastSpace+1);
        if( i >= mAnswer.size()-1 ) {
          mAnswer.insert(mAnswer.begin()+i+1, extra);
        } else {
          mAnswer[i+1].insert(0, extra);
        }
        mAnswer[i].erase(lastSpace+1);
        if( mCursorY == i && mCursorX > lastSpace ) {
          mCursorY += 1;
          mCursorX -= lastSpace+1;
        }
      }
    }
  }
}

void MultiLineQuestion::updateInputLayout() {
  cairo_font_extents_t extents;
  cairo_font_extents( mWindow->mCtx, &extents );

  wrapLines();

  mInputWidth = mMaxWidth + mPad*3;
  mInputHeight = mAnswer.size() * extents.height + mPad;
  if (mInputHeight < 150) mInputHeight = 150;
}


void MultiLineQuestion::drawInput() {
  cairo_font_extents_t extents;
  cairo_font_extents( mWindow->mCtx, &extents );

  cairo_rectangle( mWindow->mCtx,
                   mWindow->getWidth()/2 - mInputWidth/2 + mPad,
                   mInputY - extents.ascent,
                   mInputWidth - mPad*2, mInputHeight );
  cairo_set_source_rgb( mWindow->mCtx, 0.75, 0.75, 0.75 );
  if( mHighlight == HighlightInput ) {
    cairo_fill( mWindow->mCtx );
  } else {
    cairo_stroke( mWindow->mCtx );
  }

  cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
  double y = mInputY;
  for (size_t i=0; i<mAnswer.size(); i++) {
    double w,h;
    textSize( mWindow, mAnswer[i], mFontSize, w, h );
    drawText( mWindow->mCtx, mAnswer[i], mWindow->getWidth()/2 - mInputWidth/2 + mPad*3/2, y, 1, -1 );
    if( i == mCursorY && mHighlight == HighlightInput ) {
      double bw,bh;
      std::string before;

      before = mAnswer[i].substr(0, mCursorX);
      textSize( mWindow, before, mFontSize, bw, bh );

      cairo_move_to( mWindow->mCtx,
                     mWindow->getWidth()/2 - mInputWidth/2 + mPad*3/2 + bw + 1,
                     y - extents.ascent);
      cairo_line_to( mWindow->mCtx,
                     mWindow->getWidth()/2 - mInputWidth/2 + mPad*3/2 + bw + 1,
                     y + extents.descent );
      cairo_stroke( mWindow->mCtx );
    }

    y += extents.height;
  }

}

// void MultiLineQuestion::doDrawing( bool helpText ) {
//   // Need a context for this function.
//   wrapLines();

//   cairo_set_source_rgb(mWindow->mCtx, 230/255.0, 240/255.0, 255/255.0);
//   cairo_paint(mWindow->mCtx);

//   cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
//   cairo_set_font_face( mWindow->mCtx, gCairoFont );
//   cairo_set_font_size( mWindow->mCtx, mFontSize );

//   cairo_font_extents_t extents;
//   cairo_font_extents( mWindow->mCtx, &extents );

//   double qw, qh;
//   textSize( mWindow, mQuery, mFontSize, qw, qh );

//   drawText( mWindow->mCtx, mQuery, mWindow->getWidth()/2, mWindow->getHeight()/2 - 100, 0, -1 );

//   double totalh = extents.height * mAnswer.size();
//   double tinyPad = 5;
//   double maxw = mMaxWidth;
//   double boxWidth = std::max(mMaxWidth + mPad*2, qw + mPad*2);
//   for (size_t i=0; i<mAnswer.size(); i++) {
//     double w,h;
//     textSize( mWindow, mAnswer[i], mFontSize, w, h );
//     if (w > maxw) maxw = w;
//   }
//   if (totalh < 150) totalh = 150;

//   drawBoundingRect( mWindow->getWidth()/2 - boxWidth/2 - mPad,
//                     mWindow->getHeight()/2 - 100 - extents.ascent - mPad,
//                     boxWidth + mPad*2,
//                     totalh + 100 + mPad * 2 + extents.height+mPad*2 - tinyPad );
//   cairo_move_to( mWindow->mCtx,
//                  mWindow->getWidth()/2 - boxWidth/2 - mPad,
//                  mWindow->getHeight()/2 - 100 - extents.ascent + extents.height + mPad );
//   cairo_line_to( mWindow->mCtx,
//                  mWindow->getWidth()/2 + boxWidth/2 + mPad,
//                  mWindow->getHeight()/2 - 100 - extents.ascent + extents.height + mPad );
//   // Submit separator
//   cairo_move_to( mWindow->mCtx,
//                  mWindow->getWidth()/2 - boxWidth/2 - mPad,
//                  mWindow->getHeight()/2 + totalh - mPad );
//   cairo_line_to( mWindow->mCtx,
//                  mWindow->getWidth()/2 + boxWidth/2 + mPad,
//                  mWindow->getHeight()/2 + totalh - mPad);
//   cairo_stroke( mWindow->mCtx );

//   double y = mWindow->getHeight()/2-30;
//   cairo_rectangle( mWindow->mCtx,
//                    mWindow->getWidth()/2 - maxw/2 - tinyPad,
//                    y - extents.ascent - tinyPad,
//                    maxw + tinyPad*2, totalh + tinyPad*2 );
//   if( mDone ) {
//     cairo_set_source_rgb( mWindow->mCtx, 0.75, 0.75, 0.75 );
//     cairo_stroke( mWindow->mCtx );
//   } else {
//     cairo_set_source_rgb( mWindow->mCtx, 0.75, 0.75, 0.75 );
//     cairo_fill( mWindow->mCtx );
//   }

//   if( helpText ) {
//     cairo_set_source_rgb(mWindow->mCtx, 0.6, 0.6, 0.6);
//     drawText( mWindow->mCtx, "Press Enter to submit", mWindow->getWidth()/2, mWindow->getHeight()/2 - extents.ascent - mPad + totalh + mPad * 5 + 50, 0, -1 );
//   }

//   cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
//   for (size_t i=0; i<mAnswer.size(); i++) {
//     double w,h;
//     textSize( mWindow, mAnswer[i], mFontSize, w, h );
//     drawText( mWindow->mCtx, mAnswer[i], mWindow->getWidth()/2 - maxw/2, y, 1, -1 );
//     if( i == mCursorY && !mDone ) {
//       double bw,bh;
//       std::string before;

//       before = mAnswer[i].substr(0, mCursorX);
//       textSize( mWindow, before, mFontSize, bw, bh );

//       cairo_move_to( mWindow->mCtx,
//                      mWindow->getWidth()/2 - maxw/2 + bw + 1,
//                      y - extents.ascent);
//       cairo_line_to( mWindow->mCtx,
//                      mWindow->getWidth()/2 - maxw/2 + bw + 1,
//                      y + extents.descent );
//       cairo_stroke( mWindow->mCtx );
//     }

//     y += extents.height;
//   }

//   y = mWindow->getHeight()/2 - extents.ascent - mPad + totalh + mPad * 5 - tinyPad;

//   std::string done = "Done";
//   double w,h;
//   textSize( mWindow, done, mFontSize, w, h );
//   w += mPad * 2;

//   if( mDone ) {
//     cairo_rectangle( mWindow->mCtx,
//                      mWindow->getWidth()/2 - boxWidth/2,
//                      y - extents.ascent,
//                      boxWidth, extents.height );
//     cairo_set_source_rgb( mWindow->mCtx, 0.75, 0.75, 0.75 );
//     cairo_fill( mWindow->mCtx );
//   }
//   cairo_set_source_rgb(mWindow->mCtx, 0, 0, 0);
//   drawText( mWindow->mCtx, "Done", mWindow->getWidth()/2, y, 0, -1 );
// }

void MultiLineQuestion::start() {
  SDL_StartTextInput();
  open();
  clearEvents();
}

void MultiLineQuestion::finish(bool force) {
  SDL_StopTextInput();
  if( !force ) clearWindow();
}

bool MultiLineQuestion::handleEvent(const SDL_Event &e) {
  bool handled = false;

  // SDL_StartTextInput();
  // open();

  // clearEvents();

  bool update = false;
  switch( e.type ) {
  case SDL_WINDOWEVENT:
    if( e.window.event == SDL_WINDOWEVENT_EXPOSED ) {
      update = true;
      handled = true;
    }
    break;
  case SDL_TEXTINPUT:
    if( mHighlight == HighlightInput ) {
      std::string msg( e.text.text );
      mAnswer[mCursorY].insert( mCursorX, msg );
      mCursorX += msg.size();
      update = true;
    }
    handled = true;
    break;
  case SDL_KEYDOWN:
    if( mHighlight == HighlightInput ) {
      if ( e.key.keysym.sym == SDLK_RETURN ) {
        //   mAnswer.insert(mAnswer.begin()+mCursorY+1, mAnswer[mCursorY].substr(mCursorX));
        //   mAnswer[mCursorY].erase(mCursorX);
        //   mCursorY += 1;
        //   mCursorX = 0;
        //   update = true;
        size_t sz = 0;
        for( size_t i=0; i<mAnswer.size(); i++ )
          sz += mAnswer[i].size();
        if( sz > 0 ) {
          mHighlight = HighlightSubmit;
          update = true;
        }
        handled = true;
      } else if ( e.key.keysym.sym == SDLK_BACKSPACE ) {
        if (mCursorX > 0) {
          mAnswer[mCursorY].erase( mCursorX-1, 1 );
          mCursorX -= 1;
        } else {
          if (mCursorY > 0) {
            if (mAnswer[mCursorY-1].size() > 0)
              mAnswer[mCursorY-1].erase(mAnswer[mCursorY-1].end()-1);
            mCursorX = mAnswer[mCursorY-1].size();
            mAnswer[mCursorY-1].append( mAnswer[mCursorY] );
            mAnswer.erase(mAnswer.begin()+mCursorY, mAnswer.begin()+mCursorY+1);
            mCursorY -= 1;
          }
        }
        update = true;
        handled = true;
      } else if( e.key.keysym.sym == SDLK_LEFT ) {
        if (mCursorX == 0) {
          if( mCursorY > 0 ) {
            mCursorY -= 1;
            mCursorX = mAnswer[mCursorY].size();
          } else {
            mCursorX = 0;
          }
        } else {
          mCursorX -= 1;
        }
        update = true;
        handled = true;
      } else if( e.key.keysym.sym == SDLK_RIGHT ) {
        mCursorX += 1;
        if( mCursorX > mAnswer[mCursorY].size() ) {
          if( mCursorY < mAnswer.size()-1 ) {
            mCursorY += 1;
            mCursorX = 0;
          } else {
            mCursorX = mAnswer[mCursorY].size();
          }
        }
        update = true;
        handled = true;
      } else if( e.key.keysym.sym == SDLK_DOWN ) {
        mCursorY += 1;
        if( mCursorY < mAnswer.size() ) {
          // FIXME: go as straight down as possible
          if (mCursorX > mAnswer[mCursorY].size())
            mCursorX = mAnswer[mCursorY].size();
        } else if( mCursorY >= mAnswer.size() ) {
          mCursorY = mAnswer.size()-1;
          if (mCursorX == mAnswer[mCursorY].size()) {
            mHighlight = HighlightSubmit;
          } else {
            mCursorX = mAnswer[mCursorY].size();
          }
        }
        update = true;
        handled = true;
      } else if( e.key.keysym.sym == SDLK_UP ) {
        if( mCursorY > 0 ) {
          mCursorY -= 1;
          // FIXME: go as straight down as possible
          if (mCursorX > mAnswer[mCursorY].size())
            mCursorX = mAnswer[mCursorY].size();
        } else {
          mCursorY = 0;
          mCursorX = 0;
        }
        update = true;
        handled = true;
      }
      handled = true;
    } else if( mHighlight == HighlightSubmit ) {
      if ( e.key.keysym.sym == SDLK_RETURN ) {
        mDone = QuestionSubmitted;
        handled = true;
      } else if( e.key.keysym.sym == SDLK_UP ) {
        mHighlight = HighlightInput;
        update = true;
        handled = true;
      } else if( e.key.keysym.sym == SDLK_DOWN ) {
        if( mCanGoToPrevious ) mHighlight = HighlightPrev;
        update = true;
        handled = true;
      }
    } else {
      // Prev
      if ( e.key.keysym.sym == SDLK_RETURN ) {
        if( mCanGoToPrevious ) mDone = QuestionGoToPrevious;
        handled = true;
      } else if( e.key.keysym.sym == SDLK_UP ) {
        mHighlight = HighlightSubmit;
        update = true;
        handled = true;
      }
    }
  }

  if( update ) draw(true);

  // SDL_StopTextInput();

  // clearWindow();

  return handled;
}
