#include "graphics.hh"

enum QuestionDoneCode {
  QuestionInProgress = 0,
  QuestionSubmitted,
  QuestionGoToPrevious
};

const int maxQuestionQueryWidth = 500;

class Question {
protected:
  Window *mWindow;
  std::string mName;
  std::string mQuery;
  double mFontSize, mPad;
  int mSubmitDelay;

  double mQueryY, mQueryHeight, mQueryWidth;
  double mInputY, mInputHeight, mInputWidth;
  double mSubmitY, mSubmitHeight, mSubmitWidth;
  double mRectX, mRectY, mRectW, mRectH;

  std::string mSubmitHelpText, mPrevHelpText, mInputHelpText;
  QuestionDoneCode mDone;
  bool mCanGoToPrevious;

public:

  Question( Window *w, const std::string &name, const std::string &query );

  virtual ~Question();
  void clearEvents();
  void clearWindow();
  void open();

  void updateLayout();
  void updateQueryLayout();
  void updateSubmitLayout();
  void updateBoxLayout();
  virtual void updateInputLayout() = 0;
  virtual bool isSubmitHighlighted() = 0;
  virtual bool isPrevHighlighted() = 0;

  void drawBoundingRect(double x, double y, double w, double h);
  void drawSubmit();
  void drawQuery();
  virtual void drawInput() = 0;
  void drawCommon( bool helpText );
  void draw( bool helpText );
  void doDrawing( bool helpText );
  QuestionDoneCode getDoneCode();

  void setCanGoToPrevious( bool flag );

  std::string getName();

  virtual bool handleEvent(const SDL_Event &event) = 0;
  virtual std::string getAnswer() = 0;

  virtual void start() = 0;
  virtual void finish(bool force) = 0;
};


class RadioQuestion : public Question {
protected:
  std::vector<std::string> mOptions, mStrings;
  unsigned int mCurrent, mSelected;
  bool mSelectedValid;
  double mRadioRadius;
public:
  RadioQuestion(Window *w, const std::string &name, const std::string &query, std::vector<std::string> options, std::vector<std::string> strings);

  ~RadioQuestion();

  void select( unsigned int idx );

  std::string getAnswer();

  bool handleEvent(const SDL_Event &event);

  bool isSubmitHighlighted();
  bool isPrevHighlighted();

  void drawRadio(double x, double y, bool state);
  void updateInputLayout();
  void drawInput();

  void start();
  void finish(bool force);
};

class CheckboxQuestion : public Question {
protected:
  std::vector<std::string> mOptions, mStrings;
  unsigned int mCurrent;
  std::vector<bool> mSelected;
  double mCheckboxSize;
public:
  CheckboxQuestion(Window *w, const std::string &name, const std::string &query, std::vector<std::string> options, std::vector<std::string> strings);

  ~CheckboxQuestion();

  void check( unsigned int idx, bool flag = true );

  bool isSubmitHighlighted();
  bool isPrevHighlighted();
  void drawInput();
  void drawCheckbox(double x, double y, bool state);
  void updateInputLayout();

  bool handleEvent(const SDL_Event &event);
  std::string getAnswer();
  void start();
  void finish(bool force);
};

enum Highlight {
  HighlightInput,
  HighlightSubmit,
  HighlightPrev
} ;

class OneLineQuestion : public Question {
protected:
  std::string mAnswer;
  Highlight mHighlight;

public:
  OneLineQuestion(Window *w, const std::string &name, const std::string &query);

  ~OneLineQuestion();

  void setAnswer( const std::string &answer );

  bool isSubmitHighlighted();
  bool isPrevHighlighted();
  void updateInputLayout();
  void drawInput();

  bool handleEvent(const SDL_Event &event);
  std::string getAnswer();
  void start();
  void finish(bool force);
};

class MultiLineQuestion : public Question {
protected:
  std::vector<std::string> mAnswer;
  unsigned int mCursorX, mCursorY;
  unsigned int mMaxWidth;
  Highlight mHighlight;

public:
  MultiLineQuestion(Window *w, const std::string &name, const std::string &query);

  ~MultiLineQuestion();

  void setAnswer( const std::string &answer );

  bool isSubmitHighlighted();
  bool isPrevHighlighted();
  void updateInputLayout();
  void drawInput();

  void wrapLines();
  bool handleEvent(const SDL_Event &event);
  std::string getAnswer();
  void start();
  void finish(bool force);
};
