'use strict';

const game = require('./game.js');

function ShowInformation(delay, contText) {
    this.continueText = contText || "Press SPACE to continue";
    this.state = {screen: "message",
                  lines: ["Waiting for Players"],
                  continue: ['', '']};
    this.delay = delay;
}

ShowInformation.prototype = {};

ShowInformation.prototype.init = function (server) {
    this.server = server;
    this.status = ['in-progress', 'in-progress'];
    this.updateState();
    if (this.delay > 0) {
        this.canContinue = false;
        this.continueTimer = setTimeout(this.allowContinue.bind(this), this.delay);
    } else {
        this.allowContinue();
    }
};

ShowInformation.prototype.updateState = function () {
    for (let i=0; i<this.status.length; i++) {
        var p = i+1;
        if (!this.canContinue) {
            this.state.continue[i] = '';
        } else if (this.status[i] === 'ready') {
            this.state.continue[i] = 'Player '+p+' is ready';
        } else if (this.status[i] === 'disconnected') {
            this.state.continue[i] = 'Player '+p+' - Connection lost';
        } else {
            this.state.continue[i] = 'Player '+p+' - ' + this.continueText;
        }
    }
};

ShowInformation.prototype.resetStatus = function() {
    for (let i=0; i<this.status.length; i++) {
        this.status[i] = 'in-progress';
    }
};

ShowInformation.prototype.readyToContinue = function() {
    for (let i=0; i<this.status.length; i++)
        if (this.status[i] !== 'ready') return false;
    return true;
};

ShowInformation.prototype.allowContinue = function() {
    this.canContinue = true;
    this.updateState();
    this.server.broadcastScreenState(this.state);
};

ShowInformation.prototype.cleanup = function() {
    clearTimeout(this.continueTimer);
};

function SplitInformation(delay, contText) {
    this.continueText = contText || "Press SPACE to continue";
    this.playerState = [{screen: "message",
                         lines: ["Waiting for Players"],
                         continue: ''},
                        {screen: "message",
                         lines: ["Waiting for Players"],
                         continue: ''}];
    this.delay = delay;
}

SplitInformation.prototype = {};

SplitInformation.prototype.init = function (server) {
    this.server = server;
    this.status = ['in-progress', 'in-progress'];
    this.canContinue = [false, false];
    this.scheduleAllowContinue(0);
    this.scheduleAllowContinue(1);
};

SplitInformation.prototype.updateState = function (player) {
    if (this.status[player] === 'ready') {
        this.playerState[player].continue = '';
    } else if (this.canContinue[player]) {
        this.playerState[player].continue = this.continueText;
    } else {
        this.playerState[player].continue = '';
    }
};

SplitInformation.prototype.readyToContinue = function() {
    for (let i=0; i<this.status.length; i++)
        if (this.status[i] !== 'ready') return false;
    return true;
};

SplitInformation.prototype.scheduleAllowContinue = function(player) {
    this.canContinue[player] = false;
    if (this.delay > 0) {
        this.continueTimer = setTimeout(() => { this.allowContinue(player) }, this.delay);
    } else {
        this.allowContinue(player);
    }
};

SplitInformation.prototype.allowContinue = function(player) {
    this.canContinue[player] = true;
    this.updateState(player);
    this.server.sendScreenState(player, this.playerState[player]);
};

SplitInformation.prototype.cleanup = function() {
    clearTimeout(this.continueTimer);
};


function Message(lines, contText) {
    ShowInformation.call(this, 0, contText);
    this.screen_id = 'message';
    this.state.lines = lines;
}

Message.prototype = Object.create(ShowInformation.prototype);

Message.prototype.init = function (server) {
    ShowInformation.prototype.init.call(this, server);
    var bonus = this.server.formatBonus();
    for (let i=0; i<this.state.lines.length; i++) {
        if (typeof this.state.lines[i] === 'string')
            this.state.lines[i] = this.state.lines[i].replace('%bonus%', bonus);
    }
    this.server.broadcastScreenState(this.state);
    this.server.lgScreen('message', {lines: this.state.lines});
};

Message.prototype.handleInputEvent = function(event) {
    if (event.type === 'gameKeyDown' && event.key === 4 && this.canContinue) {
        this.status[event.player] = 'ready';
        this.updateState();
        this.server.broadcastScreenState(this.state);
        if (this.readyToContinue()) {
            clearTimeout(this.continueTimer);
            this.continueTimer = setTimeout(this.server.nextScreen.bind(this.server), 350);
        }
    } else if (event.type === 'connection') {
        console.log('message got connection', event);
        this.status[event.player] = event.value ? 'in-progress':'disconnected';
        this.updateState();
        this.server.broadcastScreenState(this.state);
    }
};

Message.prototype.cleanup = function() {
    ShowInformation.prototype.cleanup.call(this);
    clearTimeout(this.continueTimer);
    this.server.lg('end');
};

function End(lines) {
    this.screen_id = 'End';
    this.state = {screen: "message",
                  lines: lines,
                  continue: ['', '']};
}

End.prototype = {};

End.prototype.init = function (server) {
    this.server = server;
    var bonus = this.server.formatBonus();
    for (let i=0; i<this.state.lines.length; i++) {
        if (typeof this.state.lines[i] === 'string')
            this.state.lines[i] = this.state.lines[i].replace('%bonus%', bonus);
    }
    this.server.broadcastScreenState(this.state);
    this.server.broadcastScreenState({screen: 'clientCommand', cmd: 'end'});
    setImmediate(() => { this.server.shutdown(); });
};

End.prototype.handleInputEvent = function(event) {
    if (event.type === 'connection') {
        console.log('message got connection', event);
        if (event.value)
            this.server.sendScreenState(event.player, this.state);
    }
};

End.prototype.cleanup = function() {
};


function Intermission(duration) {
    ShowInformation.call(this, 0);
    this.duration = duration;
    this.screen_id = 'intermission';
}

Intermission.prototype = Object.create(ShowInformation.prototype);

Intermission.prototype.init = function (server) {
    // ShowInformation.prototype.init.call(this, server);
    this.server = server;
    this.status = ['in-progress', 'in-progress'];
    this.canContinue = false;
    this.updateState();
    this.state.lines = [{title: 'Intermission'},
                        '',
                        'Take a break and be ready to continue in:',
                        '',
                        ''];
    this.state.bg = 0x000000;
    this.state.fg = 0xEEEEEE;
    this.state.hl = 0x999999,

    this.startTime = (new Date()).getTime();
    this.updateCountDown();
    this.server.broadcastScreenState(this.state);
    this.server.lgScreen('intermission', {duration: this.duration,
                                          lines: this.state.lines});
};

ShowInformation.prototype.updateCountDown = function () {
    var now = (new Date()).getTime();
    var remaining = this.duration - (now - this.startTime);
    // console.log(now, this.startTime, this.duration, remaining);
    if (remaining > 0) {
        var min = Math.floor(remaining / 60000);
        var sec = Math.floor((remaining % 60000) / 1000);
        var min_s = min === 1 ? '':'s';
        this.state.lines[4] = min.toString() + ' minute'+min_s+' and ' + sec.toString() + ' seconds.';
        this.countdownTimer = setTimeout(this.updateCountDown.bind(this), 1000);
    } else {
        this.state.lines = ['The intermission has finished.',
                            '',
                            'Please return to your computer.'];
        this.allowContinue();
    }
    this.updateState();
    this.server.broadcastScreenState(this.state);
};

Intermission.prototype.handleInputEvent = function(event) {
    if (event.type === 'gameKeyDown' && event.key === 4 && this.canContinue) {
        this.status[event.player] = 'ready';
        this.updateState();
        this.server.broadcastScreenState(this.state);
        if (this.readyToContinue()) {
            clearTimeout(this.continueTimer);
            this.continueTimer = setTimeout(this.server.nextScreen.bind(this.server), 350);
        }
    } else if (event.type === 'connection') {
        this.status[event.player] = event.value ? 'in-progress':'disconnected';
        this.updateState();
        this.server.broadcastScreenState(this.state);
    }
};

Intermission.prototype.cleanup = function() {
    ShowInformation.prototype.cleanup.call(this);
    clearTimeout(this.continueTimer);
    clearTimeout(this.countdownTimer);
    this.server.lg('end');
};

function Consent() {
    this.consentPages = require('./consent');
};

Consent.prototype = {};

Consent.prototype.updateText = function (player) {
};

Consent.prototype.createQuestions = function (player) {
    this.questions[player] = [{type: 'radio',
                               name: 'consent',
                               query: "By answering 'yes' below, you agree that information in the consent form has been explained to you and all your current questions have been answered. You are encouraged to ask questions about any aspect of this research study during the course of the study and in the future. By answering 'yes' below, you agree to participate in this research study. A copy of the consent form will be given to you.",
                               options: [["yes", "Yes. I wish to participate in this study."],
                                         ["no", "No. I do not want to participate."],
                                         ["review", "Review Consent Form"]],
                               disablePrev: true}];
};

Consent.prototype.init = function (server) {
    this.server = server;
    this.status = ['in-progress', 'in-progress'];
    this.playerIndex = [0, 0];
    this.survey = [{}, {}];
    this.questions = [{},{}];
    this.currentConsentPage = [0,0];
    this.createQuestions(0);
    this.createQuestions(1);
    this.server.lgScreen('consent', {questions: this.questions[0]});
    this.server.sendScreenState(0, {screen: "question",
                                    question: this.questions[0][this.playerIndex[0]]});
    this.server.sendScreenState(1, {screen: "question",
                                    question: this.questions[1][this.playerIndex[1]]});
};

Consent.prototype.readyToContinue = function() {
    for (let i=0; i<this.status.length; i++)
        if (this.status[i] !== 'ready') return false;
    return true;
};

Consent.prototype.sendConsentPage = function(player) {
    var screen = {screen: "wrappedMessage",
                  lines: this.consentPages[this.currentConsentPage[player]],
                  continue: `Page ${this.currentConsentPage[player]+1} of ${this.consentPages.length} -- press ENTER to continue`};
    this.server.sendScreenState(player, screen);
}

Consent.prototype.handleInputEvent = function(event) {
    var update = false;
    if (event.type === 'questionSubmitted') {
        if (event.answer === 'review') {
            this.status[event.player] = 'reviewing';
            this.currentConsentPage[event.player] = 0;
            this.sendConsentPage(event.player);
        } else if (event.answer === 'no') {
            this.status[event.player] = 'declined';
            this.server.sendScreenState(event.player, {screen: "message",
                                                       lines: ["You have chosen not to participate.",
                                                               "Please contact the experimenter."],
                                                       continue: ''});
        } else {
            this.survey[event.player][event.question] = event.answer;
            this.server.lg('surveyResults', {player: event.player,
                                             survey: this.survey[event.player]});
            this.status[event.player] = 'ready';
            this.server.sendScreenState(event.player, {screen: "message",
                                                       lines: ["Waiting for other player"],
                                                       continue: ''});
            if (this.readyToContinue()) {
                clearTimeout(this.continueTimer);
                this.continueTimer = setTimeout(this.server.nextScreen.bind(this.server), 350);
            }
        }
    } else if (event.type === 'connection') {
        this.status[event.player] = event.value ? 'in-progress':'disconnected';
        if (event.value) {
            this.server.sendScreenState(event.player,
                                        {screen: "question",
                                         question: this.questions[event.player][0]});
        }
    } else if (event.type === 'questionGoToPrevious') {
        // Impossible
    } else if (event.type === 'keyDown' && event.name === 'Return') {
        if (this.status[event.player] === 'reviewing') {
            this.currentConsentPage[event.player] += 1;
            if (this.currentConsentPage[event.player] >= this.consentPages.length) {
                this.status[event.player] = 'in-progress';
                this.server.sendScreenState(event.player,
                                            {screen: "question",
                                             question: this.questions[event.player][0]});
            } else {
                this.sendConsentPage(event.player);
            }
        }
    }
};

Consent.prototype.cleanup = function() {
    clearTimeout(this.continueTimer);
    this.server.lg('end', {playerMapping: [this.server.clients[0].subjectId, this.server.clients[1].subjectId]});
};


function Demographics() {
    this.continueText = 'Completing Survey...';
    this.playerState = [{screen: "message",
                         lines: ["Waiting for other player"],
                         continue: ''},
                        {screen: "message",
                         lines: ["Waiting for other player"],
                         continue: ''}];
};

Demographics.prototype = {};

Demographics.prototype.updateText = function (player) {
};

Demographics.prototype.createQuestions = function (player) {
    var doText = this.survey[player]['currentlyPlay'] === 'no' ? 'did':'do';
    var haveText = this.survey[player]['currentlyPlay'] === 'no' ? 'did you play':'have you played';
    this.questions[player] = [{type: 'radio',
                               name: 'gender',
                               query: "What is your gender?",
                               options: [["male", "Male"],
                                         ["female", "Female"],
                                         ["other", "Other"]],
                               disablePrev: true},
                              {type: 'oneline',
                               name: 'age',
                               query: "What is your age?"},
                              {type: 'radio',
                               name: 'ethnicity',
                               query: "What is your ethnicity?",
                               options: [["white", "White or Caucasian"],
                                         ["black", "Black or African-American"],
                                         ["asian", "Asian"],
                                         ["indian", "American Indian or Alaska Native"],
                                         ["hawaiian", "Native Hawaiian or Other Pacific Islander"],
                                         ["other", "Other"]]},
                              {type: 'oneline',
                               name: 'ethnicityOther',
                               query: "Please specify your ethnicity."},
                              {type: 'radio',
                               name: 'dominantHand',
                               query: "What is your dominant hand?",
                               options: [["right", "Right"],
                                         ["left", "Left"],
                                         ["ambidextrous", "Ambidextrous"]]},
                              {type: 'radio',
                               name: 'fieldOfStudy',
                               query: "If you are a student, what is your field of study?",
                               options: [["engineering", "Engineering"],
                                         ["basicsciences", "Basic Sciences"],
                                         ["humanities", "Humanities and Social Sciences"],
                                         ["business", "Business"],
                                         ["arts", "Arts"],
                                         ["other", "Other"],
                                         ["notstudent", "Not a Student"]]},
                              {type: 'oneline',
                               name: 'fieldOfStudyOther',
                               query: "Please specify your field of study."},
                              {type: 'radio',
                               name: 'played',
                               query: "Have you ever played video games?",
                               options: [["yes", "Yes"],
                                         ["no", "No"]]},
                              {type: 'radio',
                               name: 'currentlyPlay',
                               query: "Do you currently play video games?",
                               options: [["yes", "Yes"],
                                         ["no", "No"]]},
                              {type: 'radio',
                               name: 'controllerHand',
                               query:'When using only a keyboard, which hand '+doText+' you typically use to control the movement of characters or objects?',
                               options: [['right','Right Hand'],
                                         ['left', 'Left Hand'],
                                         ['both', 'Both'],
                                         ['noexperience', 'No Experience']]},
                              {type: 'radio',
                               name: 'howLong',
                               query: 'For how long '+haveText+' video games?',
                               options: [['lt1month', 'Less than 1 month'],
                                         ['1to6months', '1 to 6 months'],
                                         ['6to12months', '6 to 12 months'],
                                         ['1to4years', '1 to 4 years'],
                                         ['4to8years', '4 to 8 years'],
                                         ['gt8years', 'More than 8 years']]},
                              {type: 'radio',
                               name: 'platform',
                               query: 'Which gaming platform '+doText+' you use the most?',
                               options: [['pc', 'PC'],
                                         ['console', 'Console'],
                                         ['handheld', 'Handheld Console'],
                                         ['mobile', 'Mobile phone/tablet']]},
                              {type: 'oneline',
                               name: 'genres',
                               query: 'What genre(s) of video games '+doText+' you play mostly?'},
                              {type: 'radio',
                               name: 'frequency',
                               query: 'How frequently '+doText+' you play video games?',
                               options: [['everyday', 'Every day'],
                                         ['weekly', 'Weekly'],
                                         ['monthly', 'Monthly'],
                                         ['irregularly', 'Irregularly']]},
                              {type: 'radio',
                               name: 'hoursPlayed',
                               query: 'How many hours a week '+doText+' you play video games?',
                               options: [['lt1hr', '< 1 Hour'],
                                         ['1to5hr', '1-5 Hours'],
                                         ['5to10hr', '5-10 Hours'],
                                         ['10to20hr', '10-20 Hours'],
                                         ['20to30hr', '20-30 Hours'],
                                         ['gt30hr', '> 30 Hours']]},
                              {type: 'check',
                               name: 'relationship',
                               query: 'Which statements describe your relationship with video games?  Check all that apply.',
                               options: [['social', 'Video games provide a common social interest that I share with my friends.'],
                                         ['challenge', 'I enjoy the mental and physical challenges video games provide.'],
                                         ['competitive', 'I enjoy competitive games with human or computer opponents more than casual games.'],
                                         ['skilled', 'I consider myself skilled (better than average) in at least one video game I play.']]},
                              {type: 'radio',
                               name: 'heard',
                               query: 'Did you hear about this study from anyone?',
                               options: [['no', 'No'],
                                         ['yes', 'Yes']]},
                              {type: 'multiline',
                               name: 'heardDescription',
                               query: 'Please describe what you heard.'}];
};

Demographics.prototype.init = function (server) {
    this.server = server;
    this.status = ['in-progress', 'in-progress'];
    this.playerIndex = [0, 0];
    this.survey = [{}, {}];
    this.questions = [{},{}];
    this.createQuestions(0);
    this.createQuestions(1);
    this.server.lgScreen('demographics', {questions: this.questions[0]});
    this.server.sendScreenState(0, {screen: "question",
                                    question: this.questions[0][this.playerIndex[0]]});
    this.server.sendScreenState(1, {screen: "question",
                                    question: this.questions[1][this.playerIndex[1]]});
};

Demographics.prototype.updateState = function (player) {
    if (this.status[player] === 'ready') {
        this.playerState[player].continue = '';
    } else if (this.canContinue[player]) {
        this.playerState[player].continue = this.continueText;
    } else {
        this.playerState[player].continue = '';
    }
};

Demographics.prototype.readyToContinue = function() {
    for (let i=0; i<this.status.length; i++)
        if (this.status[i] !== 'ready') return false;
    return true;
};

Demographics.prototype.handleInputEvent = function(event) {
    var update = false;
    if (event.type === 'questionSubmitted') {
        this.survey[event.player][event.question] = event.answer;
        if (this.questions[event.player][this.playerIndex[event.player]].name === 'ethnicity' &&
            this.survey[event.player]['ethnicity'] !== 'other') {
            this.playerIndex[event.player] += 2;
        } else if (this.questions[event.player][this.playerIndex[event.player]].name === 'fieldOfStudy' &&
                   this.survey[event.player]['fieldOfStudy'] !== 'other') {
            this.playerIndex[event.player] += 2;
        } else if (this.questions[event.player][this.playerIndex[event.player]].name === 'played' &&
                   this.survey[event.player]['played'] === 'no') {
            for (let i=0; i<this.questions[event.player].length; i++) {
                if (this.questions[event.player][i].name === 'heard') {
                    this.playerIndex[event.player] = i;
                    break;
                }
            }
        } else if (this.questions[event.player][this.playerIndex[event.player]].name === 'currentlyPlay') {
            this.createQuestions(event.player);
            this.playerIndex[event.player] += 1;
        } else if (this.questions[event.player][this.playerIndex[event.player]].name === 'heard' &&
                   this.survey[event.player]['heard'] === 'no') {
            this.playerIndex[event.player] += 2;
        } else {
            this.playerIndex[event.player] += 1;
        }
        update = true;
    } else if (event.type === 'questionGoToPrevious') {
        if (this.questions[event.player][this.playerIndex[event.player]].name === 'dominantHand' &&
            this.survey[event.player]['ethnicity'] !== 'other') {
            this.playerIndex[event.player] -= 2;
        } else if (this.questions[event.player][this.playerIndex[event.player]].name === 'played' &&
            this.survey[event.player]['fieldOfStudy'] !== 'other') {
            this.playerIndex[event.player] -= 2;
        } else if (this.questions[event.player][this.playerIndex[event.player]].name === 'heard' &&
                   this.survey[event.player]['played'] === 'no') {
                for (let i=0; i<this.questions[event.player].length; i++) {
                    if (this.questions[event.player][i].name === 'played') {
                        this.playerIndex[event.player] = i;
                        break;
                    }
                }
        } else if (this.questions[event.player][this.playerIndex[event.player]].name === 'currentlyPlay') {
            this.updateText();
            this.createQuestions();
            this.playerIndex[event.player] -= 1;
        } else {
            this.playerIndex[event.player] -= 1;
        }
        if (this.playerIndex[event.player] < 0) this.playerIndex[event.player] = 0;
        update = true;
    } else if (event.type === 'connection') {
        this.status[event.player] = event.value ? 'in-progress':'disconnected';
        if (event.value) {
            update = true;
        }
    }

    if (update) {
        if (this.playerIndex[event.player] < this.questions[event.player].length) {
            this.server.sendScreenState(event.player, {screen: "question",
                                                       answer: this.survey[event.player][this.questions[event.player][this.playerIndex[event.player]].name],
                                                       question: this.questions[event.player][this.playerIndex[event.player]]});
        } else {
            this.server.lg('surveyResults', {player: event.player,
                                             survey: this.survey[event.player]});
            this.status[event.player] = 'ready';
            this.updateState(event.player);
            this.server.sendScreenState(event.player, this.playerState[event.player]);
            if (this.readyToContinue()) {
                clearTimeout(this.continueTimer);
                this.continueTimer = setTimeout(this.server.nextScreen.bind(this.server), 350);
            }
        }
    }
};

Demographics.prototype.cleanup = function() {
    clearTimeout(this.continueTimer);
    this.server.lg('end');
};

function Feedback() {
    this.continueText = 'Completing Survey...';
    this.playerState = [{screen: "message",
                         lines: ["Waiting for other player"],
                         continue: ''},
                        {screen: "message",
                         lines: ["Waiting for other player"],
                         continue: ''}];
    this.questions = [{type:'multiline',
                       name: 'difficulties',
                       query: 'Please describe the difficulties that you faced when learning to play this computer game.',
                       disablePrev: true},
                      {type: 'multiline',
                       name: 'attemptedStrategies',
                       query: 'Please describe all the strategies that you attempted while playing this game.'},
                      {type: 'multiline',
                       name: 'successfulStrategy',
                       query: 'Please describe the strategy that you settled on and explain why you decided to use that strategy.'},
                      {type: 'radio',
                       name: 'playedBefore',
                       query: "Have you played a game which had mechanics similar to the game in the experiment?",
                       options: [["yes", "Yes"],
                                 ["no", "No"]]},
                      {type: 'oneline',
                       name: 'game',
                       query: 'The name of the similar game is:'},
                      {type: 'check',
                       name: 'physics',
                       query: 'Which of the following physics concepts are you familiar with? Check all that apply.',
                       options: [['addition', 'Vector Addition'],
                                 ['velocity', 'Velocity & Acceleration'],
                                 ['momentum', 'Conservation of Momentum']]},
                      {type: 'multiline',
                       name: 'feedback',
                       query: 'Please add any comments, questions, and/or suggestions below.'}];
};

Feedback.prototype = {};

Feedback.prototype.init = function (server) {
    this.server = server;
    this.status = ['in-progress', 'in-progress'];
    this.playerIndex = [0, 0];
    this.survey = [{}, {}];
    this.server.lgScreen('feedback', {questions: this.questions});
    this.server.sendScreenState(0, {screen: "question",
                                    question: this.questions[this.playerIndex[0]]});
    this.server.sendScreenState(1, {screen: "question",
                                    question: this.questions[this.playerIndex[1]]});
};

Feedback.prototype.updateState = function (player) {
    if (this.status[player] === 'ready') {
        this.playerState[player].continue = '';
    } else if (this.canContinue[player]) {
        this.playerState[player].continue = this.continueText;
    } else {
        this.playerState[player].continue = '';
    }
};

Feedback.prototype.readyToContinue = function() {
    for (let i=0; i<this.status.length; i++)
        if (this.status[i] !== 'ready') return false;
    return true;
};

Feedback.prototype.handleInputEvent = function(event) {
    var update = false;

    if (event.type === 'questionSubmitted') {
        this.survey[event.player][event.question] = event.answer;
        if (this.questions[this.playerIndex[event.player]].name === 'playedBefore' &&
            this.survey[event.player]['playedBefore'] === 'no') {
            this.playerIndex[event.player] += 2;
        } else {
            this.playerIndex[event.player] += 1;
        }
        update = true;
    } else if (event.type === 'questionGoToPrevious') {
        if (this.questions[this.playerIndex[event.player]].name === 'physics' &&
            this.survey[event.player]['playedBefore'] === 'no') {
            this.playerIndex[event.player] -= 2;
        } else {
            this.playerIndex[event.player] -= 1;
        }
        if (this.playerIndex[event.player] < 0) this.playerIndex[event.player] = 0;
        update = true;
    } else if (event.type === 'connection') {
        this.status[event.player] = event.value ? 'in-progress':'disconnected';
        if (event.value) {
            update = true;
        }
    }

    if (update) {
        if (this.playerIndex[event.player] < this.questions.length) {
            this.server.sendScreenState(event.player, {screen: "question",
                                                       answer: this.survey[event.player][this.questions[this.playerIndex[event.player]].name],
                                                       question: this.questions[this.playerIndex[event.player]]});
        } else {
            this.server.lg('surveyResults', {player: event.player,
                                             survey: this.survey[event.player]});
            this.status[event.player] = 'ready';
            this.updateState(event.player);
            this.server.sendScreenState(event.player, this.playerState[event.player]);
            if (this.readyToContinue()) {
                clearTimeout(this.continueTimer);
                this.continueTimer = setTimeout(this.server.nextScreen.bind(this.server), 350);
            }
        }
    }
};

Feedback.prototype.cleanup = function() {
    clearTimeout(this.continueTimer);
    this.server.lg('end');
};

function InstructionGuts(server, player, instructions) {
    this.server = server;
    this.instructions = instructions;
    this.player = player;
    this.curPage = 0;
    this.state = {screen: "message",
                  lines: [],
                  continue: ''};
    this.state.lines = this.instructions[this.curPage];
    this.canContinue = false;
    this.done = false;
    this.delay = 250;
}

InstructionGuts.prototype = {};

InstructionGuts.prototype.start = function() {
    this.updatePage();
    this.server.lg('instructions-start', {player: this.player, page:this.curPage});
}

InstructionGuts.prototype.handleInputEvent = function (event, doneFn) {
    if (event.type === 'gameKeyDown' && event.key === 1 && this.canContinue) {
        this.prevPage(doneFn);
    } else if (event.type === 'gameKeyDown' && event.key === 2 && this.canContinue) {
        this.nextPage(doneFn);
    } else if (event.type === 'connection') {
        if (event.value === true) {
            this.updatePage();
        }
    }
};

InstructionGuts.prototype.updateState = function () {
    if (this.done) {
        this.state.continue = '';
    } else if (this.canContinue) {
        this.state.continue = 'previous page-A   Page '+(this.curPage+1)+' of '+this.instructions.length+'   D-' + (this.curPage === this.instructions.length-1?'continue':'next page');
    } else {
        this.state.continue = '';
    }
};

InstructionGuts.prototype.updatePage = function() {
    this.state.lines = this.instructions[this.curPage];
    this.scheduleAllowContinue();
    this.updateState();
    this.server.sendScreenState(this.player, this.state);
};

InstructionGuts.prototype.nextPage = function(doneFn) {
    if (!this.done) {
        if( this.curPage < this.instructions.length-1 ) {
            this.curPage += 1;
            this.updatePage();
            this.server.lg('instructions-next-page', {player: this.player, page: this.curPage});
        } else {
            this.done = true;
            this.server.lg('instructions-finished', {player: this.player});
            doneFn();
        }
    }
};

InstructionGuts.prototype.prevPage = function() {
    if (!this.done) {
        if( this.curPage > 0 ) {
            this.curPage -= 1;
            this.updatePage();
            this.server.lg('instructions-prev-page', {player: this.player, page: this.curPage});
        }
    }
};

InstructionGuts.prototype.scheduleAllowContinue = function() {
    this.canContinue = false;
    if (this.delay > 0) {
        this.clearTimer();
        this.continueTimer = setTimeout(() => { this.allowContinue() }, this.delay);
    } else {
        this.allowContinue();
    }
};

InstructionGuts.prototype.clearTimer = function() {
    clearTimeout(this.continueTimer);
};

InstructionGuts.prototype.allowContinue = function() {
    this.canContinue = true;
    this.updateState();
    this.server.sendScreenState(this.player, this.state);
};


function Instructions(instructions) {
    this.instructions = instructions;
}

Instructions.prototype = Object.create(SplitInformation.prototype);

Instructions.prototype.init = function(server) {
    this.server = server;
    this.players = [new InstructionGuts(server, 0, this.instructions[0]),
                    new InstructionGuts(server, 1, this.instructions[1])];
    this.server.lgScreen('instructions', {pages: [this.instructions[0].length, this.instructions[1].length]});
    this.players[0].start();
    this.players[1].start();
};

Instructions.prototype.handleInputEvent = function(event) {
    if (event.player !== undefined) {
        if (this.players[event.player].done) {
            // ignore
        } else {
            this.players[event.player].handleInputEvent( event,
                                                         () => {
                if (this.players[0].done && this.players[1].done) {
                    this.server.nextScreen();
                } else {
                    this.players[event.player].state.lines = ["Waiting for other player"];
                    this.players[event.player].updateState();
                    this.server.sendScreenState(event.player, this.players[event.player].state);
                }
            });
        }
    }
};

Instructions.prototype.cleanup = function() {
    for (let i=0; i<this.players.length; i++)
        this.players[i].clearTimer();
    this.server.lg('end');
};

function GameStart(gnum, lines, instructions) {
    this.gnum = gnum;
    this.continueText = "Press SPACE to start the game";
    this.playerState = new Array(2);
    // share memory to save on updates
    this.continue = ['', ''];
    this.playerState[0] = {screen: "message",
                           lines: lines[0],
                           continue: this.continue};
    this.playerState[1] = {screen: "message",
                           lines: lines[1],
                           continue: this.continue};
    this.delay = 0;
    this.instructions = instructions;
}

GameStart.prototype = {};

GameStart.prototype.init = function(server) {
    this.server = server;
    this.status = ['in-progress', 'in-progress'];
    this.playerMode = ['start', 'start'];
    this.playerInstructions = [null, null];
    this.updateState();
    if (this.delay > 0) {
        this.canContinue = false;
        this.continueTimer = setTimeout(this.allowContinue.bind(this), this.delay);
    } else {
        this.allowContinue();
    }
    this.server.sendScreenState(0, this.playerState[0]);
    this.server.sendScreenState(1, this.playerState[1]);
    this.server.lgScreen('gamestart', {gnum: this.gnum});
};

GameStart.prototype.updateState = function () {
    for (let i=0; i<this.status.length; i++) {
        var p = i+1;
        if (this.status[i] === 'disconnected') {
            this.continue[i] = 'Player '+p+' - Connection lost';
        } else if (this.playerMode[i] === 'instructions') {
            this.continue[i] = 'Player '+p+' is reading the instructions';
        } else if (!this.canContinue) {
            this.continue[i] = '';
        } else if (this.status[i] === 'ready') {
            this.continue[i] = 'Player '+p+' is ready';
        } else {
            this.continue[i] = 'Player '+p+' - ' + this.continueText;
        }
    }
};

GameStart.prototype.resetReady = ShowInformation.prototype.resetReady;

GameStart.prototype.readyToContinue = ShowInformation.prototype.readyToContinue;

GameStart.prototype.allowContinue = function() {
    this.canContinue = true;
    this.updateState();
    this.server.sendScreenState(0, this.playerState[0]);
    this.server.sendScreenState(1, this.playerState[1]);
};

GameStart.prototype.maybeBroadcastState = function () {
    for (var i=0; i<this.playerMode.length; i++) {
        if (this.playerMode[i] === 'start') {
            this.server.sendScreenState(i, this.playerState[i]);
        }
    }
}

GameStart.prototype.handleInputEvent = function(event) {
    if (event.type === 'keyDown' && event.name === 'I') {
        var p = event.player || 0;
        if (this.playerMode[p] === 'start') {
            this.status[event.player] = 'in-progress';
            this.playerInstructions[p] = new InstructionGuts(this.server, p, this.instructions[p]);
            this.playerMode[p] = 'instructions';
            this.playerInstructions[p].start();
            this.updateState();
            this.maybeBroadcastState();
        }
    } else if (event.player !== undefined) {
        if (event.type === 'connection') {
            this.status[event.player] = event.value ? 'in-progress':'disconnected';
            this.updateState();
            this.maybeBroadcastState();
        }
        if (this.playerMode[event.player] === 'instructions') {
            this.playerInstructions[event.player].handleInputEvent(event,
                                                                   () => {
                                                                       this.playerMode[event.player] = 'start';
                                                                       this.updateState();
                                                                       this.maybeBroadcastState();
                                                                   });
        } else if (event.type === 'gameKeyDown' && event.key === 4 && this.canContinue) {
            this.status[event.player] = 'ready';
            this.updateState();
            this.maybeBroadcastState();
            if (this.readyToContinue()) {
                clearTimeout(this.continueTimer);
                this.continueTimer = setTimeout(this.server.nextScreen.bind(this.server), 350);
            }
        }
    }
};

GameStart.prototype.cleanup = function() {
    clearTimeout(this.continueTimer);
    this.server.lg('end');
}

function Score(game) {
    ShowInformation.call(this, 0);
    this.screen_id = 'score';
    this.game = game;
    this.delay = 1000;
}

Score.prototype = Object.create(ShowInformation.prototype);

Score.prototype.formatDollars = function (cents) {
    var d = (cents / 100).toFixed(2);
    return '$' + d;
};

Score.prototype.init = function (server) {
    ShowInformation.prototype.init.call(this, server);
    this.state.lines = [{title: 'Score'},
                        '',
                        'You scored ' + this.server.lastGame.points + ' points.',
                        'You earned a bonus of ' + this.formatDollars(this.server.lastGame.bonus) + ' this game.',
                        '',
                        'So far you have earned a total bonus of ' + this.formatDollars(this.server.bonus) + '.'];
    this.server.broadcastScreenState(this.state);
    this.server.lgScreen('score', {gnum: this.game.gnum, lines: this.state.lines});
};

Score.prototype.handleInputEvent = function(event) {
    if (event.type === 'gameKeyDown' && event.key === 4 && this.canContinue) {
        this.status[event.player] = 'ready';
        this.updateState();
        this.server.broadcastScreenState(this.state);
        if (this.readyToContinue()) {
            clearTimeout(this.continueTimer);
            this.continueTimer = setTimeout(this.server.nextScreen.bind(this.server), 350);
        }
    } else if (event.type === 'connection') {
        this.status[event.player] = event.value ? 'in-progress':'disconnected';
        this.updateState();
        this.server.broadcastScreenState(this.state);
    }
};

Score.prototype.cleanup = function() {
    ShowInformation.prototype.cleanup.call(this);
    clearTimeout(this.continueTimer);
    this.server.lg('end');
};


function Game(gnum, voiceChat, config) {
    this.screen_id = 'game';
    this.gnum = gnum;
    this.voiceChat = voiceChat;
    // There is currently only one config...
    if (config === 'shield+shell')
        this.configFn = game.shieldedConfig;
    else
        this.configFn = game.shieldedConfig;
}

Game.prototype = {};

Game.prototype.initHelper = function (server) {
    this.server = server;
    this.state = {screen: 'game',
                  ships: [{alive: true, x:0, y:0, angle:0, color: 0x00FFFF},
                          {alive: true, x:0, y:0, angle:0, color: 0xFF00FF}],
                  fortress: {alive: true, x:0, y:0, angle: 0},
                  missiles: [],
                  shells: [],
                  smallhex: {radius: 40, x: 0, y: 0, angle: 0, full: true},
                  bighex: {radius: 200, x: 0, y: 0, angle: 0, full: true},
                  points: 0,
                  vulnerability: 0
                 };
    if (this.server.debug) this.state.latency = [0,0];
    this.logStateData = [0, [0,0,0,0,0,0,0], [0,0,0,0,0,0,0], [0,0,0,0,0], [], [], 0, []];

    this.configFn(game.config);

    this.modelStateData = {tick: 0,
                           players: [{alive: false,
                                      pos: {x:0, y:0},
                                      vel: {x:0, y:0},
                                      angle: 0,
                                      thrustFlag: false,
                                      turnFlag: 0,
                                      insideBigHex: false},
                                     {alive: false,
                                      pos: {x:0, y:0},
                                      vel: {x:0, y:0},
                                      angle: 0,
                                      thrustFlag: false,
                                      turnFlag: 0,
                                      insideBigHex: false}],
                           fortress: {alive: false,
                                      vulnerable: false,
                                      pos: {x:0, y:0},
                                      angle: 0,
                                      target: 0},
                           bighex: { radius: game.config.fortress.activationRadius },
                           smallhex: { radius: game.config.fortress.smallhexRadius },
                           missiles: [],
                           shells: [],
                           points: 0,
                           events: []};
    this.game = new game.Game();
    this.updateScreenState();
    if (this.voiceChat)
        this.server.startAudioRecording(this.gnum);
    this.server.broadcastScreenState(this.state);
    this.server.lgScreen('game', {gnum: this.gnum,
                                  voiceChat: this.voiceChat,
                                  config: game.config,
                                  gameKeys: ['none', 'left', 'right', 'thrust', 'fire'],
                                  log: ['gameTicks',
                                        ['player1Alive', 'player1PosX', 'player1PosY', 'player1VelX', 'player1VelY', 'player1Angle', 'player1ThrustFlag', 'player1TurnFlag'],
                                        ['player2Alive', 'player2PosX', 'player2PosY', 'player2VelX', 'player2VelY', 'player2Angle', 'player2ThrustFlag', 'player2TurnFlag'],
                                        ['fortressAlive', 'fortressAngle', 'fortressTarget', 'fortressVulnerable','fortressExplosion'],
                                        ['missileID', 'missilePosX', 'missilePosY', 'missileAngle'],
                                        ['shellID', 'shellPosX', 'shellPosY', 'shellAngle'],
                                        'points', 'events']
                                 });
};

Game.prototype.init = function (server) {
    this.initHelper(server);
    this.requestUpdate();
};

Game.prototype.updateProjectiles = function(dest, src) {
    if (dest.length < src.length) {
        while (dest.length < src.length) {
            dest.push({x:0, y:0, angle: 0});
        }
    } else {
        dest.length = src.length;
    }
    for (let i=0; i<src.length; i++) {
        dest[i].x = src[i].position.x;
        dest[i].y = src[i].position.y;
        dest[i].angle = src[i].angle;
    }
};

Game.prototype.updateHexagon = function(dest, src) {
    dest.radius = src.radius;
    dest.x = src.position.x;
    dest.y = src.position.y;
};

Game.prototype.isExplosionVisible = function(timer) {
    return this.game.tick < timer;
};

Game.prototype.updateScreenState = function() {
    for (let i=0; i<this.game.players.length; i++) {
        this.state.ships[i].alive = this.game.players[i].alive;
        this.state.ships[i].x = this.game.players[i].position.x;
        this.state.ships[i].y = this.game.players[i].position.y;
        this.state.ships[i].angle = this.game.players[i].angle;
        if (!this.state.ships[i].alive) {
            this.state.ships[i].explosion = this.isExplosionVisible(this.game.players[i].respawnTimer);
        }
    }
    this.state.fortress.alive = this.game.fortress.alive;
    this.state.fortress.x = this.game.fortress.position.x;
    this.state.fortress.y = this.game.fortress.position.y;
    this.state.fortress.angle = this.game.fortress.angle;
    if (!this.game.fortress.alive)
        this.state.fortress.explosion = this.isExplosionVisible(this.game.fortress.respawnTimer);

    this.updateHexagon(this.state.bighex, this.game.bighex);
    this.updateHexagon(this.state.smallhex, this.game.smallhex);
    this.state.smallhex.angle = this.game.fortress.angle;

    this.state.smallhex.full = !this.game.fortress.vulnerable;

    this.updateProjectiles(this.state.missiles, this.game.missiles);
    this.updateProjectiles(this.state.shells, this.game.shells);

    this.state.points = this.game.points;

    this.server.broadcastScreenState(this.state);
};

Game.prototype.broadcastGameStateAsBuffer = function() {
    var len = 37 + (this.game.missiles.length+this.game.shells.length)*3*4 + 1;
    if (this.server.debug) len += 4*2;
    var buf = Buffer.allocUnsafe(len);

    var offset = buf.writeUInt16LE(this.game.points);
    offset = buf.writeUInt8(this.game.players[0].alive?1:0, offset);
    offset = buf.writeFloatLE(this.game.players[0].position.x, offset);
    offset = buf.writeFloatLE(this.game.players[0].position.y, offset);
    offset = buf.writeFloatLE(this.game.players[0].angle, offset);

    offset = buf.writeUInt8(this.game.players[1].alive?1:0, offset);
    offset = buf.writeFloatLE(this.game.players[1].position.x, offset);
    offset = buf.writeFloatLE(this.game.players[1].position.y, offset);
    offset = buf.writeFloatLE(this.game.players[1].angle, offset);

    offset = buf.writeUInt8(this.game.fortress.alive?1:0, offset);
    offset = buf.writeFloatLE(this.game.fortress.angle, offset);
    offset = buf.writeUInt8(this.game.fortress.vulnerable?0:1, offset);
    if (this.game.fortress.alive) {
        offset = buf.writeUInt8(0, offset);
    } else {
        offset = buf.writeUInt8(this.isExplosionVisible(this.game.fortress.respawnTimer)?1:0, offset);
    }

    offset = buf.writeUInt8(this.game.missiles.length, offset);
    offset = buf.writeUInt8(this.game.shells.length, offset);

    for (let i=0; i<this.game.missiles.length; i++) {
        offset = buf.writeFloatLE(this.game.missiles[i].position.x, offset);
        offset = buf.writeFloatLE(this.game.missiles[i].position.y, offset);
        offset = buf.writeFloatLE(this.game.missiles[i].angle, offset);
    }

    for (let i=0; i<this.game.shells.length; i++) {
        offset = buf.writeFloatLE(this.game.shells[i].position.x, offset);
        offset = buf.writeFloatLE(this.game.shells[i].position.y, offset);
        offset = buf.writeFloatLE(this.game.shells[i].angle, offset);
    }

    offset = buf.writeUInt8(this.server.debug?1:0, offset);
    if (this.server.debug) {
        offset = buf.writeUInt32LE(this.state.latency[0], offset);
        offset = buf.writeUInt32LE(this.state.latency[1], offset);
    }

    this.server.broadcastBuffer(buf);
}

Game.prototype.requestUpdate = function() {
    var now = new Date().getTime();
    if (this.lastUpdate) {
            var diff = now - this.lastUpdate;
            var compensation = diff - this.ms;
            this.ms = 1000/60 - compensation;
            if (this.ms < 0) this.ms = 0;
        } else {
            this.ms = 1000/60;
        }
    this.lastUpdate = now;
    this.updateid = setTimeout(this.update.bind(this), this.ms);
};

Game.prototype.updateProjectileLog = function(logProj, proj) {
    if (logProj.length > proj.length)
        logProj.length = proj.length;
    else if (logProj.length < proj.length) {
        for (let i=logProj.length; i<proj.length; i++) {
            logProj.push([0,0,0,0]);
        }
    }
    for (let i=0; i<proj.length; i++) {
        logProj[i][0] = proj[i].id;
        logProj[i][1] = proj[i].position.x;
        logProj[i][2] = proj[i].position.y;
        logProj[i][3] = proj[i].angle;
    }
}

Game.prototype.modelStateUpdateProjectiles = function(modelProj, proj) {
    if (modelProj.length > proj.length)
        modelProj.length = proj.length;
    else if (modelProj.length < proj.length) {
        for (let i=modelProj.length; i<proj.length; i++) {
            modelProj.push({id: 0,
                            pos: {x:0,y:0},
                            vel: {x:0,y:0},
                            angle: 0});
        }
    }
    for (let i=0; i<proj.length; i++) {
        modelProj[i].id = proj[i].id;
        modelProj[i].pos.x = proj[i].position.x;
        modelProj[i].pos.y = proj[i].position.y;
        modelProj[i].vel.x = proj[i].velocity.x;
        modelProj[i].vel.y = proj[i].velocity.y;
        modelProj[i].angle = proj[i].angle;
    }
}

Game.prototype.getModelState = function() {
    this.modelStateData.tick = this.game.tick;
    for (let i=0; i<this.game.players.length; i++) {
        this.modelStateData.players[i].alive = this.game.players[i].alive ? true:false;
        this.modelStateData.players[i].pos.x = this.game.players[i].position.x;
        this.modelStateData.players[i].pos.y = this.game.players[i].position.y;
        this.modelStateData.players[i].vel.x = this.game.players[i].velocity.x;
        this.modelStateData.players[i].vel.y = this.game.players[i].velocity.y;
        this.modelStateData.players[i].angle = this.game.players[i].angle;
        this.modelStateData.players[i].thrustFlag = this.game.players[i].thrustFlag ? true:false;
        this.modelStateData.players[i].turnFlag = this.game.players[i].turnFlag || 0;
        this.modelStateData.players[i].insideBigHex = this.game.activationHex.inside(this.game.activationHex.position, this.game.players[i].position) ? true:false;
    }
    this.modelStateData.fortress.alive = this.game.fortress.alive ? true:false;
    this.modelStateData.fortress.angle = this.game.fortress.angle;
    this.modelStateData.fortress.target = this.game.fortress.target ? this.game.players.indexOf(this.game.fortress.target) : null;
    this.modelStateData.fortress.vulnerable = this.game.fortress.vulnerable ? true:false;

    this.modelStateUpdateProjectiles(this.modelStateData.missiles, this.game.missiles);
    this.modelStateUpdateProjectiles(this.modelStateData.shells, this.game.shells);

    this.modelStateData.points = this.game.points;

    this.modelStateData.events.length = this.game.events.length;
    for (let i=0; i<this.game.events.length; i++)
        this.modelStateData.events[i] = this.game.events[i];

    return this.modelStateData;
};


Game.prototype.logState = function() {
    this.logStateData[0] = this.game.tick;
    for (let i=0; i<this.game.players.length; i++) {
        this.logStateData[i+1][0] = this.game.players[i].alive ? 1:0;
        this.logStateData[i+1][1] = this.game.players[i].position.x;
        this.logStateData[i+1][2] = this.game.players[i].position.y;
        this.logStateData[i+1][3] = this.game.players[i].velocity.x;
        this.logStateData[i+1][4] = this.game.players[i].velocity.y;
        this.logStateData[i+1][5] = this.game.players[i].angle;
        this.logStateData[i+1][6] = this.game.players[i].thrustFlag ? 1:0;
        this.logStateData[i+1][7] = this.game.players[i].turnFlag || 0;
    }
    this.logStateData[3][0] = this.game.fortress.alive ? 1:0;
    this.logStateData[3][1] = this.game.fortress.angle;
    this.logStateData[3][2] = this.game.fortress.target ? this.game.players.indexOf(this.game.fortress.target) : null;
    this.logStateData[3][3] = this.game.fortress.vulnerable ? 1:0;
    this.logStateData[3][4] = !this.game.fortress.alive && this.game.tick < this.game.fortress.respawnTimer ? 1:0;

    this.updateProjectileLog(this.logStateData[4], this.game.missiles);
    this.updateProjectileLog(this.logStateData[5], this.game.shells);

    this.logStateData[6] = this.game.points;
    // this.logStateData[7] = this.game.vulnerability;
    this.logStateData[7] = this.game.events;

    this.server.lg('f', {'s': this.logStateData});
};

Game.prototype.updateHelper = function() {
    this.game.stepOneTick();
    //this.updateScreenState();
    this.broadcastGameStateAsBuffer();
    this.logState();
};

Game.prototype.update = function() {
    this.updateHelper();
    if (this.game.isFinished()) {
        // console.log('baaang its done!');
        this.server.nextScreen();
    } else {
        this.requestUpdate();
    }
};

Game.prototype.cancelUpdates = function() {
    clearTimeout(this.updateid);
    this.updateid = null;
};

Game.prototype.cleanup = function() {
    if (this.voiceChat)
        this.server.stopAudioRecording();
    this.game.calculateBonus();
    this.server.bonus += this.game.bonus;
    this.server.lastGame.bonus = this.game.bonus;
    this.server.lastGame.points = this.game.points;
    this.cancelUpdates();
    this.server.lg('end', {'points': this.game.points,
                           'rawPoints': this.game.rawPoints,
                           'bonus': this.game.bonus,
                           'total': this.server.bonus});
};

Game.prototype.handleInputEvent = function(event) {
    if (event.type === 'gameKeyDown') {
        this.game.pressKey(event.player, event.key);
    } else if (event.type === 'gameKeyUp') {
        this.game.releaseKey(event.player, event.key);
    } else if (event.type === 'pong') {
        if (this.server.debug) this.state.latency[event.player] = event.latency;
    } else if (event.type === 'connection') {
        console.log('got connection', event);
        if (this.server.areAllClientsConnected()) {
            this.state.disconnected = undefined;
            this.lastUpdate = undefined;
            this.requestUpdate();
            if (this.voiceChat)
                this.server.resumeAudioRecording();
        } else {
            this.cancelUpdates();
            this.updateScreenState();
            this.state.disconnected = true;
            this.server.broadcastScreenState(this.state);
            if (this.voiceChat)
                this.server.pauseAudioRecording();
        }
    }
};

//

function Chat() {
    this.screen_id = 'chat';
}

Chat.prototype = {};

Chat.prototype.init = function (server) {
    this.server = server;
    this.history = [[],[]];
    this.playerText = ['', ''];
    this.playerStartedTyping = [false, false];
    this.status = ['in-progress', 'in-progress'];
    this.playerState = ['text', 'text'];
    this.finishedState = {screen: 'message',
                          lines: ["Waiting for other player"],
                          continue: ''};
    this.state = [{screen: 'chat',
                   history: [],
                   otherPlayerStartedTyping: false,
                   otherPlayerTyping: false,
                   text: ''},
                  {screen: 'chat',
                   history: [],
                   otherPlayerStartedTyping: false,
                   otherPlayerTyping: false,
                   text: ''}];
    this.server.broadcastClientCommand('textInput', {state: true});
    this.updateAndSendState();
    this.server.lgScreen('chat');
};

Chat.prototype.updateAndSendState = function () {
    for (let i=0; i<this.state.length; i++) {
        this.state[i].focus = this.playerState[i];
        this.state[i].history = this.history[i];
        this.state[i].text = this.playerText[i];
        this.state[(i+1)%2].otherPlayerStartedTyping = this.playerStartedTyping[i];
        this.state[(i+1)%2].otherPlayerTyping = this.state[i].text.length > 0;
    }
    for (let i=0; i<this.state.length; i++) {
        if (this.playerState[i] === 'finished')
            this.server.sendScreenState(i, this.finishedState);
        else
            this.server.sendScreenState(i, this.state[i]);
    }
}

Chat.prototype.addHistory = function (player, text) {
    this.playerStartedTyping[player] = this.playerText[player].length > 0;
    this.playerStartedTyping[(player+1)%2] = this.playerText[(player+1)%2].length > 0;
    this.history[player].push({'column': 0, 'text': text});
    this.history[(player+1)%2].push({'column': 1, 'text': text});
};

Chat.prototype.addStatusText = function (player, text) {
    this.history[(player+1)%2].push({'type': 'status', 'column': 1, 'text': text});
};

Chat.prototype.readyToContinue = function () {
    for (let i=0; i<this.playerState.length; i++)
        if (this.playerState[i] !== 'finished') return false;
    return true;
}

Chat.prototype.handleInputEvent = function (event) {
    var update = false;
    if (this.playerState === 'finished') {
    } else {
        if (event.type === 'keyDown') {
            if (event.name == 'Return') {
                if (this.playerState[event.player] === 'text') {
                    if (this.state[event.player].text.length > 0) {
                        var tmp = this.playerText[event.player];
                        this.playerText[event.player] = '';
                        this.playerStartedTyping[event.player] = false;
                        this.addHistory(event.player, tmp);
                        this.server.lg('send-message', {player: event.player, text: tmp});
                        update = true;
                    }
                } else if (this.playerState[event.player] === 'continue') {
                    this.addStatusText(event.player, 'Player has left the chat.');
                    this.playerState[event.player] = 'finished';
                    this.playerText[event.player] = '';
                    this.server.lg('leave-chat', {player: event.player});
                    update = true;
                    if (this.readyToContinue()) {
                        clearTimeout(this.continueTimer);
                        this.continueTimer = setTimeout(this.server.nextScreen.bind(this.server), 350);
                    }
                }
            } else if (event.name == 'Backspace') {
                this.playerText[event.player] = this.playerText[event.player].slice(0, this.playerText[event.player].length-1);
                update = true;
            } else if (event.name == 'Up') {
                if (this.playerState[event.player] == 'continue') {
                    this.playerState[event.player] = 'text';
                    update = true;
                }
            } else if (event.name == 'Down') {
                if (this.playerState[event.player] == 'text') {
                    this.playerState[event.player] = 'continue';
                    update = true;
                }
            } else if (event.name == 'Tab') {
                if (this.playerState[event.player] == 'text') {
                    this.playerState[event.player] = 'continue';
                } else {
                    this.playerState[event.player] = 'text';
                }
                update = true;
            }
        } else if (event.type === 'textInput') {
            if (this.playerState[event.player] == 'text') {
                this.playerText[event.player] += event.text;
                this.playerStartedTyping[event.player] = true;
                update = true;
            }
        }
    }
    if (event.type === 'connection') {
        if (event.value) {
            this.server.broadcastClientCommand('textInput', {state: true});
            update = true;
        }
    }


    if (update) this.updateAndSendState();
};

Chat.prototype.cleanup = function () {
    this.server.broadcastClientCommand('textInput', {state: false});
    this.server.lg('end');
}

function AudioTest () {
    this.screen_id = 'audiotest';
    this.state = {screen: "message",
                  lines: [""],
                  continue: ['', '']};
}

AudioTest.prototype = {};

AudioTest.prototype.init = function (server) {
    this.server = server;
    this.status = ['in-progress', 'in-progress'];
    this.server.broadcastScreenState({screen: 'clientCommand', cmd: 'audiotest'});
    this.server.lgScreen('audiotest');
}

AudioTest.prototype.readyToContinue = function() {
    for (let i=0; i<this.status.length; i++)
        if (this.status[i] !== 'ready') return false;
    return true;
};

AudioTest.prototype.handleInputEvent = function(event) {
    if (event.type === 'audioTestComplete') {
        this.status[event.player] = 'ready';
        this.server.sendScreenState(event.player, {screen: "message",
                                                   lines: ["Waiting for other player"],
                                                   continue: ''});
        if (this.readyToContinue()) {
            clearTimeout(this.continueTimer);
            this.continueTimer = setTimeout(this.server.nextScreen.bind(this.server), 350);
        }
    } else if (event.type === 'connection') {
        this.status[event.player] = event.value ? 'in-progress':'disconnected';
        if (event.value) {
            this.server.sendScreenState(event.player,
                                        {screen: 'clientCommand', cmd: 'audiotest'});
        }
    }
}

AudioTest.prototype.cleanup = function () {
    this.server.lg('end');
}


exports.Message = Message;
exports.Score = Score;
exports.Game = Game;
exports.Consent = Consent;
exports.Demographics = Demographics;
exports.Feedback = Feedback;
exports.Instructions = Instructions;
exports.GameStart = GameStart;
exports.Intermission = Intermission;
exports.Chat = Chat;
exports.End = End;
exports.AudioTest = AudioTest;
