var util = require('./util');
var game = require('./game');
var V = util.V;
var rad2deg = util.rad2deg;
var deg2rad = util.deg2rad;
var angle_diff = util.angle_diff;
var rotateInto = util.rotateInto;
var dotProduct = util.dotProduct;

var FIRE = game.FIRE; //'fire';
var THRUST = game.THRUST; //'thrust';
var LEFT = game.LEFT; //'left';
var RIGHT = game.RIGHT; //'right';

// Coop Space Fortress model 1.19:
// * this model is a modification of model 1.17 that:
//	- slightly increases desired speed and desired speed inside
//	- slightly improves aim precision
//	- 
// developped by Cvetomir Dimov


/*
- shell detection is very insensitive
*/

// constants that determine size of elements in the playing field
var maxX = 350;
var maxY = 310;
var hexRadius = 180;
var fortressRadius = 40;
var tickLength = 1000/60; //duration of tick in milliseconds
var turnRate = 4; //degrees/tick
var acceleration = 0.05; //pixels/tick^2

// all functions required by model
function normalizeAngle(angle){
	if (angle < 0) angle += 360;
	else if (angle >= 360) angle -= 360;
	return angle;
}

function vector_angle(vector){
	return normalizeAngle(rad2deg(Math.atan2(vector.y,vector.x)));
}


function lineCirlceIntersection(vec1,vec2,r){
	let dx = vec2.x - vec1.x;
	let dy = vec2.y - vec1.y;
	let dr = Math.sqrt(Math.pow(dx,2)+Math.pow(dy,2));
	let D = vec1.x * vec2.y - vec2.x * vec1.y;
	let det = Math.pow(r,2) * Math.pow(dr,2) - Math.pow(D,2);
	let p1 = V((D * dy + Math.sign(dy) * dx * Math.sqrt(det))/Math.pow(dr,2), (- D * dx + Math.abs(dy) * Math.sqrt(det))/Math.pow(dr,2));
	let p2 = V((D * dy - Math.sign(dy) * dx * Math.sqrt(det))/Math.pow(dr,2), (- D * dx - Math.abs(dy) * Math.sqrt(det))/Math.pow(dr,2));
	if (det < 0) {
		return [undefined, undefined];
	} else {
		return [p1, p2];
	}
}

function approachingPointP(pos,vel,point){
	let nextPos = pos.duplicate().add(vel);
	let distanceCurrent = (point.duplicate().sub(pos)).norm();
	let distanceNext = (point.duplicate().sub(nextPos)).norm();
	return distanceNext < distanceCurrent;
}

function timeToIntersect(pos,vel,radius){
	let start = V(pos.x + 10000*vel.x,pos.y + 10000*vel.y);
	let end = V(pos.x - 10000*vel.x,pos.y - 10000*vel.y);
	let [intersection1, intersection2] = lineCirlceIntersection(start,end,radius);
	if (intersection1===undefined){
		return Infinity;
	} else if (approachingPointP(pos,vel,intersection1) && approachingPointP(pos,vel,intersection2)) {
		if (intersection1.duplicate().sub(pos).norm() < intersection2.duplicate().sub(pos).norm()) {
			return intersection1.duplicate().sub(pos).norm()/vel.norm();
		} else {
			return intersection2.duplicate().sub(pos).norm()/vel.norm();
		}
	} else if (approachingPointP(pos,vel,intersection1)){
		return intersection1.duplicate().sub(pos).norm()/vel.norm();
	} else if (approachingPointP(pos,vel,intersection2)){
		return intersection2.duplicate().sub(pos).norm()/vel.norm();
	} else {
		return Infinity;
	}
}

function anglesWithin(angle1,angle2,halfRange){
	let angleDifference = angle_diff(angle2,angle1);
	return Math.abs(angleDifference) < halfRange;
}

function angleTowardFortress(position){
	return normalizeAngle(vector_angle(position) + 180);
}


// constants required by the agent
let timeThresh = 100; // in units of time ticks; 75 ticks = 1.25s 
let slowSpeed = 0.2; // in pixels/tick
let aimPrecision = 4.5; // degrees
let thrustPrecision = 15; //degrees
let desiredSpeed = 0.75; // in pixels/tick
let desiredSpeedInside = 0.65; // in pixels/tick

// functions required by agent
desiredVelocity = function(angle){
	return V(Math.cos(deg2rad(angle)), Math.sin(deg2rad(angle))).scalar_mult(desiredSpeed);
}

slowVelocity = function(angle){
	return V(Math.cos(deg2rad(angle)), Math.sin(deg2rad(angle))).scalar_mult(slowSpeed);
}

// agent model
function Agent(game, player){
    this.currentAction = {pressUntil: 0, keyPressed: false};
	this.goal = 'normal';
    this.actionStack = [];
    this.game = game;
    this.player = player;
}

Agent.prototype = {}

// parameters that determine agent profficiency
Agent.prototype.perceptualDelay = 50/tickLength; // ms
Agent.prototype.motorDelay = 250/tickLength; // in ms; to model preparation of motor actions
Agent.prototype.motorPrepDelay = 150/tickLength; // in ms; to model preparation of motor actions
Agent.prototype.cognitiveDelay = 50/tickLength; // in ms; to model non-motor processes
Agent.prototype.speedThreshold = 0.1; // in pixels/tick
Agent.prototype.delayNoise = 0.33; // presumably, there would be less noise added to all non-motor processes than the huge amount of motor noise
Agent.prototype.motorNoise = 0.2; // as a fraction of motor duration; affects thrust duration and turn duration

Agent.prototype.angleTowardFortress = function(){
	return angleTowardFortress(this.player.position);
}

Agent.prototype.angleTowardOutside = function(){
	let angle = vector_angle(this.player.position);
	return normalizeAngle(angle);
}

Agent.prototype.angleAvoidShell = function(){
	let avoidAngle;
	if  (this.movingClockwiseP()) {
		avoidAngle = normalizeAngle(this.angleTowardFortress() + 60);
	} else {
		avoidAngle = normalizeAngle(this.angleTowardFortress() - 60);
	}
	return avoidAngle;
}

Agent.prototype.angleAvoidExit = function(){
	let avoidAngle;
	if  (this.movingClockwiseP()) {
		avoidAngle = normalizeAngle(this.angleTowardFortress() + 70);
	} else {
		avoidAngle = normalizeAngle(this.angleTowardFortress() - 70);
	}
	return avoidAngle;
}

Agent.prototype.velocityAvoidExit = function(){
	return this.slowVelocity(this.angleAvoidExit());
}

Agent.prototype.velocityTowardInside = function(){
	return desiredVelocity(this.angleTowardFortress());
}

Agent.prototype.entryVelocity = function(){
	return desiredVelocity(this.angleTowardFortress()).scalar_mult(0.6);
}

Agent.prototype.velocityTowardOutside = function(){
	return desiredVelocity(this.angleTowardOutside());
}

Agent.prototype.velocityAvoidShell = function(){
	return slowVelocity(this.angleAvoidShell());
}

// -----------------------------------
// functions for computing agent state
// -----------------------------------
Agent.prototype.closeToOuterP = function(){
	pos = this.player.position;
	vel = this.player.velocity;
	let timeX = 100;
	let timeY = 100;	
	if (vel.x < 0){
		timeX = (pos.x + maxX)/(-vel.x);
	} else {
		timeX = (maxX - pos.x)/vel.x;
	}
	if (vel.y < 0){
		timeY = (pos.y + maxY)/(-vel.y);
	} else {
		timeY = (maxY - pos.y)/vel.y;
	}
	return (timeX < timeThresh) || (timeY < timeThresh);
}

Agent.prototype.closeToFortressP = function(fortress){
	// avoid fortress if close to it, it is alive, and it is not vulnerable + a target
	pos = this.player.position;
	vel = this.player.velocity;
	return (timeToIntersect(pos,vel,fortressRadius) < timeThresh) && fortress.alive && !(fortress.vulnerable && this.behindFortressP(fortress.angle));
}

Agent.prototype.willIntersectHexP = function(){
	pos = this.player.position;
	vel = this.player.velocity;
	return timeToIntersect(pos,vel,hexRadius) != Infinity;
}

Agent.prototype.aboutToCrossHexP = function(){
	pos = this.player.position;
	vel = this.player.velocity;
	return pos.norm() < this.hexRadius && timeToIntersect(pos,vel,this.hexRadius) < this.timeThresh;
}

behindFortressP = function(playerPosition,fortressAngle){
	return anglesWithin(angleTowardFortress(playerPosition),fortressAngle,45);	
}

Agent.prototype.behindFortressP = function(fortressAngle){
	return behindFortressP(this.player.position,fortressAngle);	
}

Agent.prototype.movingBehindFortressP = function(fortressAngle){
	//  && this.behindFortressP()
	if (angle_diff(Math.abs(angle_diff(this.angleTowardFortress(),fortressAngle)),Math.abs(angle_diff(angleTowardFortress(this.player.position.duplicate().add(this.player.velocity)),fortressAngle))) > 0  && !this.playerSlow() && this.behindFortressP()) {
		return true;
	} else {
		return false;
	}	
}

Agent.prototype.approachingShellP = function(shellPosition,shellVelocity){
	let shellNearby = this.player.position.duplicate().sub(shellPosition).norm() < 60;
	let shellPerpendicular = V(0,0);
	rotateInto(90,shellVelocity,shellPerpendicular);
	let perpendicularVelocity = Math.abs(dotProduct(this.player.velocity,shellPerpendicular)/shellPerpendicular.norm());
	let shellApproaching = approachingPointP(shellPosition,shellVelocity,this.player.position);
	return shellNearby && shellApproaching && (perpendicularVelocity < 0.1);
}

Agent.prototype.approachingShellsP = function(shells,insideHex){
	if (shells.length==0){
		return false;
	} else {
		let truthValues = [];
		var i;
		for (i = 0; i < shells.length; i++){
			truthValues.push(this.approachingShellP(shells[i].position,shells[i].velocity));
		}
		let cumulativeTruthValue = false;
		for (i = 0; i < truthValues.length; i++){
			cumulativeTruthValue = cumulativeTruthValue || truthValues[i];
		}
		return cumulativeTruthValue && insideHex;
	}
}

Agent.prototype.willLeaveFortressBackP = function(fortressAngle){
	playerFuturePosition = this.player.position.duplicate().add(this.player.velocity.duplicate().scalar_mult(30)); // player position in half a second
	return this.behindFortressP(fortressAngle) && !behindFortressP(playerFuturePosition,fortressAngle);
}

Agent.prototype.playerSlow = function(){
	return this.player.velocity.norm() < 1.5 * slowSpeed;
}

Agent.prototype.playerSlowish = function(){
	return this.player.velocity.norm() < 2.5 * slowSpeed;
}

Agent.prototype.goodDirectionOutside = function(outsideVelocity){
	let headingOut = anglesWithin(vector_angle(outsideVelocity),vector_angle(this.player.velocity),30) && !this.playerSlow();
	return headingOut;// || !this.playerSlowish();
}

Agent.prototype.movingClockwiseP = function(){
	return (vector_angle(this.player.position) - vector_angle(this.player.position.duplicate().add(this.player.velocity))) > 0;
}

Agent.prototype.withinRange = function(missiles){
	return this.player.position.norm() < 160 && missiles.length == 0;
}

Agent.prototype.inDanger = function(fortress,shells,insideHex){
	return this.closeToFortressP(fortress) || this.approachingShellsP(shells,insideHex) || this.closeToOuterP();
}

// --------------------------------
// functions for executing actions
// --------------------------------
uniformNoise = function(){
	return 2*(Math.random()-0.5);
}

Agent.prototype.keyPressed = function(){
    return this.currentAction.keyPressed;
};

Agent.prototype.pressKey = function(key,duration,currentTime){
    if (key != "nothing"){
	this.game.pressKey(this.game.getPlayerIndex(this.player), key);
    }
    this.currentAction.keyPressed = key;
    this.currentAction.pressUntil = currentTime + duration;
};

Agent.prototype.releaseKey = function(currentTime) {
    if (this.currentAction.keyPressed != false){
		if (currentTime >= this.currentAction.pressUntil){
                    if (this.currentAction.keyPressed !== "nothing") {
			this.game.releaseKey(this.game.getPlayerIndex(this.player), this.currentAction.keyPressed);
                    }
			this.currentAction.pressUntil = 0;
			this.currentAction.keyPressed = false;
		}
    }
};

Agent.prototype.emergencyReleaseKey = function(currentTime) {
    if (this.currentAction.keyPressed != false){
        if (this.currentAction.keyPressed !== "nothing") {
	    this.game.releaseKey(this.game.getPlayerIndex(this.player), this.currentAction.keyPressed);
        }
	this.currentAction.pressUntil = 0;
	this.currentAction.keyPressed = false;
    }
};

Agent.prototype.actionsWaitingP = function(){
    return this.actionStack.length > 0;
};

Agent.prototype.getNextAction = function(){
    return this.actionStack.shift();
};

Agent.prototype.executeAction = function(currentTime){
    let action = this.getNextAction();
    this.pressKey(action[0],action[1],currentTime);
}

Agent.prototype.appendPerceptualCost = function(){
	this.actionStack.push(["nothing",this.perceptualDelay*(1+this.delayNoise*uniformNoise())]);
}

Agent.prototype.appendCognitiveCost = function(){
	this.actionStack.push(["nothing",this.cognitiveDelay*(1+this.delayNoise*uniformNoise())]);
}

Agent.prototype.appendMotorCost = function(){
	this.actionStack.push(["nothing",this.motorDelay*(1+this.delayNoise*uniformNoise())]);
}

Agent.prototype.appendPrepMotorCost = function(){
	this.actionStack.push(["nothing",this.motorPrepDelay*(1+this.delayNoise*uniformNoise())]);
}

Agent.prototype.appendAction = function(action){
	action[1] = Math.round(Math.max(action[1]*(1+this.motorNoise*uniformNoise()) + uniformNoise(),1));
    this.actionStack.push(action);
}

Agent.prototype.clearActions = function(){
    if (this.currentAction.keyPressed !== "nothing") {
	this.game.releaseKey(this.game.getPlayerIndex(this.player), this.currentAction.keyPressed);
    }
    this.currentAction.pressUntil = 0;
    this.currentAction.keyPressed = false;
    this.actionStack = [];
}

Agent.prototype.turnTowards = function(desiredAngle){
    let turnAngle = normalizeAngle(desiredAngle - this.player.angle);
	this.appendMotorCost();
	if ((turnAngle < 358) || (turnAngle > 2)){
		if (turnAngle > 180) {
			this.appendAction([RIGHT,Math.abs(360-turnAngle)/turnRate]);
		} else {
			this.appendAction([LEFT,turnAngle/turnRate]);
		}
	}
}

Agent.prototype.thrust = function(duration){
	this.appendAction([THRUST,duration]);
}

Agent.prototype.changeVelocity = function(desiredVelocity){
	this.appendPerceptualCost();
	this.appendCognitiveCost();
    let thrustVelocity = desiredVelocity.duplicate().sub(this.player.velocity);
    let thrustAngle = vector_angle(thrustVelocity);
    if (thrustVelocity.norm() > 0.1){
//		if (!anglesWithin(this.player.angle,thrustAngle,thrustPrecision)) {
		this.turnTowards(thrustAngle);
		this.appendPrepMotorCost();
//		}
		this.thrust(thrustVelocity.norm()/acceleration);
    }
}

Agent.prototype.thrustAction = function(){
	this.appendPerceptualCost();
	this.appendCognitiveCost();
	this.appendMotorCost();
	this.thrust(desiredSpeed/acceleration);
}

Agent.prototype.aimAndFire = function(fortressAngle,fortressVulnerable){
	this.appendPerceptualCost();
	this.appendCognitiveCost();
	let fortressAngle2;
	if (!this.playerSlow()){
		if (this.movingClockwiseP()) {
			fortressAngle2 = normalizeAngle(vector_angle(this.player.position) + 180 - 5)
		} else {
			fortressAngle2 = normalizeAngle(vector_angle(this.player.position) + 180 + 5);
		}
	} else {
		fortressAngle2 = normalizeAngle(vector_angle(this.player.position) + 180);
	}
	if (!anglesWithin(this.player.angle,fortressAngle2,aimPrecision)){
		this.turnTowards(fortressAngle2);
		this.appendPrepMotorCost();
	}
	if (fortressVulnerable) {
		this.appendAction([FIRE,1]);
	}
}

Agent.prototype.aim = function(fortressAngle){
	this.appendPerceptualCost();
	this.appendCognitiveCost();
	this.turnTowards(fortressAngle);
}

Agent.prototype.stop = function(){
    this.changeVelocity(V(0,0));
}

Agent.prototype.slowDown = function(){
	this.changeVelocity(this.player.velocity.duplicate().scalar_mult(slowSpeed/this.player.velocity.norm()));
}

Agent.prototype.moveBehindFortress = function(fortressAngle){
	let fortressBackAngle = fortressAngle + 180;
	let agentAngle = normalizeAngle(vector_angle(this.player.position));
	let desiredLocation;
	if (anglesWithin(agentAngle,normalizeAngle(fortressBackAngle),70)){
		desiredLocation = V(120*Math.cos(deg2rad(normalizeAngle(fortressBackAngle))),120*Math.sin(deg2rad(normalizeAngle(fortressBackAngle))));
	} else if (angle_diff(agentAngle,normalizeAngle(fortressBackAngle)) < 0) {
		desiredLocation = V(120*Math.cos(deg2rad(normalizeAngle(agentAngle+90))),120*Math.sin(deg2rad(normalizeAngle(agentAngle+90))));
	} else {
		desiredLocation = V(120*Math.cos(deg2rad(normalizeAngle(agentAngle-90))),120*Math.sin(deg2rad(normalizeAngle(agentAngle-90))));
	}
	let desiredAngle = vector_angle(desiredLocation.duplicate().sub(this.player.position));
	let desiredVelocity = V(desiredSpeedInside*Math.cos(deg2rad(desiredAngle)),desiredSpeedInside*Math.sin(deg2rad(desiredAngle)));
	this.changeVelocity(desiredVelocity);
}

Agent.prototype.emergencyChangeVelocity = function(emergencyVelocity){
	this.emergencyReleaseKey();
	this.clearActions();
	this.goal='avoiding';
	if (anglesWithin(vector_angle(emergencyVelocity),this.player.angle,30)){
		this.thrustAction();
	} else {
		this.changeVelocity(emergencyVelocity);
	}
}

Agent.prototype.update = function() {
    if (this.game.tick > 120) {
        var other_player = this.game.players[(this.game.getPlayerIndex(this.player)+1)%this.game.players.length];
	this.releaseKey(this.game.tick);
	if (this.player.alive===false) {
	    this.clearActions();
            // this.game.addEvent({tag: 'ai-actions-cleared', player: this.game.getPlayerIndex(this.player)});
	} else if (this.inDanger(this.game.fortress,this.game.shells,this.game.activationHex.inside(this.game.activationHex.position, this.player.position)) && this.goal!='avoiding') {
	    // an emergency actions need to be undertaken
	    if (this.closeToFortressP(this.game.fortress)) {
		this.emergencyChangeVelocity(this.velocityTowardOutside());
                this.game.addEvent({tag: 'ai-avoid-fortress', player: this.game.getPlayerIndex(this.player)});
	    } else if (this.approachingShellsP(this.game.shells,this.game.activationHex.inside(this.game.activationHex.position, this.player.position))) {
		if (this.player.velocity.duplicate().sub(this.velocityAvoidShell()).norm() < slowSpeed) {
		    this.emergencyChangeVelocity(this.velocityAvoidShell());
                    this.game.addEvent({tag: 'ai-avoid-shell', player: this.game.getPlayerIndex(this.player)});
		}
	    } else if (this.closeToOuterP()) {
		this.emergencyChangeVelocity(this.velocityTowardInside());
                this.game.addEvent({tag: 'ai-avoid-outer', player: this.game.getPlayerIndex(this.player)});
	    } 
	} else if (!this.keyPressed()) {
	    if (this.actionsWaitingP()) {
		//execute actions if any actions are waiting to be executed
		this.executeAction(this.game.tick);
	    } else {
		this.goal='normal';
		if (!this.game.activationHex.inside(this.game.activationHex.position, this.player.position)) {
		    // if outside hexagon
		    if (this.game.fortress.alive) {
			// and fortress alive
			if (!this.willIntersectHexP() || this.playerSlow()) {
			    this.changeVelocity(this.entryVelocity());
                            this.game.addEvent({tag: 'ai-move-inside', player: this.game.getPlayerIndex(this.player)});
			} 
		    } else if (!this.game.fortress.alive) {
			// and fortress not alive
			if (!this.playerSlow()) {
			    this.stop();
                            this.game.addEvent({tag: 'ai-stop-outside', player: this.game.getPlayerIndex(this.player)});
			} else if (this.playerSlow()){
			    if (!anglesWithin(this.player.angle,this.angleTowardFortress(),30)){
				this.aim(this.angleTowardFortress());
                                this.game.addEvent({tag: 'ai-turn-torwards-inside', player: this.game.getPlayerIndex(this.player)});
			    }
			}
		    }
		} else if (this.game.activationHex.inside(this.game.activationHex.position, this.player.position)) {
		    // if inside hexagon
		    if (this.game.fortress.alive) {
			// and fortress alive
			if (this.aboutToCrossHexP()) {
			    //	this.changeVelocity(this.velocityTowardInside().scalar_mult(0.4));
			    this.game.changeVelocity(this.game.velocityAvoidExit());
                            this.game.addEvent({tag: 'ai-avoiding-exit', player: this.game.getPlayerIndex(this.player)});
			} else if (this.game.fortress.target===this.player) {
			    if (this.player.position.norm() < 80 && this.player.velocity.duplicate().sub(this.velocityTowardOutside().scalar_mult(0.5)).norm() > 0.2) {	
				this.changeVelocity(this.velocityTowardOutside().scalar_mult(0.4));
                                this.game.addEvent({tag: 'ai-move-away-from-fortress', player: this.game.getPlayerIndex(this.player)});
			    } else {
				if (!this.playerSlow()) {
				    this.changeVelocity(this.velocityAvoidShell());
                                    this.game.addEvent({tag: 'ai-slowly-move-sideways-to-missiles', player: this.game.getPlayerIndex(this.player)});
				} else if (this.playerSlow()) {
				    if (!anglesWithin(this.angleAvoidShell(),this.player.angle,30)) {
					this.aim(this.angleAvoidShell());
                                        this.game.addEvent({tag: 'ai-prepare-to-avoid', player: this.game.getPlayerIndex(this.player)});
				    }
				}
			    }
			} else if (this.game.fortress.target===other_player) {
			    if (this.behindFortressP(this.game.fortress.angle)) {
				if (!this.game.fortress.vulnerable){
				    if (!this.playerSlow() && this.withinRange(this.game.missiles)) {
					this.stop();
                                        this.game.addEvent({tag: 'ai-stop-behind-fortress', player: this.game.getPlayerIndex(this.player)});
				    } else if (this.willLeaveFortressBackP(this.game.fortress.angle)) {
					this.moveBehindFortress(this.game.fortress.angle); 
                                        this.game.addEvent({tag: 'ai-stay-behind-fortress', player: this.game.getPlayerIndex(this.player)});
				    } else {
					if (!anglesWithin(this.angleTowardFortress(),this.player.angle,30)) {
					    this.aim(this.angleTowardFortress());
                                            this.game.addEvent({tag: 'ai-aim', player: this.game.getPlayerIndex(this.player)});
					}
				    }
				} else {
				    if (this.withinRange(this.game.missiles)){
					this.aimAndFire(this.angleTowardFortress(),this.game.fortress.vulnerable);
                                        this.game.addEvent({tag: 'ai-aim-and-fire', player: this.game.getPlayerIndex(this.player)});
				    }
				}
			    } else {
				if (!this.movingBehindFortressP(this.game.fortress.angle)) {
				    this.moveBehindFortress(this.game.fortress.angle); 
                                    this.game.addEvent({tag: 'ai-move-behind-fortress', player: this.game.getPlayerIndex(this.player)});
				} else {
				    if (!anglesWithin(this.angleTowardFortress(),this.player.angle,30)){
					this.aim(this.game.fortress.angle);
                                        this.game.addEvent({tag: 'ai-aim-towards-fortress-behind', player: this.game.getPlayerIndex(this.player)});
				    }
				}
			    } 
			}
		    } else if (!this.game.fortress.alive) {
			// and fortress not alive
			if (!this.goodDirectionOutside(this.velocityTowardOutside())) {
                            this.game.addEvent({tag: 'ai-move-outside', player: this.game.getPlayerIndex(this.player)});
			    this.changeVelocity(this.velocityTowardOutside());
			} else if (this.goodDirectionOutside(this.velocityTowardOutside())) {
			    if (!anglesWithin(this.player.angle,this.angleTowardFortress(),45)) {
				this.aim(this.angleTowardFortress());
                                this.game.addEvent({tag: 'ai-anticipatory-turn-torwards-inside', player: this.game.getPlayerIndex(this.player)});
			    }
			}
		    }
		}
	    }
	}
    }
}

module.exports = Agent;
