'use strict';

var os = require('os');
var JSONStream = require('JSONStream');
var net = require('net');
var game = require('./game');
var server = require('./server');
var experiment = require('./experiment');
var graphics = require('./serverGraphics');

const keyMapping = {};
keyMapping['A'] = game.LEFT;
keyMapping['D'] = game.RIGHT;
keyMapping['W'] = game.THRUST;
keyMapping['Space'] = game.FIRE;
// // Alternate bindings
// keyMapping['Left'] = game.LEFT;
// keyMapping['Right'] = game.RIGHT;
// keyMapping['Up'] = game.THRUST;
// // Keypad bindings
// keyMapping['Keypad 4'] = game.LEFT;
// keyMapping['Keypad 6'] = game.RIGHT;
// keyMapping['Keypad 8'] = game.THRUST;

function TCPServer(port, subjectId, session, datadir, screens, debug, broadcast) {
    server.Server.call(this, subjectId, session, datadir, screens, debug);
    this.port = port;
    this.broadcast = broadcast;
    this.nh = new graphics.NetworkHelper();
    this.screenState = {screen:'blank'};
    this.grHelper = new graphics.GraphicsHelper(this.resourceDir);
}

TCPServer.prototype = Object.create(server.Server.prototype);

TCPServer.prototype.startAudioRecording = function (gnum) {
    this.broadcastClientCommand("audio", {state:true});
    graphics.serverStartLoggingAudio(this.getLogFile(this.subjectId, '-p0-g'+(gnum<10?'0':'')+gnum+'-audio', '.wav'),
                                     this.getLogFile(this.subjectId, '-p1-g'+(gnum<10?'0':'')+gnum+'-audio', '.wav'));
};

TCPServer.prototype.stopAudioRecording = function () {
    this.broadcastClientCommand("audio", {state:false});
    graphics.serverStopLoggingAudio();
};

TCPServer.prototype.broadcastClientCommand = function(command, options) {
    var cmd = {'screen': 'clientCommand', 'cmd': command};
    if (options) {
        for (let k in options) cmd[k] = options[k];
    }
    this.nh.broadcastScreenState(cmd);
};

TCPServer.prototype.broadcastScreenState = function (state) {
    this.addToScreenState(state);
    this.nh.broadcastScreenState(state);
};

TCPServer.prototype.sendScreenState = function (player, state) {
    this.addToScreenState(state);
    this.nh.sendScreenState(player, state);
};

TCPServer.prototype.init = function () {
    if (graphics.numWindows() > 0) {
        this.screenState = {screen:'message', lines: ['Waiting for players'], continue: ['','']};
        this.grHelper.drawScreenState(this.screenState);
    }
    graphics.startServer(this.port);
    this.waitForClients(this.broadcast);

    server.Server.prototype.init.call(this);

    var ips = os.networkInterfaces();
    // console.log(ips);
    for (var k in ips) {
        for (var i=0; i<ips[k].length; i++) {
            if (ips[k][i].family === 'IPv4') {
                console.log(k+': listening on', ips[k][i].address,'port', this.port);
            }
        }
    }
};

// the server decides everything--what screens to display, when to
// start the game, what game number it is. the clients are complete
// dumb terminals.  All they do is obey. logging happens on the
// server. the clietns, maybe we don't have to worry about it. it'd be
// good to record the latency though.

TCPServer.prototype.currentScreenChanged = function() {
    console.log('Screen ' + (this.currentScreen+1) + '/' + this.screens.length + ' ' + this.formatBonus());
    if (graphics.numWindows() > 0) {
        this.screenState = {screen:'message', lines: ['Running Experiment',
                                                      '',
                                                      'Screen ' + (this.currentScreen+1) + '/' + this.screens.length], continue: ['','']};
        this.grHelper.drawScreenState(this.screenState);
    }
};

TCPServer.prototype.waitForClients = function() {
    if (graphics.numWindows() > 0) {
        graphics.processEvents((e) => {
            if (e.type === 'windowEvent' && e.windowEvent === 'exposed') {
                this.grHelper.drawScreenState(this.screenState);
            }
        });
    }

    this.nh.waitForClients(this.broadcast);

    if (graphics.getNumClients() >= 2) {
        console.log('start the experiment');
        this.startExperiment();
        this.processEvents();
        process.on('SIGINT', () => { this.inputEvent({type:"quit"}); });
    } else {
        if (graphics.numWindows() > 0) {
            this.screenState = {screen:'message', lines: ['Waiting for players...',
                                                          '',
                                                          'Players connected: ' + graphics.getNumClients()], continue: ['','']};
            this.grHelper.drawScreenState(this.screenState);
        }
        setTimeout(this.waitForClients.bind(this), 1000);
    }
};

TCPServer.prototype.processEvents = function() {
    if (graphics.numWindows() > 0) graphics.processEvents(() => {});
    for (let pnum=0; pnum<graphics.getNumClients(); pnum++) {
        this.nh.processClient(pnum, pnum, keyMapping, this.inputEvent.bind(this));
    }
    if( graphics.serverIsConnected() ) {
        this.eventTimer = setTimeout(this.processEvents.bind(this), 0);
    } else {
        console.log('Disconnected');
        this.shutdown();
    }
};

TCPServer.prototype.shutdown = function() {
    server.Server.prototype.shutdown.call(this);
    clearTimeout(this.eventTimer);
    console.log('TCP shutdown complete.');
};

exports.TCPServer = TCPServer;
