;;;  -*- mode: LISP; Syntax: COMMON-LISP;  Base: 10 -*-
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Dan Bothell & Shawn Betts
;;; Copyright   : (c) 2018 Dan Bothell
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;; Address     : Department of Psychology 
;;;             : Carnegie Mellon University
;;;             : Pittsburgh, PA 15213-3890
;;;             : db30@andrew.cmu.edu
;;; 
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : sfcoop-module.lisp
;;; Version     : 3.0
;;; 
;;; Description : Module and device for connecting to the co-op version of Space
;;;             : Fortress through a socket connection, and supporting multiple
;;;             : models playing simultaneously.
;;;
;;; Bugs        : 
;;;
;;; To do       : [ ] Investigate the chunk table settings on performance.
;;;
;;; 
;;; ----- History -----
;;;
;;; 2018.09.21 Shawn [1.0]
;;;             : * Initial version based on previous sf-module code (which was
;;;                 built on an interface written for Unreal Tournament).
;;; 2018.09.25 Dan
;;;             : * Added the header (stripped the old sf history info).
;;;             : * Cleaned up some of the old comments.
;;;             : * Changed to version 7.5 as required for now.
;;;             : * Just load all .lisp files from ACT-R:space-fortress-modules.
;;;             : * Reworked the whole visual feature system to just provide 
;;;             :   constant interface items (my-ship, my-explosion, other-ship, 
;;;             :   other-explosion, fortresss, fortress-explosion, and score), 
;;;             :   dynamic items (shells and missles), and user items (anything
;;;             :   else) without the need for a "visible" slot to determine if 
;;;             :   they're available.  
;;;             : * Also removed all the funny business with proc-display at 
;;;             :   attention time.
;;;             : * For now, putting all the features into both the loc and obj
;;;             :   chunks.
;;;             : * Removed all of the hooks except the data hook.
;;; 2018.09.26 Dan
;;;             : * Added parameters for print-comm-data and the time to wait
;;;             :   before sending the continue automatically between games.
;;;             : * Save the data records in order (push-last instead of push).
;;;             : * Updated play-sf and play-sf-games to work in a multi-model
;;;             :   situation.  Requires a player to set the :sf-player-number
;;;             :   parameter to enable playing.
;;;             : * Set the server-mode to a symbol instead of string for easier
;;;             :   use in queries (and then use case instead of cond).
;;;             : * Added the update-time slot to the module which will get the
;;;             :   interval length from the game.
;;;             : * Don't queue up the responses -- send them immediately.
;;; 2018.09.27 Dan
;;;             : * Realized there's an issue with synchronizing things at the
;;;             :   start and between games.  Can't just "wait" for the game
;;;             :   start signal because the model will fall behind the real
;;;             :   time clock so it needs to continue to advance the model time
;;;             :   too.  The "right" solution would be to get access to the
;;;             :   real time clock so it could wait and then schedule the
;;;             :   update/change to happen at the current real time, but that
;;;             :   will require a custom clock function and gets ugly (at least
;;;             :   for 7.5).  So, for now just putting in a dummy check event
;;;             :   that'll just burn through time at a fourth the game tick size
;;;             :   (so it isn't wasting too much time spinning but should
;;;             :   still keep the model withing a tick of the real start).
;;;             : * Added a scene-change flag because only need to schedule
;;;             :   proc-display if something updated for efficiency.
;;; 2018.09.28 Dan [1.1]
;;;             : * Create a real schedule of update times based on the fps value.
;;;             : * In model mode can't wait for the step response because 
;;;             :   multiple models are going to send step sequentially.
;;;             : * Fixed some issues with respect to sending the step and 
;;;             :   continue -- it only needs to send a single continue
;;;             :   while waiting for a new game to start.
;;; 2018.10.01 Dan
;;;             : * Working on updating the visual features for the model.
;;; 2018.10.02 Dan [1.2]
;;;             : * Seems stable now and visual features seem to be working.
;;; 2018.10.03 Dan
;;;             : * Fixed the reason slot of ship explosions since the event
;;;             :   only comes over on the initial death message.
;;;             : * Added the other features to the fortress chunks.
;;;             :   [Translate the 1/2 value for target to self/other but that
;;;             :    may change to 0/1 in a future update.]
;;;             : * Fixed the missing id slot in shots and missiles.
;;;             : * Fixed a bug with remove-visual-feature.
;;;             : * Updated for 0.53 server that sends 0/1 for fortress targets.
;;; 2018.10.04 Dan [1.3]
;;;             : * Use require-compiled to load the extensions and move them
;;;             :   into a subdirectory under space-fortress-modules.
;;; 2018.10.18 Dan [2.0]
;;;             : * Start converting this over to 7.12.
;;;             :   - Use usocket instead of uni-files.
;;;             :   - Use built-in json encoder/decoder.
;;;             :   - Don't use the module as a device.
;;;             :   - Add locks around the module code.
;;;             :   - Use visicon-features instead of chunks.
;;;             :   - Keep it single-threaded right now, but eventually will
;;;             :     switch to having a second thread parse the incoming data,
;;;             :     and in timer mode just schedule the updates as they come
;;;             :     in (model mode should still use the fixed update events).
;;;             : * Simplify the game counting so it doesn't have to iterate
;;;             :   over all the connected models.
;;; 2018.10.19 Dan
;;;             : * CLJSON is slow -- half the run time for the game is spent
;;;             :   in the decoder!  Going back to Shawn's simple JSON parsers.
;;; 2018.10.22 Dan
;;;             : * Add the fortress-explosion and fortress chunks.
;;;             : * Still using the repeated checks for start, but once that's
;;;             :   in a background process it could just schedule for "now".
;;; 2018.11.05 Dan
;;;             : * Added a chunk named bonus.
;;;             : * Ignore the world parameter in create/update player-visual-feature
;;;             :   (should probably just remove it).
;;; 2020.01.28 Dan
;;;             : * Fixed the call to set-default-vis-loc-slots since it's a 
;;;             :   macro as of 7.13.2.
;;; 2020.10.14 Dan [3.0]
;;;             : * Schedule all the events with a priority of :min so that they
;;;             :   will cycle among models if they're all scheduling events for
;;;             :   the same time repeatedly.
;;;             : * Don't schedule the next update for a model immediately upon
;;;             :   the game starting.  Instead, wait for all the models to be
;;;             :   connected and then schedule them all together.
;;;             : * Make sure that handle-server-message always gets called and
;;;             :   have it clear pending-step for continue messages too.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; The code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Standard ACT-R checks for putting things in an appropriate package as needed.

#+:packaged-actr (in-package :act-r)
#+(and :clean-actr (not :packaged-actr) :ALLEGRO-IDE) (in-package :cg-user)
#-(or (not :clean-actr) :packaged-actr :ALLEGRO-IDE) (in-package :cl-user)

;;; Do a version check.

(written-for-act-r-version "7.21" (format nil "Space Fortress module.~%           Needs fully updated extended-motor-actions."))

;;; Load the necessary ACT-R extensions.  Some are included with ACT-R and others
;;; are specific to the SF interface and included separately.
;;; Also create a logical host called SF-DATA pointing to the directory where
;;; this file is located.

(require-extra "extended-motor-actions")

;; require the loader which will load all of the .lisp files from the 
;; space-fortress-modules/load directory.

(require-compiled "SF-MODULES" "ACT-R:space-fortress-modules;loader.lisp")

;; Create the logical SF-DATA to point to the directory of this file.

(setf (logical-pathname-translations "SF-DATA")
  `(("**;*.*" ,(namestring (merge-pathnames "**/*.*" *load-truename*)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Set the system parameters to adjust the chunk table initial size.  Improved
;;; performance in original SF due to the number of chunks created while playing
;;; (mostly visual and imaginal).
;;; May or may not be necessary now. 

(ssp :mcts 45000 :mctrt .75)

;;; Simple json parser for game communication

(defun sexp2json (sexp)
  "Convert a simple s-expression into a JSON representation in a string ."
  (cond
    ((and (listp sexp)
          (evenp (length sexp))
          (loop for rest = sexp then (cddr rest)
                for k = (first rest)
                while rest
                always (keywordp k)))
     (format nil "{~{~A:~A~^,~}}" (mapcar #'sexp2json sexp)))
    ((listp sexp)
     (format nil "[~{~A~^,~}]" (mapcar #'sexp2json sexp)))
    ((symbolp sexp)
     (with-output-to-string (out)
       (write-char #\" out)
       (loop with s = (symbol-name sexp)
             with capitalize = nil
             for i below (length s)
             do (cond
                  ((char= (elt s i) #\-)
                   (setf capitalize t))
                  (capitalize
                   (write-char (char-upcase (elt s i)) out)
                   (setf capitalize nil))
                  (t
                   (write-char (char-downcase (elt s i)) out))))
       (write-char #\" out)))
    ((stringp sexp) (format nil "~S" sexp))
    ((numberp sexp) (format nil "~A" sexp))
    (t (warn "Dunno how to translate ~S to JSON" sexp)
       "null")))

(defun json2sexp (str &optional ( start 0))
  "A json parser. Not complete but good enough for communicating with
the server."
  (let ((ch (char str start)))
    (cond
      ((char= ch #\{)
       (incf start)
       (loop with k = nil
             with v = nil
             with ret = nil
             with done = nil
             until done
             do (multiple-value-setq (k start) (json2sexp str start))
                (when (char/= (char str start) #\:) (error "dict parse error"))
                (incf start)
                (multiple-value-setq (v start) (json2sexp str start))
                (push (intern (string-upcase k) :keyword) ret)
                (push v ret)
                (if (char= (char str start) #\,)
                    (incf start)
                    (progn
                      (when (char/= (char str start) #\}) (error "dict parse error"))
                      (incf start)
                      (setf done t)))
             finally (return (values (nreverse ret) start))))
      ((char= ch #\[)
       (incf start)
       (when (char= (char str start) #\])
         (incf start)
         (return-from json2sexp (values nil start)))
       (loop with v = nil
             with ret = nil
             with done = nil
             until done
             do (multiple-value-setq (v start) (json2sexp str start))
                (push v ret)
                (if (char= (char str start) #\,)
                    (incf start)
                    (progn
                      (when (char/= (char str start) #\]) (error "array parse error"))
                      (incf start)
                      (setf done t)))
             finally (return (values (nreverse ret) start))))
      ((char= ch #\")
       (incf start)
       (loop for i from start below (length str)
             when (char= (char str i) #\")
               do (return (values (subseq str start i) (1+ i)))
             finally (error "string parse error")))
      ((or (and (char>= ch #\0)
                (char<= ch #\9))
           (char= ch #\-))
       (loop with beg = (if (char= ch #\-) (1+ start) start)
             with dot = nil
             with exponent = nil
             with done = nil
             for i from beg below (length str)
             until done
             do (cond
                  ((char= (char str i) #\.)
                   (if (or dot exponent)
                       (error "number parse error")
                       (setf dot t)))
                  ((char= (char str i) #\e)
                   (if exponent
                       (error "exponent parse error")
                       (setf exponent t)))
                  ((char= (char str i) #\-)
                   (unless (char= (char str (1- i)) #\e)
                     (error "exponent minus parse error")))
                  ((and (char>= (char str i) #\0)
                        (char<= (char str i) #\9)))
                  (t (setf done t)))
             finally (return (values (read-from-string (subseq str start (1- i)) t) (1- i)))))
      ((string= str "true" :start1 start :end1 (+ start 4))
       (values t (+ start 4)))
      ((string= str "false" :start1 start :end1 (+ start 5))
       (values nil (+ start 5)))
      ((string= str "null" :start1 start :end1 (+ start 4))
       (values nil (+ start 4)))
      (t (error "top-level parse error")))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Structure to keep track of models playing, games played, and syncing game starts

(defstruct game-counter ready-players players current-game limit updated-players (lock (bt:make-lock "game-counter")))
(defvar *game-counter* (make-game-counter))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Create a class which will be a module for each model.

(defclass sfcoop ()
  (;; connection info
   (host :initform nil :accessor sf-host)
   (port :initform nil :accessor sf-port)
   (socket :initform nil :accessor socket)
   (server-mode :initform nil :accessor server-mode)
   (connection-state :initform nil :accessor connection-state)
   (server-player :initform nil :accessor server-player)
   (server-stepping :initform nil :accessor server-stepping)

   ;; communication logging
   (collect-incoming :initform nil :accessor collect-incoming)
   (record-outgoing :initform nil :accessor record-outgoing)
   (data-record :initform nil :accessor data-record)
   (print-communication :initform t :accessor print-communication)
   
   ;; visual features
   (my-ship :initform nil :accessor my-ship)
   (other-ship :initform nil :accessor other-ship)
   (fortress :initform nil :accessor fortress)
   (pnts :initform nil :accessor pnts)
   (bonus :initform nil :accessor bonus)
   (dynamics :initform nil :accessor dynamics)

   ;; user hook parameters
   (data-hook :initform nil :accessor data-hook)

   ;; track some game stats as it plays
   (game-stats :initform (list (list 0 0)) :accessor game-stats)

   ;; connection and running control
   (pending-step :initform nil :accessor pending-step)
   (continue-sent :initform nil :accessor continue-sent)
   (continue-time :initform 5 :accessor continue-time)
   (update-time :initform 0 :accessor update-time)
   (update-times :initform '(1000) :accessor update-times)
   (game-freq :initform 1 :accessor game-freq)
   (name :initform nil :accessor name)
   (change-drawing-mode :initform nil :accessor change-drawing-mode)
   
   ;; locks for various pieces
   (comm-lock :initform (bt:make-recursive-lock "sf-comm") :accessor comm-lock)
   ; socket server-mode connection-state data-record server-stepping
   
   (param-lock :initform (bt:make-recursive-lock "sf-param") :accessor param-lock)
   ; host port server-player collect-incoming record-outgoing data-hook continue-time print-communication
   
   (state-lock :initform (bt:make-recursive-lock "sf-state") :accessor state-lock)
   ; everything else...
   ))


(defun model-keypress (model key)
  (declare (ignore model))
  (server-send-command (list :cmd "keydown" :value (model-key-to-sf-key key))))

(defun model-keyrelease (model key)
  (declare (ignore model))
  (server-send-command (list :cmd "keyup" :value (model-key-to-sf-key key))))

(add-act-r-command "sf-keypress" 'model-keypress "Monitor for output-key actions for space fortress module.")
(add-act-r-command "sf-keyrelease" 'model-keyrelease "Monitor for release-key actions for space fortress module.")

(monitor-act-r-command "output-key" "sf-keypress")
(monitor-act-r-command "release-key" "sf-keyrelease")


;; Functions for use if one wants to do something other than
;; space fortress with current ACT-R session and then return to space
;; fortress afterwards.

(defun remove-sf-monitors ()
  (remove-act-r-command-monitor "output-key" "sf-keypress")
  (remove-act-r-command-monitor "release-key" "sf-keyrelease"))

(defun restore-sf-monitors ()
  (monitor-act-r-command "output-key" "sf-keypress")
  (monitor-act-r-command "release-key" "sf-keyrelease"))

;;; Since the model actions are generated based on hand and finger specifications
;;; just assume the default hand positions (home row) to keep things simple since
;;; hand position doesn't affect the punch timing (there's no hand/finger movement
;;; necessary to play).
;;;

(defun model-key-to-sf-key (key)
  (cond ((string-equal key "d")
         "thrust")
        ((string-equal key "s")
         "left")
        ((string-equal key "f")
         "right")
        ((string-equal key "j")
         "fire")
        (t "nothing")))



;;; JSON parser in ACT-R uses conses for JSON objects (in and out)
;;; use these to convert the lists used for output and get data from input items.
;;;
;;; Since I switched back to the simple parser, these aren't really needed
;;; but keeping them around just incase I decide to switch to a different
;;; version again at some point.

(defun get-value (data key)
  (getf data key)) ; (cdr (assoc key data)))

(defun list-to-conses (data) data) 
#|
  (do* ((c (pop data) (pop data))
        (d (pop data) (pop data))
        (i (cons (string-downcase (symbol-name c)) d) (cons (string-downcase (symbol-name c)) d))
        (r (list i) (push-last i r)))
       ((null data) r)))
|#

;;; communication parsing

(defun server-send-command (cmd &optional module)
  "Write a command to the server socket."
  (let ((world (or module (get-module :sfcoop))))
    (bt:with-recursive-lock-held ((param-lock world))
      (if (eq (connection-state world) 'disconnected)
          (print-warning "Can't send command {~S} because connection closed" cmd))
      (bt:with-recursive-lock-held ((comm-lock world))
        (when (record-outgoing world)
          (push-last (list :out (mp-time) (get-internal-real-time) cmd)
                     (data-record world)))
        (let ((js (sexp2json (list-to-conses cmd))))
          (when (print-communication world) (model-output ">>> ~A~%" js))
          (format (usocket:socket-stream (socket world)) "~a" js)
          (force-output (usocket:socket-stream (socket world))))))))

(defun server-read-message (world)
  "Read a line of data from the server skipping blank lines and terminating the run if there's an error encountered."
  (loop
    (bt:with-recursive-lock-held ((comm-lock world))
      (multiple-value-bind (line err)
          (ignore-errors (read-line (usocket:socket-stream (socket world))))

        (when (or (null line) (subtypep (type-of err) 'condition))
          (setf (connection-state world) 'disconnected)
          (print-warning "Error reading from game server.  Cannot continue.")
          (schedule-break-relative 0 :priority :max :details (format nil "Game connection error for model ~s" (current-model)) :time-in-ms t)
          (return nil))
        
        (unless (string= line "")
          (let (print collect hook (data (json2sexp line)))
            (bt:with-recursive-lock-held ((param-lock world))
              (setf print (print-communication world))
              (setf collect (collect-incoming world))
              (setf hook (data-hook world)))
            
            (when print (model-output "<<< ~A~%" line))
            
            (when hook
              (multiple-value-bind (res e)
                  (ignore-errors (funcall hook data))
                (declare (ignore res))
                (when (subtypep (type-of e) 'condition)
                  (print-warning "Condition: ~s encountered while calling data-hook ~s" e hook))))
            
            (when collect
              (push-last (list :in (mp-time) (get-internal-real-time) data) (data-record world)))
            
            (return data)))))))


;; parse the player info into a list of feature value lists

(defun default-player-features (player)
  (let ((x (get-value (get-value player :pos) :x))
        (y (get-value (get-value player :pos) :y))
        (vx (get-value (get-value player :vel) :x))
        (vy (get-value (get-value player :vel) :y)))
    
    (list 'x x
          'y y
          'vx vx
          'vy vy
          'angle (get-value player :angle)
          'inside-hex (get-value player :insidebighex)
          'vel (sqrt (+ (expt vx 2) (expt vy 2)))
          'dist (sqrt (+ (expt x 2) (expt y 2))))))


(defun default-shot-features (shot)
  (let ((x (get-value (get-value shot :pos) :x))
        (y (get-value (get-value shot :pos) :y))
        (vx (get-value (get-value shot :vel) :x))
        (vy (get-value (get-value shot :vel) :y)))
    
    (list 'x x
          'y y
          'vx vx
          'vy vy
          'angle (get-value shot :angle)
          'id (get-value shot :id)
          'vel (sqrt (+ (expt vx 2) (expt vy 2))))))

;; List of update lengths for 1 second based on frequency

(defun create-times (freq)
  (multiple-value-bind (base rem) (truncate 1000 freq)
    (let ((times (make-list freq :initial-element base)))
      (unless (zerop rem)
        (let* ((total (/ rem freq))
               (step total)
               (inc 0))
          (dotimes (i freq)
            (when (> (truncate total) inc)
              (incf (nth i times))
              (incf inc))
            (incf total step))))
      times)))



(defun wait-for-game-start (world)
  (if (listen (usocket:socket-stream (socket world)))
    (let ((data (server-read-message world)))
      ;(format t "Wait ~S got ~S~%" (current-model) data)
      (if (and (equal (getf data :msgtype) "response")
               (equal (getf data :cmd) "continue"))
          (progn
            (handle-server-message world data)
            (schedule-next-update world 0))
          (if (and (equal (getf data :msgtype) "stateChange")
                   (equal (getf data :state) "game"))
              (progn 
                (signal-player-ready data))
              (progn
                (print-warning "Unexpected incoming message ~s while waiting for game start." data)
                (schedule-next-update world 0)))))
  ;; if there wasn't any data reschedule to try again at current time
    (schedule-next-update world 0)))


(defun signal-player-ready (data)
  (bt:with-lock-held ((game-counter-lock *game-counter*))
    (push (cons (current-model) data) (game-counter-ready-players *game-counter*))
    (when (= (length (game-counter-ready-players *game-counter*)) 
                     (game-counter-players *game-counter*))
    (dolist (m (game-counter-ready-players *game-counter*))
      (with-model-eval (car m)
        (handle-server-message (get-module :sfcoop) (cdr m))
        (schedule-next-update (get-module :sfcoop))))
    (setf (game-counter-ready-players *game-counter*) nil))))


(defun handle-server-message (world data)
  (cond
   ((equal (get-value data :msgtype) "response")
    (if (get-value data :success)
        (cond
         ((or (equal (get-value data :cmd) "gameState")
              (equal (get-value data :cmd) "step")
             )
          (when (equal (get-value data :cmd) "step")
            (bt:with-recursive-lock-held ((state-lock world))
              (setf (pending-step world) nil)))
          (handle-state-game world (get-value data :gamestate)))
         ((equal (get-value data :cmd) "continue")
          (bt:with-recursive-lock-held ((state-lock world))
              (setf (pending-step world) nil)))
          (t
          ;; ignore for now
          ))
      (cond
       ((equal (get-value data :cmd) "step")
        (bt:with-recursive-lock-held ((state-lock world))
          (setf (pending-step world) nil)))
       (t (print-warning "Response is unsuccessful: ~S" data)))))
   ((equal (get-value data :msgtype) "stateChange")
    (handle-state-change world data))
   ((equal (get-value data :msgtype) "gameStateUpdate")
    (handle-state-game world (get-value data :gamestate)))
   (t
    (print-warning "Unknown message type: ~S" data))))


(defun handle-state-change (world data)
  (let ((state (read-from-string (get-value data :state))))
    (bt:with-recursive-lock-held ((comm-lock world)) (setf (server-mode world) state))
    (case state
      (score
       (handle-state-score world data))
      (game
       (handle-state-game world (get-value data :gamestate)))
      (t
       (print-warning "Unknown state change: ~S" data)))))

(defun handle-state-score (world data)
  (bt:with-recursive-lock-held ((state-lock world))
    (push-last (get-value data :rawPoints) (first (game-stats world)))
    (push-last (get-value data :bonus) (first (game-stats world)))
    (push-last (get-value data :total-bonus) (first (game-stats world)))
  
    (push (list 0 0) (game-stats world))
 
    (update-game-count)
  
    (setf (continue-sent world) nil)
    
    ;; delete all of the default features
    ;; don't use delete-all-visicon-features because
    ;; the modeler may have added some that should remain
    
    (let (remove)
      
      (awhen (my-ship world)
             (push it remove)
             (setf (my-ship world) nil))
      (awhen (other-ship world)
             (push it remove)
             (setf (other-ship world) nil))
      (awhen (fortress world)
             (push it remove)
             (setf (fortress world) nil))
      (awhen (pnts world)
             (push it remove)
             (setf (pnts world) nil))
      
      (dolist (x (dynamics world))
        (push (cdr x) remove))
      (setf (dynamics world) nil)
      
      (apply 'delete-visicon-features remove))
    
    ;; add a bonus feature
    (setf (bonus world) (add-visicon-features (list 'points (get-value data :points) 'bonus (get-value data :bonus) 
                                                    'x 0 'y 0 'type 'bonus)))))


(defun create-player-visual-feature (world data player-data who)
  (declare (ignore world))
  (car (add-visicon-features (append (default-player-features player-data) 
                                       (if (get-value player-data :alive)
                                           (list 'value who 'type 'ship)
                                         (let ((reason (cond ((find "player-hit-shield" (get-value data :events)
                                                                    :test #'string=
                                                                    :key (lambda (e) (get-value e :tag)))
                                                              'inner)
                                                             ((find "player-hit-border" (get-value data :events)
                                                                    :test #'string=
                                                                    :key (lambda (e) (get-value e :tag)))
                                                              'outer)
                                                             ((find "shell-hit-player" (get-value data :events)
                                                                    :test #'string=
                                                                    :key (lambda (e) (get-value e :tag)))
                                                              'shell)
                                                             (t 'unknown))))
                                           (list 'reason reason 'type 'ship-explosion 'value who)))))))


(defun update-player-visual-features (world data player-data feature)
  (declare (ignore world))
  (modify-visicon-features (append (list feature) (default-player-features player-data) 
                                     (if (get-value player-data :alive)
                                         (list 'reason nil 'type 'ship)
                                       (let ((reason (cond ((find "player-hit-shield" (get-value data :events)
                                                                  :test #'string=
                                                                  :key (lambda (e) (get-value e :tag)))
                                                            'inner)
                                                           ((find "player-hit-border" (get-value data :events)
                                                                  :test #'string=
                                                                  :key (lambda (e) (get-value e :tag)))
                                                            'outer)
                                                           ((find "shell-hit-player" (get-value data :events)
                                                                  :test #'string=
                                                                  :key (lambda (e) (get-value e :tag)))
                                                            'shell))))
                                         (if reason
                                             (list 'reason reason 'type 'ship-explosion)
                                           (list 'type 'ship-explosion)))))))

(defun handle-state-game (world data)
  
  (let ((player (bt:with-recursive-lock-held ((param-lock world))
                  (server-player world))))
    
    (bt:with-recursive-lock-held ((state-lock world))
      
      ;; increment the time until next update
      (incf (update-time world) (nth (mod (get-value data :tick) (game-freq world)) (update-times world)))
    
      ;; record deaths and fortress kills
      (when (find "kill-player" (get-value data :events)
                  :test #'string=
                  :key (lambda (e) (get-value e :tag)))
        (incf (first (first (game-stats world)))))
      
      (when (find "kill-fortress" (get-value data :events)
                  :test #'string=
                  :key (lambda (e) (get-value e :tag)))
        (incf (second (first (game-stats world)))))
      
      ;; Update players
      (let ((self (elt (get-value data :players) player))
            (other (elt (get-value data :players) (if (eql player 0) 1 0))))
                                       
      (if (my-ship world)
          (update-player-visual-features world data self (my-ship world))
        (setf (my-ship world)
          (create-player-visual-feature world data self 'self)))
      
      (if (other-ship world)
          (update-player-visual-features world data other (other-ship world))
        (setf (other-ship world)
          (create-player-visual-feature world data other 'other)))
      
  
      ;; Fortress
        (let* ((fort (get-value data :fortress))
               (alive (get-value fort :alive))
               (fort-feats (list 'x 0 'y 0 'angle (when alive (get-value fort :angle))
                               'vulnerable (when alive (get-value fort :vulnerable))
                                 'target (when alive (awhen (get-value fort :target)
                                                            (if (= player it)
                                                                'self 'other)))
                                 'type (if (get-value fort :alive) 'fortress 'fortress-explosion))))
        (if (fortress world)
            (modify-visicon-features (push (fortress world) fort-feats))
          (setf (fortress world)
            (car (add-visicon-features fort-feats)))))
      
      ;; points
      (let ((points (get-value data :points)))
        (if (pnts world)
            (modify-visicon-features (list (pnts world) 'points points))
          (setf (pnts world)
            (car (add-visicon-features (list 'x 0 'y -230 'type 'score 'points points))))))
      

      ;; update all the shells and missiles
      
        (let* ((shells nil) (missiles nil) 
               (my-pos (get-value self :pos))
               (mx (get-value my-pos :x))
               (my (get-value my-pos :y))
               (o-pos (get-value other :pos))
               (ox (get-value o-pos :x))
               (oy (get-value o-pos :y)))
          
          (dolist (shell (get-value data :shells))
          (let* ((id (get-value shell :id))
                 (details (default-shot-features shell))
                 (self-dist (dist (list (second details) (fourth details))
                                  (list mx my)))
                 (other-dist (dist (list (second details) (fourth details))
                                   (list ox oy)))
                 (all-feats (append details (list 'type 'shell)
                                    (if (< other-dist self-dist) 
                                        (list 'closest 'other 'dist other-dist)
                                      (list 'closest 'self 'dist self-dist))))
                 (feat (assoc id (dynamics world))))
            
            (if feat
                (progn
                  (push feat shells)
                  (modify-visicon-features (push (cdr feat) all-feats))
                  (setf (dynamics world) (remove feat (dynamics world))))
              (push (cons id (car (add-visicon-features all-feats))) shells))))
    
          (dolist (missile (get-value data :missiles))
            (let* ((id (get-value missile :id))
                   (details (default-shot-features missile))
                   (all-feats (append details (list 'type 'missile)
                                      (list 'fort-dir
                                            (if (< (dist '(0 0) (list (+ (second details) (sixth details))
                                                                      (+ (fourth details) (eighth details))))
                                                   (dist '(0 0) (list (second details) (fourth details))))
                                                'toward
                                              'away))))
                   (feat (assoc id (dynamics world))))
              
              (if feat
                  (progn
                    (push feat missiles)
                    (modify-visicon-features (push (cdr feat) all-feats))
                    (setf (dynamics world) (remove feat (dynamics world))))
                (push (cons id (car (add-visicon-features all-feats))) missiles))))
          
          (apply 'delete-visicon-features (mapcar 'cdr (dynamics world)))
          
          (setf (dynamics world) (append shells missiles)))))))

(defun send-continue (world)
  (server-send-command '(:cmd "continue") world)
  (let ((data (server-read-message world)))
    (if (and (equal (get-value data :msgtype) "response")
             (equal (get-value data :cmd) "continue"))
        (unless (get-value data :success)
          (print-warning "Continue command failed somehow"))
      (print-warning "Unexpected incoming data ~s when sending a continue" data))
    (when data (handle-server-message world data))))


(defun schedule-next-update (world &optional delay)
  (case (bt:with-recursive-lock-held ((comm-lock world)) (server-mode world))
    (ready ;; shouldn't happen now because we wait for the game data
     (schedule-event-relative (or (and (numberp delay) delay)
                                  0)
                              'wait-for-game-start :destination :sfcoop :module :sfcoop :output 'high :maintenance t :time-in-ms t :priority :min))
    (game
     (schedule-event-relative (or (and (numberp delay) delay)
                                  (bt:with-recursive-lock-held ((state-lock world)) (update-time world)))
                              'update-world :destination :sfcoop :module :sfcoop :maintenance t :time-in-ms t :output 'high :priority :min))
    (score
     (bt:with-recursive-lock-held ((state-lock world))
       (unless (continue-sent world)
         (schedule-event-relative (or (and (numberp delay) delay)
                                      (bt:with-recursive-lock-held ((param-lock world))(continue-time world)))
                                  'send-continue :destination :sfcoop :module :sfcoop :maintenance t :time-in-ms t :output 'high :priority :min)
         (setf (continue-sent world) t))
       (schedule-event-relative (or (and (numberp delay) delay)
                                    (bt:with-recursive-lock-held ((param-lock world))(continue-time world))) 
                                'wait-for-game-start :destination :sfcoop :module :sfcoop :output 'high :maintenance t :time-in-ms t :priority :min)))))



(defun update-world (world) ;; only called in game mode now
  (awhen (change-drawing-mode world)
         (server-send-command (list :cmd "drawing" :value it) world)
         (setf (change-drawing-mode world) nil))
  
  ;; will be incremented in game state handler as needed
  
  (setf (update-time world) 0)
  
  (case (server-stepping world)
    (:model ;; this is synchronous
            ;; shouldn't see data on socket
            ;; unless there's a state change
            ;; or a drawing change was sent
  
     ;; Drain the socket
     (loop while (listen (usocket:socket-stream (socket world)))
         for data = (server-read-message world)
         while data
         do 
           (handle-server-message world data))
  
     (when (eq (server-mode world) 'game) ;; may have transitioned to score
       (unless (pending-step world)
         (server-send-command (list :cmd "step") world)
         (setf (pending-step world) t))))
    
    (:timer ;; this is asynchronous
            ;; we could have 0 updates (the game is lagged or we're early)
            ;; could be 1 or more if we're behind 
            ;; for now just process all of them since the last
            ;; will be the most recent.
            ;; Update time will be based on how many game
            ;; states arrived.
     
     (loop while (listen (usocket:socket-stream (socket world)))
         for data = (server-read-message world)
         while data
         do 
           (handle-server-message world data)))
     
    (t
     (print-warning "Unknown server-mode: ~S" (server-mode world))))
  (schedule-next-update world))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Some functions to handle trig info (some of these come over directly now 
;;; and don't need to be computed on the model side).

(defun vdir (x y vx vy)
  (let* ((dir (atan vy vx))
         (actual (to-fortress-orient x y))
         (diff (rad->deg (- dir actual))))
    (if (> diff 180)
        (- diff 360)
      (if (< diff -180)
          (+ diff 360)
        diff))))


(defun angle-to-fortress (x y orient)
  (let* ((actual (rad->deg (to-fortress-orient x y)))
         (diff (- orient actual)))
        (if (> diff 180)
        (- diff 360)
      (if (< diff -180)
          (+ diff 360)
        diff))))

(defun to-fortress-orient (x y)
  ;; Fortress is at 0,0
  (atan y x))

(defun closest-projectile (target-x target-y projectiles)
  (let (best
        best-dist)
    (dolist (s projectiles)
      (let* ((s-x (get-value s :x))
             (s-y (get-value s :y))
             (d (dist (vector target-x target-y) (vector s-x s-y))))
        (when (or (null best) (< d best-dist))
          (setf best s)
          (setf best-dist d))))
    best))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The module to provide the interface, parameters, etc

(defun create-sfcoop-module (name)
  (declare (ignore name))
  (make-instance 'sfcoop))

(defun pre-reset-sf-module (world)

  ;; Kill any previous connection
  
  (bt:with-recursive-lock-held ((comm-lock world))
    (when (usocket::usocket-p (socket world))
      (usocket:socket-close (socket world)))
    
    (setf (connection-state world) 'disconnected)
    (setf (data-record world) nil)
    (setf (server-mode world) nil))
  
  ;; clear out the slots as needed
  
  (bt:with-recursive-lock-held ((state-lock world))
  
    (setf (update-time world) 0)
    (setf (pending-step world) nil)
  
    ;; track deaths and kills per game
    (setf (game-stats world) (list (list 0 0)))

    ;; name for the data file
    (setf (name world) nil)

    ;; track complete games
    (setf (change-drawing-mode world) nil)))


(defun post-reset-sf-module (world)

  ;; create custom chunk-types for the features and objects
  
  (chunk-type (sf-visual-location (:include visual-location)) x y z type)
  (chunk-type (sf-object (:include visual-object)) x y z type)
  
  (chunk-type (ship (:include sf-object)) (type ship) angle vel dist vx vy inside-hex)
  
  (chunk-type (ship-explosion (:include ship)) (type ship-explosion) reason)
  
  (chunk-type (fortress (:include sf-object)) (type fortress) angle vulnerable target)
  
  (chunk-type (fortress-explosion (:include sf-object)) (type fortress-explosion))
              
  (chunk-type (projectile (:include sf-object)) vx vy vel angle id)
  
  (chunk-type (missile (:include projectile)) (type missile) fort-dir)
  
  (chunk-type (shell (:include projectile)) (type shell) closest dist)
  
  (chunk-type (score (:include sf-object)) (type score) points)
  
  (chunk-type (bonus (:include score)) (type bonus) bonus)
  
  
  ;; slots for the features will be set as needed
  
  (bt:with-recursive-lock-held ((state-lock world))
    (setf (my-ship world) nil)
    (setf (other-ship world) nil)
    (setf (fortress world) nil)
    (setf (dynamics world) nil)
    (setf (pnts world) nil)
    (setf (bonus world) nil))
  
  ;; the type values for the chunk-types, query values,
  ;; and some values used in the default features
    
  (define-chunks ship ship-explosion missile shell score fortress fortress-explosion
    ready game connecting connected disconnected
    other inner outer unknown toward away bonus)

  
  ;; Set :tracking-clear to nil so it re-encodes explosions
  ;; if visually tracking the ship when it explodes
  (sgp :tracking-clear nil)

  ;;; Have vision use slots named x, y, and z for location info by default
  (set-default-vis-loc-slots x y z)
  
  ;; effectively turn off buffer stuffing by default
  (set-visloc-default color free)
  
  ;; install a keyboard device
  (install-device '("motor" "keyboard")))


(defun delete-sfcoop-module (world)
  ;; Kill an open connection
  (bt:with-recursive-lock-held ((comm-lock world))
    (when (usocket::usocket-p (socket world))
      (usocket:socket-close (socket world)))))


;; respond to the game buffer queries
(defun sf-query (world buffer-name slot value)
  (bt:with-recursive-lock-held ((comm-lock world))
    (case buffer-name
      (game (case slot
              (state
               (case value
                 (busy nil)
                 (free t)
                 (error nil)
                 (t (print-warning "Unknown state query ~S to game buffer" value)
                    nil)))
              (connection-state
               (eq (connection-state world) value))
              (game-state
               (eq value (server-mode world)))
              (t (print-warning "Unknown query ~S ~S to the game buffer" slot value)))))))


;; custom function for printing the buffer-status of the game buffer
(defun print-sf-connection-state ()
  (let ((world (get-module :sfcoop)))
    (bt:with-recursive-lock-held ((comm-lock world))
      (with-output-to-string (s)
        (format s "  connection-state connecting  : ~S"
          (eq (connection-state world) 'connecting))
        (format s "  connection-state connected   : ~S"
          (eq (connection-state world) 'connected))
        (format s "  connection-state disconnected: ~S"
          (eq (connection-state world) 'disconnected))
        (format s "  game-state ready: ~S"
          (eq 'ready (server-mode world)))
        (format s "  game-state game: ~S"
          (eq 'game (server-mode world)))
        (format s "  game-state score: ~S"
          (eq 'score (server-mode world)))))))

;; function to handle the setting and getting of the module's parameters

(defun sf-params (world param)
  (bt:with-recursive-lock-held ((param-lock world))
    (cond ((consp param)
           (case (car param)
             (:sf-host (setf (sf-host world) (cdr param)))
             (:sf-port (setf (sf-port world) (cdr param)))
             (:log-sf-data-in (setf (collect-incoming world) (cdr param)))
             (:log-sf-data-out (setf (record-outgoing world) (cdr param)))
             (:sf-data-hook (setf (data-hook world) (cdr param)))
             (:continue-time 
              (setf (continue-time world) (seconds->ms (cdr param)))
              (cdr param))
             (:print-comm-data (setf (print-communication world) (cdr param)))
             (:sf-player (setf (server-player world) (cdr param)))))
          (t
           (case param
             (:sf-host (sf-host world))
             (:sf-port (sf-port world))
             (:log-sf-data-in (collect-incoming world))
             (:log-sf-data-out (record-outgoing world))
             (:sf-data-hook (data-hook world))
             (:continue-time (ms->seconds (continue-time world)))
             (:print-comm-data (print-communication world))
             (:sf-player (server-player world)))))))


;; define the module itself

(define-module-fct :sfcoop '((game nil nil (connection-state game-state) print-sf-connection-state))
  (list
   (define-parameter :sf-host :valid-test #'stringp :default-value "127.0.0.1"
     :warning "a string" :documentation "IP address or host name of SF server")
   (define-parameter :sf-port :valid-test 'integerp  :default-value 3000
     :warning "an integer" :documentation "Port of the server for connecting")
   (define-parameter :log-sf-data-in :valid-test #'tornil :default-value nil
     :warning "T or nil" :documentation "Record all incoming SF data")
   (define-parameter :log-sf-data-out :valid-test #'tornil :default-value nil
     :warning "T or nil" :documentation "Record all commands sent to SF")
   (define-parameter :print-comm-data :valid-test #'tornil :default-value nil
     :warning "T or nil" :documentation "Print the communication information between the model and game")
   (define-parameter :sf-data-hook :valid-test #'fctornil
     :default-value nil
     :warning "a function or nil" :documentation "Called with each data line received from the game interface")
   (define-parameter :sf-player :valid-test 'numornil :default-value nil
     :warning "a number or nil" :documentation "Enable playing Space Fortress and indicate which player")
   
   (define-parameter :continue-time :valid-test #'posnum
     :default-value 5
     :warning "a positive number" :documentation "Seconds to wait before the model automatically presses to continue between games"))

  :version "3.0"
  :documentation "Module to allow ACT-R 7.12+ to connect and play Co-op Space Fortress game"
  :creation 'create-sfcoop-module
  :reset (list 'pre-reset-sf-module 'post-reset-sf-module)
  :delete 'delete-sfcoop-module
  :query 'sf-query
  :params 'sf-params)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Handle all the config info. for the current game servers when a new session gets
;;; started with one of the running functions
;;;
;;; - read the first config line
;;; - send draw and id settings verifying that the result is t for each
;;;
;;; If anything doesn't work out punt and close the connection

(defun get-model-name ()
  (if (and (boundp '*cluster-run-start*) (boundp '*cluster-run-id*))
      (progn
        (format nil "model_~d_~d" (symbol-value '*cluster-run-start*) (symbol-value '*cluster-run-id*)))
      (let (number)
        (if (probe-file (translate-logical-pathname "SF-DATA:sf-model-number.txt"))
            (with-open-file (f (translate-logical-pathname "SF-DATA:sf-model-number.txt") :direction :input)
              (setf number (ignore-errors (read f nil :done))))
            (setf number -1))
        (if (numberp number)
            (incf number)
            (setf number 0))

        (with-open-file (f (translate-logical-pathname "SF-DATA:sf-model-number.txt") :direction :output :if-exists :supersede :if-does-not-exist :create)
          (write number :stream f))
        (format nil "model_~d" number))))

(defun connect-to-game (module draw stepping)
  (labels ((bad-connection-exit (message)
             (print-warning message)
             (when (usocket::usocket-p (socket module))
               (usocket:socket-close (socket module))
               (setf (socket module) nil))
                                
             (setf (connection-state module) 'disconnected)
             (return-from connect-to-game nil))
           
           (parse-data-line (message)
             (let ((data (server-read-message module)))
               (if data
                   data
                   (bad-connection-exit message))))
           
           (send-command (cmd error-message)
             (server-send-command cmd module)
             (let ((data (parse-data-line error-message)))
               (unless (and (equal (get-value data :msgtype) "response")
                            (equal (get-value data :cmd) (getf cmd :cmd))
                            (get-value data :success))
                 (bad-connection-exit (format nil "Did not get a successful response for command ~S: ~S" cmd data)))
               data))
           
           (set-config-parameter (key value)
             (send-command (list :cmd key :value value)
                           (format nil "Failed reading response to ~A setting" key))))
    
    (bt:with-recursive-lock-held ((comm-lock module))
      
      (if (eq (connection-state module) 'disconnected)
          
          (let ((socket (handler-case (usocket:socket-connect (sf-host module) (sf-port module) :nodelay :if-supported)
                          (error (x) (bad-connection-exit
                                      (format nil "Error ~/print-error-message/ occurred while trying to open a socket connection to host ~s port ~s" 
                                        x (sf-host module) (sf-port module)))))))
            
            (setf (socket module) socket)
            (setf (connection-state module) 'connecting)
            
            (let ((data (parse-data-line "Failed to read initial config")))
              
              (bt:with-recursive-lock-held ((state-lock module))
              
                ;; create the update times to to total a second
                (setf (game-freq module) (get-value data :fps))
                (setf (update-times module) (create-times (game-freq module)))
                
                (when (not (and (equal (get-value data :msgtype) "stateChange")
                                (equal (get-value data :state) "config")))
                  (bad-connection-exit (format nil "Expected initial config line but got ~s" data)))
                
                (set-config-parameter "drawing" (if draw 1 0))
                (set-config-parameter "player" (server-player module))
                (set-config-parameter "stepping" (ecase stepping (:model "model") (:timer "timer")))
                (setf (server-stepping module) stepping)
                (setf (name module) (get-model-name))
                (set-config-parameter "id" (name module))
                (send-command (list :cmd "continue") "Failed to progress from config screen")
                ;; The next message will be a state change into the
                ;; "game" state. So consider the module connected.
                (setf (connection-state module) 'connected
                  (server-mode module) 'ready)
                t)))
        (print-warning "Model ~s is already connected cannot connect." (current-model))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Functions for actually playing the game -- call these to run the model.
;;;
;;; The game can be run for a specific time in seconds using the play-sf function
;;; or a fixed number of games using the play-sf-games function.  Each of those
;;; has one required parameter for the indicated value (time or games).
;;;
;;; They also have 4 keyword parameters:
;;;
;;; :speed should be a number and indicates how fast to run the game relative
;;;        to real-time.  The default is 1 which means at real time.  A bigger
;;;        value runs the model faster than real time and a smaller value slower.
;;;
;;; :cont whether the current run should be continuted.  If this is specified
;;;       as t then it does not reconnect the model before starting running
;;;       and if it is nil then it resets the model and starts a new session.
;;;       The default is nil.
;;;
;;; :draw can be t or nil.  Indicates whether the game server displays
;;;       the game as the model plays.  A value of t means show everything and nil
;;;       means show nothing.  The default value is t.
;;;
;;; :stepping can be :timer or :model.  Determines how the model gets the game data.
;;;       In model mode the game waits for the model at each step, but in timer mode
;;;       the game just sends the updates when they happen based on the real time
;;;       clock.
;;;
;;;       When running with a person as one of the players :timer mode should be used
;;;       along with a speed of 1 (:model mode should also work with a speed of 1 for
;;;       playing with a person but it has the potential to affect the person's game).
;;;       When running two models :model mode should be used so that it can play as 
;;;       fast or slow as needed, and stepping or stopping won't affect the game time.
;;;       
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun update-game-count ()
  (bt:with-lock-held ((game-counter-lock *game-counter*))
    (pushnew (current-model) (game-counter-updated-players *game-counter*))
    (when (= (length (game-counter-updated-players *game-counter*))
             (game-counter-players *game-counter*))
      (incf (game-counter-current-game *game-counter*))
      (setf (game-counter-updated-players *game-counter*) nil))))

(defun initialize-game-counter (&optional (limit 0))
  (bt:with-lock-held ((game-counter-lock *game-counter*))
    (setf (game-counter-players *game-counter*) 0)
    (setf (game-counter-updated-players *game-counter*) nil)
    (setf (game-counter-ready-players *game-counter*) nil)
    (setf (game-counter-limit *game-counter*) limit)
    (setf (game-counter-current-game *game-counter*) 0)))

(defun increment-player-count ()
  (bt:with-lock-held ((game-counter-lock *game-counter*))
    (incf (game-counter-players *game-counter*))))

(defun check-game-count (x)
  ;; for safety allow it to have run over to catch for a bad update,
  ;; spurious result from the server, or other unforseen issue.
  (declare (ignore x))
  
  (bt:with-lock-held ((game-counter-lock *game-counter*))
    (>= (game-counter-current-game *game-counter*) (game-counter-limit *game-counter*))))


(defun play-sf (time &key (speed 1) (cont nil) (draw t) (stepping :model))
  (run-sf-coop time :time speed cont draw stepping))

(defun play-sf-games (games &key (speed 1) (cont nil) (draw t) (stepping :model))
  (run-sf-coop games :games speed cont draw stepping))


(defun run-sf-coop (value mode speed cont draw stepping)
  (if cont
      (progn
        (initialize-game-counter (if (eq mode :games) value 0))
        (dolist (m (mp-models))
          (with-model-eval m
            (let ((module (get-module :sfcoop)))
              (when (and module (bt:with-recursive-lock-held ((param-lock module)) (server-player module)))
                (increment-player-count)
                (bt:with-recursive-lock-held ((state-lock module))
                  (setf (change-drawing-mode module) (if draw 1 0)))))))
        (if (eq mode :games)
            (run-until-condition 'check-game-count speed)
          (run-full-time value speed)))
      (let (players)
        (reset)
        (initialize-game-counter (if (eq mode :games) value 0))
        (dolist (m (mp-models))
          (with-model-eval m
            (let ((module (get-module :sfcoop)))
              (awhen (and module (bt:with-recursive-lock-held ((param-lock module)) (server-player module)))
                     (if (find it players)
                         (progn
                           (print-warning "Duplicate player number ~s for model ~s.  Cannot run Space fortress." it m)
                           (return-from run-sf-coop))
                       (progn
                         (increment-player-count)
                         (push it players)
                         (unless (connect-to-game module draw stepping)
                           (print-warning "Error connecting model ~s to game." m)
                           (return-from run-sf-coop))))
                     
                     (schedule-next-update module)))))
        (if (eq mode :games)
            (run-until-condition 'check-game-count speed)
          (run-full-time value speed)))))


#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#
